(* Module implementing oracles with ATI *)
module E = Errormsg
module B = Bitmap 
module H = Hashtbl

open Logic

open Pretty

let debug = ref false
let reduceAssumptionMatches = false

let atiDepth = ref 3

class ['clause] clauseData (init: 'clause) = object
  method name = "<clauseData>"
  method conc = Bool true
  method hyps: pred list = []
  method maxUnif = -1
  method clause = init
end

let dclause () (c: 'clause clauseData) = 
  dprintf "Clause %s: %a%a  @[%a@]@!"
    c#name d_pred c#conc 
    insert (if c#hyps == [] then nil else line) 
    (docList ~sep:(chr ',' ++ line) (d_pred ())) c#hyps


(**** Nodes ****)
type ornode = 
    {         or_const: (string, andnode) H.t;  (* For constants *)
              or_any:   B.t;                    (* For a uvar *)
              or_all:   B.t;                    (* The union of all bitmaps 
                                                 * in this subtree *)
    } 

  (* Each andnode corresponds to the application of a constant. It has as 
     many ornode children as the arity of the constant *)
and andnode = 
    { mutable and_count: int;           (* The cardinality of the union of 
                                         * all bitmaps in this subtree. So we 
                                         * know when to collect this  *)
              and_data:  anddata;
    } 
     
and anddata = 
    Nullary of B.t
  | Nary of ornode list  (* One for each argument *)

(***** PRINTING OF NODES ******)
let hToList (h : ('a, 'b) H.t)  = 
  let acc : ('a * 'b) list ref = ref [] in
  H.iter (fun key data -> acc := (key, data) :: !acc) h;
  !acc

let intListRange () (il : int list) = 
  let rec loop start last = function
      [] -> [(start, last)]
    | i :: rest -> 
        if i = last + 1 then loop start i rest
        else (start, last) :: loop i i rest
  in
  let ranges = match il with
    [] -> []
  | i :: rest -> loop i i rest
  in
  docList ~sep:(chr ',' ++ break) 
    (fun (s, l) -> 
      if s = l then num s else 
      if s + 1 = l then num s ++ chr ',' ++ break ++ num l else
      num s ++ text "-" ++ num l) ()
    ranges

let rec dor () o =
  dprintf "OR@[(@[match: %a@])@!%a@]"
    intListRange (List.rev (B.toList o.or_all))
    (docList ~sep:line dand) (("_", { and_count = B.card o.or_any;
                                 and_data = Nullary o.or_any}) 
                         :: (hToList o.or_const))

and dand (n, a) = 
  let danddata = function
      Nullary bm -> dprintf "%s (@[%d match: %a@])"
          n a.and_count intListRange (List.rev (B.toList bm))
    | Nary ol -> 
        dprintf "%s@[%a@]" n 
          (docList ~sep:line (fun on -> text "-" ++ dor () on)) ol
  in
  dprintf "@[%a@]" insert (danddata a.and_data)


(****************************)


(****** SCANNING THE TREE ************)
  (* A generic expression scanner *)
let rec processExp (fv : subst -> 'a) 
                   (fc : subst -> string -> exp list -> int -> 'a) 
                   (s : subst) 
                   (level : int)
    : (exp -> 'a) = 
  let fc' s name args = 
    if level >= !atiDepth then 
      fv s
    else
      fc s name args level
  in
  function
      Int n -> fc' s (Int32.to_string n) []
    | Float f -> fc' s (string_of_float f) []
    | Var n -> E.s (E.bug "ATI: Found variable %d\n" n)
    | NVar n -> E.s (E.bug "ATI: Found variable %s\n" n)
    | UVar i -> begin 
        match s.(i) with
          Unif -> fv s
        | Done (e', s') -> processExp fv fc s' level e'
    end
    | UnOp(uo, e) -> fc' s (unop_to_string uo) [e]
    | BinOp(bo, e1, e2) -> fc' s (binop_to_string bo) [e1; e2]
    | Fun(n, el) -> fc' s n el
    | Ite(b,e1,e2) -> fc' s "ite" [b; e1; e2]
          
and processAtom fv fc s = function
  | PFun(n, el) -> fc s n el 0
  | PNez e -> processExp fv fc s 0 e
  | Bool true -> fc s "true" [] 0
  | Bool false -> fc s "false" [] 0
  | _ -> E.s (E.bug "ATI: found unexpected predicate")



type 'clause state = 
    {          initClause: 'clause clauseData; (* To be used for 
                                                * initialization *)
      mutable clauses: 'clause clauseData array;  (* Last one is the last 
                                                   * added *)
      mutable nextClause: int;   (* 0 <= ?? <= Array.length clauses *)
      root: ornode;                     (* The root of the ATI *)
    } 

let dumpState state = 
  ignore (E.log "ATI: state has %d clauses:\n" state.nextClause);
  Array.iteri (fun i c -> 
    if i < state.nextClause then
      ignore (E.log " %-3d @[%a@]" i dclause c))
    state.clauses;
  ignore (E.log "ATI: tree:\n");
  withPrintDepth 999 (fun () -> ignore (E.log "%a@!" dor state.root))
  

let getClause (state: 'clause state) (idx: int) : 'clause clauseData = 
  state.clauses.(idx)

   (* Add a clause to the state *)    
let addClause 
    (state:'clause state)
    ~(clause:'clause clauseData)
    : unit = 
  if state.nextClause >= Array.length state.clauses then begin
    let newdim = 64 + Array.length state.clauses in
    let na = Array.make newdim state.initClause in
    Array.blit state.clauses 0 na 0 state.nextClause;
    state.clauses <- na
  end;
  assert(state.nextClause < Array.length state.clauses);
  state.clauses.(state.nextClause) <- clause;
  let clauseid = state.nextClause in
  state.nextClause <- 1 + clauseid;
  if !debug then  
    ignore (E.log "ATI: adding clause %d: %a"
              (state.nextClause - 1)
              dclause clause);
  (* Now add it to the decision tree *)
  (* The constant processor *)
  let rec doConst node s n args level = 
    B.setTo node.or_all clauseid true;   (* Set this bit *)
    let andnode =                      (* Get the and_node for this constant*)
      try H.find node.or_const n with 
        Not_found -> begin
          let anode' = { and_count = 0;
                         and_data = 
                         if List.length args = 0 then 
                           Nullary (B.make state.nextClause)
                         else
                           Nary (List.map 
                                   (fun _ -> 
                                     { or_any = B.make state.nextClause;
                                       or_all = B.make state.nextClause;
                                       or_const = H.create 17 
                                     }) args) 
                       }
          in
          H.add node.or_const n anode';
          anode'
        end
    in
    andnode.and_count <- 1 + andnode.and_count;
    match andnode.and_data with
      Nullary bm -> begin 
        assert (List.length args = 0);
        B.setTo bm clauseid true
      end
    | Nary olist -> begin
        assert (List.length args = List.length olist && List.length args > 0);
        List.iter2 
          (fun arg onode -> 
            processExp (doVar onode) (doConst onode) s (level + 1) arg) 
          args olist
    end

  (* The variable processor *)
  and doVar node s = 
    B.setTo node.or_any clauseid true;
    B.setTo node.or_all clauseid true

  in
  processAtom (doVar state.root) 
              (doConst state.root) (newSubst (1 + clause#maxUnif)) clause#conc


  (**** DELETE A CLAUSE ****)
let deleteLastClause state = 
  assert(state.nextClause > 0);
  let clauseId = state.nextClause - 1 in
  let c = state.clauses.(clauseId) in
  if !debug then
    ignore (E.log "ATI: retracting clause %d (%s)@!" clauseId c#name);

(*
  state.clauses.(clauseId) <- dummyclause state.initClause;
*)
  (* To please the GC *)
  state.nextClause <- clauseId;

  (* The constant processor *)
  let rec doConst node s n args level = 
    B.setTo node.or_all clauseId false;   (* Reset this bit *)
    let andnode =                      (* Get the and_node for this constant*)
      try H.find node.or_const n with 
        Not_found -> E.s (E.bug "Cannot find node for %s while deleting clause"
                            n)
    in
    andnode.and_count <- andnode.and_count - 1;
    if andnode.and_count = 0 then
      H.remove node.or_const n 
    else begin
      match andnode.and_data with
        Nullary bm -> begin 
          assert (List.length args = 0);
          B.setTo bm clauseId false
        end
      | Nary olist -> begin
          assert (List.length args = List.length olist);
          List.iter2 
            (fun arg onode -> 
              processExp (doVar onode) (doConst onode) s (level + 1) arg) 
            args olist
      end
    end
    
  (* The variable processor *)
  and doVar node s = 
    B.setTo node.or_all clauseId false;
    B.setTo node.or_any clauseId false

  in
  processAtom (doVar state.root) 
              (doConst state.root) (newSubst (1 + c#maxUnif)) c#conc



  (********* FIND CHOICES FOR A GIVEN ATOM *******)
let findChoices (state: 'clause state) 
                ((goal: pred), (subst: subst)) : 'clause clauseData list = 
  
  (* The constant processor. Return a bitmap of choices *)
  let rec doConst node s n args level = 
    try (* Get the and_node for this constant*)
      let andnode = H.find node.or_const n in
      match andnode.and_data with
        Nullary bm -> 
          let newbitmap = B.clone bm in
          ignore (B.union newbitmap node.or_any);
          newbitmap
      | Nary olist ->
          let bitmaps_args = 
            List.map2 (fun arg onode -> 
              processExp (doVar onode) 
                (doConst onode) s (level + 1) arg) args olist   
          in
          match bitmaps_args with
            [] -> E.s (E.bug "empty olist")
          | bm :: rest -> 
              let newbitmap = B.clone bm in
              List.iter (B.inters newbitmap) rest;
              ignore (B.union newbitmap node.or_any);
              newbitmap
    with Not_found -> 
      node.or_any 

  (* The variable processor *)
  and doVar node s = node.or_all
  
  in
  let bmres = 
    processAtom (doVar state.root) (doConst state.root) subst goal in
  let choices = List.rev (B.toList bmres) in
  if !debug then 
    ignore (E.log "ATI: choices for %a are: @[%a@]@!"
              d_spred (goal, subst) intListRange choices);
  if choices = [] then 
    E.s (E.bug "No choices found for %a@!" d_spred (goal, subst));
  let choices' = 
    if not reduceAssumptionMatches then 
      choices
    else begin
      try
        (* If the atom does not have unification variables and if it matches 
           an assumption axiom also without unification variables then we 
           reduce to a singleton choice *)
        if List.length choices = 1 then choices
        else begin
          let achoices = 
            List.filter 
              (fun cidx -> 
                let cls = state.clauses.(cidx) in
                cls#maxUnif = -1 && List.length cls#hyps = 0 &&
                punify (fun _ -> raise HasUVar) 
                  (goal, subst) 
                  (cls#conc, emptySubst)) choices in
          match achoices with
            c :: _ -> begin 
              if !debug then 
                ignore (E.log "ATI: reduce choices to: @[%a@]@!" 
                          intListRange achoices);
              achoices                   (* One is sufficient *)
            end
          | _ -> choices
        end
      with HasUVar -> choices
    end
  in
  List.map (fun cidx -> state.clauses.(cidx)) choices'


    

let make (init: 'clause clauseData) : 'clause state = 
  { clauses    = Array.make 0 init;
    initClause = init;
    nextClause = 0;
    root = { or_any = B.make 256; 
             or_all = B.make 256;
             or_const = H.create 17; 
           };
  } 

    


