

(** This file contains the cache management *)

module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module O = Options

open Pretty
open Edag

let log fmt = ignore (E.log "CACHE: "); E.log fmt

(** Keep here the proof cache. For each goal we keep multiple entries, but we 
 * remember which entries are stale.  *)
type proofCacheEntry = { 
    pcAssumptions: (U.symbol * L.pred) list; 
                                (** The assumptions, along with their name. 
                                 * These names are bound in the proof, if 
                                 * present.  *)
    pcProof: P.proof option; (** None, if we did not create full proofs *)
    mutable pcStale: bool;  (** We use this flag to remember which cache 
                             * entries have been used in this run. We only 
                             * save those entries that were used in this run *)
  }
let proofCache: (L.pred, proofCacheEntry list ref) H.t ref = ref (H.create 113)
    
let useProofCache = ref false     

(** The name of the cache file *)
let cacheFileName = ref ""

(** Change the cache version number when you change the format of the cache, 
 * the format of the proofs, predicates.  *)
let currentCacheVersion = 5



let loadProofCache (fname: string) = 
  (* make the cache name *)
  cacheFileName:=(Filename.chop_extension fname) ^ ".cache";                   
  (* sm: drop occurrence of "errorN" so that during multitest, the
   * expected-erroneous variants will use the cache from the original run *)
  cacheFileName := (Str.global_replace (Str.regexp "\\.error[0-9]+\\.") "." 
                    !cacheFileName);
  (try 
    let f = open_in_bin !cacheFileName in
    let savedVersion = (Marshal.from_channel f : int) in
    if savedVersion <> currentCacheVersion then begin
      ignore (E.warn "Ignoring proof cache %s because it is for an old version %d (now %d)\n"
                !cacheFileName savedVersion currentCacheVersion);
      close_in f;
      raise Not_found
    end;
      
    proofCache := 
      (Marshal.from_channel f : (L.pred, proofCacheEntry list ref) H.t);
    close_in f;
    (* Mark all entries as stale *)
    H.iter (fun _ pcl -> 
      List.iter (fun pc -> pc.pcStale <- true) !pcl)
     ! proofCache;
    ignore (E.log "Loaded proof cache from %s\n" !cacheFileName);
  with _ -> 
    proofCache := H.create 113;
    ignore (E.warn "No proof cache found in %s" !cacheFileName))

let saveProofCache () = 
  assert (!cacheFileName <> "");
  try
    let f = open_out_bin !cacheFileName in
    (* Clear out the stale entries *)
    H.iter (fun _ pce -> 
      let rec filter f = 
        function [] -> [] 
               | h::t -> let t' = filter f t in if f h then h :: t' else t' 
      in
      pce := filter (fun pc -> not pc.pcStale) !pce;
    ) !proofCache;
      
    Marshal.to_channel f currentCacheVersion [];
    Marshal.to_channel f !proofCache [];
    close_out f;
    ignore (E.log "Saved proof cache to %s\n" !cacheFileName);
  with _ -> 
    ignore (E.warn "Error saving the proof cache %s" !cacheFileName)

let debugCache = false
let debugCacheHeavy = debugCache && true

(** And keep a table of the assumptions, indexed by hypothesis *)
let assumeTable: L.pred IH.t = IH.create 15
let assumeTablePred: (L.pred, U.symbol) H.t 
    = H.create 115 (** And indexed by predicate *)



let pushAssumption (assumvar: U.symbol) (assumpred: L.pred) : unit = 
  if !useProofCache then begin 
    IH.add assumeTable assumvar assumpred;
    H.add assumeTablePred assumpred assumvar;
    if debugCacheHeavy then 
      ignore (log "adding assumption %a\n" L.d_pred assumpred);
    registerUndoRun 
      (fun _ -> IH.remove assumeTable assumvar;
        if debugCacheHeavy then 
          ignore (log "removing %a\n" L.d_pred assumpred);
        H.remove assumeTablePred assumpred)
  end



let pFalseMagic = P.Magic("droped after irrelevant case", L.Bool false)
(* For each relevant hypothesis, get the predicate, and the name of the 
 * assumption in the proof. *)
let getRelevantHyps (prf: P.proof) : (U.symbol * L.pred) list = 
  let relevant = IH.create 13 in 
  let bound = IH.create 13 in 

  
  P.scanHyps 
    ~pushBound:(fun h -> IH.add bound h (); true)
    ~popBound:(fun h -> IH.remove bound h)
    ~seenHyp:(fun s -> 
      if not (IH.mem bound s) then 
        IH.replace relevant s ())
    ~seenFreeHyps:(fun rs p -> 
      computeFreeHyps rs p;
      P.Bitmap.iter 
        (fun i -> 
          IH.replace relevant (i + rs.P.relfirst) ())
        rs.P.relbm;
      false (* Do not descend, because the proof might be missing *))
    
    prf;

  (* Now look then up in the assumeTable *)
  IH.fold
    (fun h _ acc -> 
      let p = IH.find assumeTable h in 
      (h, p) :: acc)
    relevant
    []
 


exception MissingHypothesis of L.pred 

let cacheHits = ref 0 
let cacheMisses = ref 0
let totalCacheHits = ref 0
let totalCacheMisses = ref 0

let checkProofCache (goal: L.pred) 
                    (invalid:bool) 
                    (proveit: L.pred -> P.proof option) : P.proof option = 
  (* Reset the relevant hypotheses. In the process of finding the proofs 
   * we will accumulate there *)
  if !useProofCache then begin
    (* Prepare the entry for the goal *)
    let goalentries: proofCacheEntry list ref = 
      try 
        let r = H.find !proofCache goal in
        if debugCache then 
          ignore (log "found %d entries for the goal in the cache\n" 
                    (List.length !r));
        r
      with Not_found -> begin 
        if debugCache then 
          ignore (log "did not find the goal in the cache. We will add it\n");
        let r = ref [] in 
        H.add !proofCache goal r; 
        r
      end
    in
    
    (* Now check that we have the hypotheses around in some entry. If we do, 
     * move the entry to the front. *)
    let rec loopEntries = function
        [] -> 
          if debugCache then begin
            ignore (log "  all entries failed. current hypotheses are\n @[%a@]\n"
                      (docList ~sep:line (fun (p, h) -> L.d_pred () p)) 
                      (U.hash_to_list assumeTablePred));
            
          end;
          None

      | pc :: restpc -> begin
          (* Check one entry in the cache. Return a list of current names for 
           * the assumptions that are needed (in reverse order). *)
          let rec loopAssume 
              (acc: (U.symbol * (U.symbol * L.pred)) list) 
              (assums: (U.symbol * L.pred) list) : 
              (U.symbol * (U.symbol * L.pred)) list option = 
            match assums with
              [] -> Some acc
            | ((_, a) as thisa) :: resta -> begin
                try
                  let assumeh = H.find assumeTablePred a in 
                  if debugCacheHeavy then 
                    ignore (log "  found hypothesis %a\n" L.d_pred a);
                  loopAssume ((assumeh, thisa) :: acc) resta
                with Not_found -> begin 
                  if debugCache then begin
                    ignore (log "  missing hypothesis %a\n" L.d_pred a);
                  end;
                  None
                end
            end
          in
          match loopAssume [] pc.pcAssumptions with 
            Some assume_map -> begin
              if debugCache then 
                ignore (log "  found all %d hypotheses\n"
                          (List.length pc.pcAssumptions));
              Some (pc, assume_map)
            end

          | _ -> loopEntries restpc
      end
    in
    match loopEntries !goalentries with 
      Some (pc, assume_map) -> begin (* found a match *)
        incr cacheHits; incr totalCacheHits;
        pc.pcStale <- false; 
        match pc.pcProof with 
          Some p when !P.checkProofs -> 
            (* Now use the assume_map to create the right proof *)
            Some 
              (List.fold_left
                 (fun prf (assumeh, (prfh, hpred)) -> 
                   if assumeh == prfh then
                     prf
                   else
                     P.Cut(P.Hyp assumeh, prfh, prf))
                 p
                 assume_map)

        | _ -> Some noProof
      end
          
    | None -> begin (* no perfect match *)
        incr cacheMisses; incr totalCacheMisses;
        let prfo = proveit goal in 
        (* Add it to the cache, unless we had an invalid goal *)
        (match prfo with
          Some prf when not invalid -> begin
            (* Find the relevant hypotheses *)
            let relh: (U.symbol * L.pred) list = getRelevantHyps prf in 
            let entry = 
              { pcProof = (* Save proofs only if we have complete proofs *)
                (if !P.checkProofs then Some prf else None);
                
                pcStale = false;
                pcAssumptions = relh
              }
            in
            
            (* put it first *)
            if debugCacheHeavy then 
              ignore (log "  add entry with %d hypotheses:\n  @[%a@]\n" 
                        (List.length entry.pcAssumptions)
                        (docList ~sep:line 
                           (fun (_, a) -> L.d_pred () a)) entry.pcAssumptions);
            goalentries := entry :: !goalentries;
          end
        | _ -> ());

        prfo
    end
  end else 
    proveit goal


let start (fname: string) = 
  if !useProofCache then begin
    if not !genProofs then begin
      ignore (E.warn "Turning off the use of caches because you turn off generation of proofs\n");
      useProofCache := false;
    end else 
      loadProofCache fname
  end

let printCacheStats () = 
  ignore (E.log " Cache hit ratio: %d/%d\n" 
            !totalCacheHits
            (!totalCacheHits + !totalCacheMisses));
  ()


let options = [
  { O.optInUI = "";
    O.optRestart = false;
    O.optKind = O.OBool useProofCache;
    O.optExtra = (fun _ -> ());
    O.optCommandLine = "-cache";
    O.optHelp = "use proof caches";
  };
]
