%{
module I = Input
module C = Cvc
module E = Errormsg
open Pretty


let parse_error msg : unit =
  E.parse_error msg


%}


%token <string> Id STRING
%token <int> Num
%token LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE SEMICOLON  ASSIGN
%token <Input.loc> QUERY QUERY_INVALID ASSERT RETRACT BLOCK ENDBLOCK IMPL ALL 
%token <Input.loc> TRIGGER EQ
%token REAL BOOLEAN NOT IFF XOR IF THEN ELSE LET IN
%token COLON ENDIF ELSIF COMMA ARROW SQHASH HASHSQ
%token PLUS MINUS MULT STRING
%token MOD DIV FLOOR CEIL
%token LT GT LEQ GEQ NEQ
%token TRUE FALSE AND OR EXISTS
%token EOF DOT 

%token ARRAY DATATYPE LAMBDA OF TYPE WITH TCC
%token DBG TRACE OPTION HELP TRANSFORM PRINT CALL ECHO INCLUDE
%token DUMP_ASSUMPTIONS DUMP_PROOF DUMP_SIG WHERE PUSH POP POPTO
%token RESET CONTEXT FORGET GET_TYPE CHECK_TYPE END CHECK_TYPE
%token GET_CHILD GET_OP SUBSTITUTE

%nonassoc LAMBDA
%nonassoc ALL

%nonassoc IN
%left IFF
%right IMPL
%left OR XOR 
%left AND
%left NOT

%left WITH
%nonassoc EQ NEQ
%nonassoc LEQ LT GEQ GT
%left PLUS MINUS
%left MULT INTDIV DIV MOD
%left UMINUS

%nonassoc LBRACKET LBRACE LPAREN

%start input

%type <Cvc.expr> expr

%type <Cvc.input> input
%%
 
  
  /*(* Here is the parser for one element of the file *)*/  
input: 
/*(* Type declarations *)*/
| Id COLON TYPE   SEMICOLON          { C.TypeConstDecl ([$1]) }
| Id COMMA ids COLON TYPE SEMICOLON  { C.TypeConstDecl ($1 :: $3) }
| Id COLON TYPE EQ typ SEMICOLON     { C.TypeConstInitDecl ($1, $5) }

/*(*Constant declarations *)*/
| Id COLON typ SEMICOLON             { C.ConstDecl    ([$1], $3) }
| Id COMMA ids COLON typ SEMICOLON   { C.ConstDecl    ($1 :: $3, $5) }
| Id COLON typ EQ expr SEMICOLON     { C.ConstInitDecl ($1, $3, $5, $4) }

            /*(* A query *)*/
| QUERY expr SEMICOLON             { C.Query ($2, $1) }
| QUERY_INVALID expr SEMICOLON;    { C.QueryInvalid ($2, $1) }
| ASSERT expr SEMICOLON            { C.Assert ($2, $1) }
| RETRACT SEMICOLON                { C.Retract $1 } 
| PRINT expr SEMICOLON             { C.Print $2 }
| ECHO STRING SEMICOLON            { C.Echo $2 }
| BLOCK LBRACKET block_pars RBRACKET 
                                   { C.Block (("truefile", $1.E.file) ::
                                              ("trueline", 
                                               string_of_int $1.E.line) :: $3, 
                                              $1) }
| ENDBLOCK                         { C.EndBlock $1 }
| EOF                              { C.Eof }
;

block_pars: 
     /*empty*/                          { [] }
  |  block_pars_ne                      { $1 }
  |  error                              { [] }
;
block_pars_ne:
  |  block_par                          { [ $1 ] }
  |  block_par COMMA block_pars_ne      { $1 :: $3 }
;
block_par:
     Id COLON STRING                    { ($1, $3) }
  |  Id COLON Num                       { ($1, string_of_int $3) }
  |  Id EQ STRING                       { ($1, $3) }
  |  Id EQ Num                          { ($1, string_of_int $3) }
;

ids: 
| Id                              { [$1] }
| Id COMMA ids                    { $1 :: $3 }
;

typ :
| Id                                  { C.TId $1 }
| REAL                                { C.TReal }
| BOOLEAN                             { C.TBool }
| LBRACKET typ ARROW typ RBRACKET     { C.TFun ($2, $4) }
| LBRACKET typs RBRACKET              { C.TTuple ($2) }
| SQHASH fields HASHSQ                { C.TRecord ($2) }
| ARRAY typ OF typ                    { C.TArray ($2, $4) }
;

typs:
  typ                                 { [$1] }
| typ COMMA typs                      { $1 :: $3 }
;

fields:
  Id COLON typ                        { [($1, $3)] }
| Id COLON typ COMMA fields           { ($1, $3) :: $5 }
;

expr:
| Id                                  { C.EId $1 } 
| Num                                 { C.ENum $1 }
| TRUE                                { C.EBool true }
| FALSE                               { C.EBool false }
| NOT expr                            { C.ENot $2 }
| LPAREN expr RPAREN                  { $2 }
| expr EQ expr                        { C.EEq ($1, $3) }
| expr NEQ expr                       { C.ENeq ($1, $3) }
| expr LT expr                        { C.ELt ($1, $3) }
| expr LEQ expr                       { C.ELeq ($1, $3) }
| expr GT expr                        { C.EGt ($1, $3) }
| expr GEQ expr                       { C.EGeq ($1, $3) }
    
| expr AND expr                       { C.EAnd ($1, $3) }
| expr OR expr                        { C.EOr ($1, $3) }
| expr IMPL expr                      { C.EImpl ($1, $3) }
| expr IFF expr                       { C.EIff ($1, $3) }
| expr XOR expr                       { C.EXor ($1, $3) }

| expr PLUS expr                      { C.EPlus ($1, $3) }
| expr MINUS expr                     { C.EMinus ($1, $3) }
| expr MULT expr                      { C.EMult ($1, $3) }
    
| IF expr THEN expr elserest          { C.EIf($2, $4, $5) }
| LET letdecls IN expr                { C.ELet ($2, $4) } 

| expr WITH indices ASSIGN expr   %prec WITH
                                      { C.EUpd($1, $3, $5) }
| expr LBRACKET expr RBRACKET         { C.ESel($1, $3) } 

| expr tupleliteral                   
           { match $1 with 
               C.EId n -> C.EApp (n, $2)
             | _ -> 
                 parse_error "only constant functions are supported";
                 C.EApp ("error", $2) 
           }
| ALL Id COLON typ DOT expr  %prec ALL      
                                      { C.EAll($2, $4, $6) }
| ALL LPAREN Id COLON typ RPAREN COLON expr  %prec ALL      
                                      { C.EAll($3, $5, $8) }
| EXISTS Id COLON typ DOT expr  %prec ALL      
                                      { C.EExist($2, $4, $6) }

| EXISTS LPAREN Id COLON typ RPAREN COLON expr  %prec ALL      
                                      { C.EExist($3, $5, $8) }

| TRIGGER LPAREN expr COMMA expr RPAREN
                                      { C.EAnnot(C.ATrigger $3, $5) }
| LAMBDA LPAREN fields RPAREN COLON expr %prec LAMBDA
                                      { C.ELambda($3, $6) }
;

tupleliteral: 
  LPAREN exprs RPAREN                 { $2 }
;

indices: 
  LBRACKET expr RBRACKET indices_empty { $2 :: $4 }
;
indices_empty: 
  /*(*empty*)*/                        { [] }
| indices                              { $1 }
;

exprs:
  expr                                { [$1] }
| expr COMMA exprs                    { $1 :: $3 }
;

elserest: 
| ELSE expr ENDIF                    { $2 }
| ELSIF expr THEN expr elserest      { C.EIf($2, $4, $5) }
;

letdecls: 
  Id COLON typ EQ expr                  { [($1, $3, $5)] }
| Id COLON typ EQ expr COMMA letdecls   { ($1, $3, $5) :: $7 }
;





