module Eng = Engine

open Edag
open Input
open Pretty

(** LF generation of proofs *)
(*** LFi proof generator ****)
type pstate = 
    Assume
  | Conjunct of int              (* How many unclosed andi there are *)

type lfpgen = 
    { mutable proofs: pstate list;
    } 

let lfstate =                           (* The global state of the LFi proof 
                                         * generator *)
  { proofs = [];
  } 
let indentProofs = ref false

let proofChannel = ref stderr

let start (proofFile: string) = 
  proofChannel := 
    if proofFile = "" then stderr else 
    (try open_out proofFile with _ ->
      raise (Arg.Bad "Cannot open proof file"));
  ignore (fprintf !proofChannel
            "ANN_STARTLF@!IMPORT %s @!ANN_ENDLF@!"
            "")


let finish () = 
  if !proofChannel != stderr then close_out !proofChannel 

let startVC name = 
  lfstate.proofs <- [Conjunct 0];
  let name' = 
    if String.length name >= 4 && String.sub name 0 3 = "Sat" then
      String.sub name 3 (String.length name - 3)
    else
      name
  in
  ignore (fprintf !proofChannel
            "\nANN_PROOF(%s,@! %%LF_" name')

let printLF thunk = 
  let olf = !printLFSyntax in
  printLFSyntax := true;
  let opd = !printDepth in
  printDepth := 999999;
  let opi = !printIndent in
  printIndent := !indentProofs;
  thunk ();
  printLFSyntax := olf;
  printDepth := opd;
  printIndent := opi

let assume (mkname: unit -> string) (p: L.pred) = 
  printLF 
    (fun () -> 
      let name = mkname () in
      ignore (fprintf !proofChannel
                "(andi (impi [%s : pf @[%a@]]@!" name 
                L.d_pred p));
  match lfstate.proofs with
    Conjunct n :: rest -> 
      lfstate.proofs <- Conjunct 0 :: Assume :: Conjunct (n + 1) :: rest
  | _ -> E.s (E.bug "expecting conjunct in assume")

let finishConjuncts () =
  let rec nclosepar n = 
    if n = 0 then nil else text ")" ++ nclosepar (n - 1) 
  in
  match lfstate.proofs with
  | Conjunct n :: rest -> 
      lfstate.proofs <- rest;
      printLF 
        (fun () -> 
          (ignore (fprintf !proofChannel 
                     "truei%a" insert (nclosepar n))))
  | _ -> E.s (E.bug "expecting conjunct in finishConjuncts")
  
let unassume () = 
  (* Finish off the IMPI proof *)
  finishConjuncts ();
  ignore (fprintf !proofChannel ")@!");
  match lfstate.proofs with
    Assume :: rest -> lfstate.proofs <- rest
  | _ -> E.s (E.bug "Conjunct not finished")

let pick v = 
  ignore (fprintf !proofChannel
            "(alli [%s : _]@!" v)

let unpick () = 
  ignore (fprintf !proofChannel ")@!")

     
let finishVC () = 
  finishConjuncts ();
  ignore (fprintf !proofChannel "%%_LF)\n\n");
  match lfstate.proofs with
    [] -> ()
  | _ -> 
      ignore (E.warn "There are still %d un-retracted proofs in finishVC"
                (List.length lfstate.proofs));
      List.iter (fun _ -> ignore (fprintf !proofChannel ")@!");) 
        lfstate.proofs;
      lfstate.proofs <- []

let conjunct (p: L.pred) (proof: P.proof) = 
  match lfstate.proofs with
    Conjunct n :: rest -> begin
      printLF 
        (fun () -> 
          ignore (fprintf !proofChannel "(andi @[%a@]@!" P.d_prf proof));
      lfstate.proofs <- Conjunct (n + 1) :: rest
    end
  | _ -> E.s (E.bug "expecting a conjunct")


let lfiProofGen : Eng.proofgen = 
  { Eng.startVC = startVC;
    Eng.start = start;
    Eng.finish = finish; 
    Eng.assume = assume;
    Eng.unassume = unassume;
    Eng.pick = pick;
    Eng.unpick = unpick;
    Eng.finishVC = finishVC;
    Eng.conjunct = conjunct;
    Eng.choose = 
        fun ~thechoice ~nrchoices -> E.s (E.bug "choose is not supported with LF")
  } 
