open Pretty 
module E = Errormsg
module I = Input
module L = Logic
module S = Satproc
module Eng = Engine
module P = Proof
module U = Util

 (* The top-level function of the prover *)
(* The module Features is produce by the Makefile based on ./configure *)  
let allSatProcs = Features.allSatProcs

let debug = false



    (* A way to collect argument options *)
let rec mapConcat (f: 'a -> 'b list) (lst: 'a list) : 'b list = 
  match lst with 
    [] -> []
  | h :: t -> f h @ mapConcat f t


let proofGen : Eng.proofgen option ref = ref None 

(***** MAIN *****)  
let rec theMain () =
  let usageMsg = "Usage: kettle [options] files" in
  let files : string list ref = ref [] in
  let recordFile fname = files := fname :: (!files) in
  let openLog lfile = 
    if !E.verboseFlag then
      ignore (Printf.printf "Setting log file to %s\n" lfile);
    try E.logChannel := open_out lfile with _ -> 
      raise (Arg.Bad "Cannot open log file") in
  (* Initially all loaded procedures are enabled *)
  let includes : U.symbol list ref = 
    ref (List.fold_left 
           (fun acc sp -> 
             if sp.S.defaultEnabled then sp.S.ident :: acc else acc)
           []
           allSatProcs)
  in
  let enableProc name =
    if name = "ALL" then 
      includes := (List.map (fun p -> p.S.ident) allSatProcs)
    else begin
      let nid = U.registerSymbolName name in 
      if not (List.exists (fun sp -> sp.S.ident = nid) allSatProcs) then
        raise (Arg.Bad ("Don't know anything about satproc " ^ name));
      includes := nid :: !includes 
    end
  in
  let disableProc name = 
    if name = "ALL" then 
      includes := []
    else begin
      let nid = U.registerSymbolName name in 
      if not (List.exists (fun sp -> sp.S.ident = nid) allSatProcs) then
        raise (Arg.Bad ("Don't know anything about satproc " ^ name));
      includes := (List.filter (fun x -> x <> nid) !includes)
    end
  in
  let importString : string ref = ref "" in
  let listProcs () = 
    ignore (E.log "Kettle uses satprocs:@!  @[EG   : main prover engine,@!%a@]\n"
        (docList ~sep:(chr ',' ++ line) (fun sp -> dprintf "%s   : %s"
            (U.symbolName sp.S.ident) sp.S.descr)) allSatProcs) in
  let setDebugFlag v name = 
    if name = "EG" then
      Eng.debug := v
    else if name = "CA" then 
      Cases.debug := v
    else if name = "MA" then 
      Matching.debug := v
    else if name= "DAG" then 
      Edag.debug := v
    else 
      let nid = U.registerSymbolName name in 
      let rec loop = function
          sp :: _ when sp.S.ident = nid -> sp.S.debug := v
        | _ :: rest -> loop rest
        | [] -> E.s (E.error "Don't know about satproc %s\n" name)
      in
      loop allSatProcs
  in
  let axiomFiles : string list ref = ref [] in
  let setAxiomFile af = axiomFiles := af :: !axiomFiles in
  let setProofGen (which: string) = 
    if which = "none" then proofGen := None
    else if which = "lfi" then proofGen := Some Lfprf.lfiProofGen
    else if which = "oracle" then proofGen := Some Oracleprf.oracleProofGen 
    else if which = "coq" then proofGen := Some Coqprf.coqProofGen 
    else E.s (E.error "Don't know about proof generator %s\n" which)
  in
  let usePerfCount = ref false in
  (*********** COMMAND LINE ARGUMENTS *****************)
  (* Prepare the UI options *)
  let uiOptions = 
    mapConcat
      (fun (ui: Eng.uiType) -> 
        let base = List.concat (List.map Options.optionToArgs ui#options) in
        if ui#name <> "" then begin
          ("-" ^ ui#name,
           Arg.Unit (fun _ -> Eng.theUI := ui),
           "use " ^ ui#name ^ " interface") ::
          base
        end else
          base)
      Features.uis
  in

  let argDescr = [
    "-listprocs", Arg.Unit listProcs, "list the loaded satprocs";
    "-debug", Arg.String (setDebugFlag true), 
                     "<xxx> turns on debugging for proc xxx";
    "-verbose", Arg.Unit (fun _ -> E.verboseFlag := true), 
                "turn on verbose mode";
    "-timing", Arg.Set Eng.timingFlag, 
                "print timing information, and reduce printing";
    "-nodebug", Arg.String (setDebugFlag false), 
                      "<xxx> turns off debugging for proc xxx";
    "-indent", Arg.Unit (fun _ -> Lfprf.indentProofs := true),
                "turn on indentation in printed proofs";
    "-log", Arg.String openLog, "the name of the log file";
    "-oimport", Arg.String (fun s -> importString := s), 
               "additional LF files to be added to the proof's import list";
    "-enable", Arg.String enableProc,  "name of satproc to enable (ALL for all loaded)";
    "-disable", Arg.String disableProc,  "name of satproc to disable (ALL for all loaded)";

    "-printinfix", Arg.Set Eng.printInfix,
                "print expressions using infix notation";
    "-printprefix", Arg.Clear Eng.printInfix,
                "print expressions using infix notation";
             
    "-pfgen", Arg.String setProofGen, 
                "select the proof generator: (none, oracle, lfi)";
    "-oracleann", Arg.Unit (fun x -> Oracleprf.useOracleAnn := true), 
                "use oracle for annotations as well";
    "-atilevel", Arg.Int (fun a -> Ati.atiDepth := a), 
                "set the maximum depth of ATI oracles";
    "-axioms", Arg.String setAxiomFile,
                 "specify the file containing the axioms";
    "-no-witness", Arg.Unit (fun _ -> Eng.printWitness := false),
                 "do not print the witness on failure";
    "-assumptions", Arg.Unit (fun _ -> Eng.printAssumptions := true),
                 "print assumptions on failure";
    "-perfcount", Arg.Unit (fun _ -> usePerfCount := true),
                 "use the performance counters";
    "-checkproofs", Arg.Unit (fun _ -> P.checkProofs := true;
                                       Edag.genProofs := true),
                 "whether Kettle should check proofs";
    "-noproofs", Arg.Unit (fun _ -> Edag.genProofs := false; 
                                    P.checkProofs := false),
                 "whether Kettle should generate proofs";
    "-nostop", Arg.Unit (fun _ -> Eng.continue_on_error := true;
                                  Eng.continue_without_asking := true),
                 "continue after errors";
    "-timeout", Arg.Set_int Eng.timeout, 
                 "number of seconds for each goal";
    "-pause", Arg.Unit (fun _ -> Eng.continue_on_error := true;
                                 Eng.continue_without_asking := false),
                 "continue after errors";
    "-stdin", Arg.Unit (fun _ -> recordFile "-"),
          "read goals from stdin";
    "-debugstats", Arg.Unit (fun _ -> Edag.debugStats := true),
          "collect stats for debugging";
    "-trace", Arg.Unit (fun _ -> Edag.doProofTrace := true),
          "print a proof trace";
  ] @ (List.concat (List.map Options.optionToArgs 
                      (Engine.options @
                       Cache.options))) 
    @ Features.options 
    @ uiOptions
  in
  (* Add the parser features *)
  S.selectedParser :=
      (* Select the first one *)
    (match Features.allParsers with 
      p :: _ -> p
    | _ -> E.s (E.bug "no parsers are present"));

  let argDescr = 
    argDescr @
    (List.map 
       (fun p -> 
         ("-" ^ p.S.language ^ "input", 
          Arg.Unit (fun _ -> S.selectedParser := p), 
          "parse " ^ p.S.language ^ " input"))
       Features.allParsers)
  in 

  let argDescr = 
    argDescr @
    (mapConcat Options.optionToArgs Cases.options)
  in
    
  let argDescr = 
    argDescr @
    (mapConcat
       (fun sp -> mapConcat Options.optionToArgs sp.S.options)
       Features.allSatProcs)
  in
  Arg.parse argDescr recordFile usageMsg;

  Stats.reset !usePerfCount;
  
  (* Grab the sat procs to be used. *)
  let satprocs = 
    List.filter 
      (fun sp -> 
        List.mem sp.S.ident !includes 
          && sp.S.canEnable !includes
      ) 
      allSatProcs  
  in
  Logic.usePFun := false;

  (* Now initialize the satprocs *)
  Edag.initialize ();
  Engine.init ();
  Matching.init ();
  Cases.init ();
  List.iter (fun sp -> sp.S.init ()) satprocs;

  if not !Edag.genProofs && !Cache.useProofCache then begin
    ignore (E.warn "Turning off the proof cache because we don't generate proofs\n");
    Cache.useProofCache := false;
  end;

  ignore (E.log "Starting Kettle using satprocs: %a\n"
            (docList (fun sp -> text (U.symbolName sp.S.ident))) satprocs);

  
  (* Construct the list of files to look at. Start with the axiom files *)
  files := (List.rev !axiomFiles) @ (List.rev !files);
  
  (* Write a function that serializes the result from the parser *)
  let inputs : I.input list ref = ref [] in
  let rec getNextInput lexbuf = 
    match !inputs with 
      i :: rest -> 
        inputs := rest; 
        (* ignore (E.log "Got %a\n" I.di i); *)
        i
    | [] -> (* Refresh *)
        inputs := !S.selectedParser.S.nextFewInputs lexbuf;
        assert (!inputs <> []);
          getNextInput lexbuf
  in
  
  (* Now process the global assumptions *)
  List.iter (fun sp -> 
    match sp.S.globalAssumeString () with 
      None -> ()
    | Some (pname, str) -> begin
        (* Get the corresponding parser *)
        S.selectedParser := 
            (try 
              List.find
                (fun p -> p.S.language = pname)
                Features.allParsers
            with Not_found -> 
              E.s (E.bug "Satproc %s needs non-existent parser %s for its global assumptions"
                     sp.S.descr pname));
        let fname = "globalAssume for " ^ sp.S.descr in
        let lexbuf = E.startParsingFromString ~file:fname  str in
        !S.selectedParser.S.initParser fname lexbuf;
        inputs := [];
        let rec doGlobAssume () = 
          match getNextInput lexbuf with 
            I.Eof -> ()
          | I.StartGoal _ -> (); doGlobAssume ()
          | I.Assume (p, l) -> 
              Eng.globalAssume p l;
              doGlobAssume ()
          | i -> 
              E.s (E.error "unexpected input in global assumptions: %a\n" 
                     I.di i); 
              doGlobAssume ()
        in
        doGlobAssume ();
        E.finishParsing ()
    end)
    satprocs;

  let rec processOneFile fname = 
    let engstate = Eng.start fname satprocs !proofGen in

    let lexbuf = E.startParsing ~useBasename:false fname in
    (* Now skip to the first LF token *)
    !S.selectedParser.S.initParser fname lexbuf;
    
    let getNextInput () = getNextInput lexbuf in 

    (* iterate over the toplevel forms in the file *)
    let rec doInputs () = 
      match getNextInput () with
        I.Axiom ax -> begin
          (match !proofGen with Some pg when pg == Oracleprf.oracleProofGen -> 
            Oracleprf.Oracle.addAxiom Oracleprf.ostate.Oracleprf.oracle ax
          | _ -> ());
          doInputs ()
        end
      | I.Import _ -> doInputs ()
      | I.Decl _ -> doInputs ()
      | I.Eof -> ()
      | I.TopChoose(i, nr) -> begin
          Eng.choose i nr;
          doInputs()
      end
      | I.StartGoal (name, l) -> begin (* Now parse the elements of the goal *)
          (* Invoke the engine to prove this VC *)
          Eng.currLoc := l;
          if !E.verboseFlag then 
            ignore (E.log "Proving goal %s\n" name);
          Stats.time "proveGoal" 
            (Eng.proveGoal name) getNextInput;
          doInputs ();
      end
      | I.Block (args, l) -> 
          Eng.currLoc := l;
          !Eng.theUI#enterBlock [] args; doInputs ()
            
      | I.UnBlock l -> 
          Eng.currLoc := l;
          !Eng.theUI#exitBlock; doInputs ()

      | e -> E.s (E.bug "Input %a occurs at top-level" I.di e)
    in

    doInputs ();
    E.finishParsing ();
    Parsing.clear_parser ();
    Eng.finish engstate;
  in
  List.iter processOneFile !files;

  if !Eng.timingFlag || !E.verboseFlag then begin 
    Stats.print stderr "Timings:\n";
    Cache.printCacheStats ();
  end;



  !Eng.theUI#run;

;;

(* Define here code to be called before termination *)
let cleanup () : unit = 
  if !E.logChannel != stderr then 
    close_out (! E.logChannel); 
  (match !proofGen with 
    Some p -> p.Eng.finish ()
  | _ -> ());
  ()
in

theMain ();
cleanup ();
let _= 
  if not !E.hadErrors then begin
    ignore (E.log "Kettle success\n");
    exit 0;
  end else begin
    ignore (E.log "Kettle failure\n");
    exit 1;
  end
in
()
;;
