module E = Errormsg
module H = Hashtbl
module IH = Inthash
module P = Proof

open Satproc
open Logic
open Pretty
open Edag


let matchid = Util.registerSymbolName "MA"        (* satproc id *)

let debug = ref false

let log fmt = ignore (E.log "MA: "); E.log fmt
let error fmt = ignore (E.log "MA: "); E.error fmt
let bug fmt = ignore (E.log "MA: "); E.bug fmt
let unimp fmt = ignore (E.log "MA: "); E.unimp fmt
let warn fmt = ignore (E.log "MA: "); E.warn fmt

let debugTrigger = true

(** We interpret matching programs ***)
type matchInstr = 
     MatchLeaf of int * int
        (** [MatchLeaf(v, c)] matches a leaf [c] and assigns the root of 
         * its class to [v] **)

   | MatchConstr of int * int * var list * var option * var list 
         (** [MatchConstr(v,head,args,child,inst)] matches a node 
          * [f(a1,...,an)] and assigns the root to [v]. Optionally we 
          * restrict the node to be among the parents of [child]. We also 
          * give the list of pattern variables that are instantiated *)

   | MatchCheck of binop * var * var
         (** Succeeds only if the two arguments are manifestly equal or not 
          * equal, depending on the binop *)


and arg = 
     APatVar of int  (** A patern variable *)
   | ANode of Edag.node (** A node *)

and matchBlock = 
   | MInst of int * matchInstr  (** a program counter and an instruction *)

   | MAnd of matchBlock list     (** a conjunction **)

   | MOr of matchBlock list

and var = int


and matchProg = 
    { mBlock: matchBlock;
      mNrPattVars: int;  (** The number of pattern variables *)
      mBody: pred; (** The body of the matching rule *)
      mId: int; (** The id of the matching rule *)
      mPatVarName: string Inthash.t;
      mForallProof: Proof.proof; 
      mForallPred: pred; 
           (** Given a mapping from variables to expressions, make a proof of 
            * this body *)
    } 


let currentProg: matchProg ref = 
  ref { mBlock = MAnd [];
        mId = (-1);
        mNrPattVars = 0;
        mBody = Bool true;
        mForallProof = P.Unimplemented "matching";
        mForallPred = L.Bool true;
        mPatVarName = IH.create 1 }

(** Print a matching program **)
let d_var () (v: var) = 
  let n = 
    if v < !currentProg.mNrPattVars then 
      (":" ^ (try IH.find !currentProg.mPatVarName v 
              with Not_found -> "@@not found@@"))
    else
      ""
  in
  num v ++ text n

let d_instr () = function
    MatchLeaf (v, l) -> dprintf "%a <- leaf %s" 
        d_var v (headToString l)
  | MatchConstr (v, h, args, po, unif) -> 
      dprintf "%a <- constr %s(%a) parentof=%a unif=@[%a@]"
        d_var v (headToString h)
        (docList (d_var ())) args
        insert 
        (match po with 
          None -> text "None"
        | Some v -> d_var () v)
        (docList (d_var ())) unif
  | MatchCheck (bop, v1, v2) -> 
      dprintf "check %a %s %a" 
        d_var v1 
        (match pbinop_to_string bop with 
          None -> binop_to_string bop
        | Some bs -> bs) 
        d_var v2

let rec d_block () = function
    MInst (pc, i) -> dprintf "%d: %a\n" pc d_instr i
  | MAnd bl -> dprintf "@[And(@[%a@])@]\n"
        (docList ~sep:nil (d_block ())) bl
  | MOr bl -> dprintf "@[Or(@[%a@])@]\n"
        (docList ~sep:nil (d_block ())) bl

let printCurrentProg () = 
  let p = !currentProg in 
  ignore (E.log "Rule %d, nrPattVars:%d, body: %a\n  %a\n"
            p.mId p.mNrPattVars d_pred p.mBody
            d_block p.mBlock)

let separateArgumentChecks = true

module BM = Bitmap

exception FoundCovering of L.exp

let inferTrigger (depth: int) (p: L.pred) : L.pred = 
  if debugTrigger then 
    ignore (log "Find trigger for %a\n" d_pred p);
  
  (* Compute the set of free variables in an expression. Raise the 
   * FoundCovering exception with a covering expression. *)
  let rec getFreeVars (e: L.exp) : BM.t = 
    (* Do some subexpressions *)
    let empty = BM.make depth in (* Careful how you share this *)

    let rec doSubexps (el: L.exp list) : BM.t = 
      let acc = BM.make depth in 
      (* Do first each expression, and accumulate the result *)
      let rec loopSubexps = function 
          [] -> (* We get here only if no individual subexpression is 
                 * covering. See if this expression is covering *) 
            if BM.card acc = depth then
              raise (FoundCovering e)
            else
              acc
        | e' :: reste -> 
            let e's = getFreeVars e' in 
            (* Only there is no covering in e' *)
            ignore (BM.union acc e's);
            loopSubexps reste
      in
      loopSubexps el
    in
    match e with 
    | L.Var idx -> 
        assert (0 <= idx && idx < depth);
        let r = BM.make depth in 
        BM.setTo r idx true;
        r

    | L.UnOp (_, e1) -> getFreeVars e1

    | L.BinOp (_, e1, e2) -> doSubexps [e1; e2]

    | L.Fun(_, el) -> doSubexps el

    | L.Ite (e1, e2, e3) -> assert false

    | L.Int _ | L.Float _ | L.NVar _ | L.UVar _ -> empty

  in (* getFreeVars *)

  (* Scan a predicate for a covering expression *)
  let rec scanPred (p: L.pred) = 
    match p with 
      Bool _ -> ()
    | PNez e -> ignore (getFreeVars e)
    | PFun _ | NegPFun _ -> assert false
    | And pl | Or pl -> List.iter scanPred pl
    | Not p -> scanPred p
    | Imply (p1, p2) -> scanPred p1; scanPred p2
    | All _ | Exist _ -> (* do not scan under quantifiers *) ()
    | Pannot (_, p) -> scanPred p
  in

  (try 
    scanPred p;
    E.s (E.unimp "Cannot find a trigger for %a\n" L.d_pred p)
  with FoundCovering e -> 
    let trg = L.eq (L.Fun("have", [ e ])) L.trueExp in 
    if debugTrigger then 
      ignore (log "  found trigger %a\n" L.d_pred trg);
    trg)
      
(** Compile a predicate into a matching program ***)
module Compile = struct
  
  
      (* The compiler maintains some state *)
  type state = { 
              lastPc: int ref; (** A global last Pc used *)

              lastVar: int ref; (** a global last variable *)

              lastPatVar: int; (** The last pattern variable *)

      mutable usedExps: (exp, int) H.t; (** A map of expressions and the 
                                         * variables they were assigned to *)

      mutable unifPattVars: int list; (** a list of pattern variables that 
                                       * were already unified *)

      mutable progSoFar: matchBlock list;
      (* accumulate here the program in reverse order *)
    } 

        (** Initialize the state **)
  let initState (nrPattVars: int) : state = 
     { lastPc = ref (-1);
       lastVar = ref (nrPattVars - 1);
       lastPatVar = (nrPattVars - 1);

       unifPattVars = [];

       usedExps = H.create 13;
       progSoFar = [] 
     } 

        (** Prepare a state for a branch *)
  let branchState (s: state) : state = 
    { s with progSoFar = [];
             usedExps = H.copy s.usedExps }

  let addBlock (s: state) (p: matchBlock) : unit = 
    s.progSoFar <- p :: s.progSoFar

  let addInstr (s: state) (i: matchInstr) = 
    incr s.lastPc; 
    addBlock s (MInst (! (s.lastPc), i))

  let newVar (s: state) = 
    incr s.lastVar;
    ! (s.lastVar)

  let stateToMatchBlock (s: state) : matchBlock = 
    match List.rev s.progSoFar with 
      [ p ] -> p
    | pl -> MAnd pl

      (** Compile an expression and augment the state. Return the variable 
       * that will hold the node of the expression *)
  let rec compileExp (s: state) (e: exp) : var = 
    try H.find s.usedExps e
    with Not_found -> begin
      let v = compileExpNoMemo s e in 
      H.add s.usedExps e v;
      v
    end

  and compileExpNoMemo (s: state) (e: exp) : var = 
    (* convert into head and arguments *)
    let (head: int), (args: exp list) = splitExpHeadArgs e in 
    if args = [] then begin 
      (* A leaf *)
      let res = newVar s in 
      addInstr s (MatchLeaf (res, head));
      res
    end else begin
      (* Compile the arguments *)


      (* Process the arguments, bottom-up first. Make a first pass to handle 
       * the constructed arguments *)
      let args': var list = 
        let rec loopArgs = function
            [] -> []
          | (Var v) :: resta -> begin (* A pattern variable *)
              if v > s.lastPatVar then 
                E.s (bug "invalid pattern variable");
              v :: loopArgs resta
          end
          | arg :: resta -> (* Not a pattern variable. Compile it *)
              let v = compileExp s arg in
              v :: loopArgs resta
        in
        loopArgs args
      in
      (** Now all arguments are variables. They may be pattern variables or 
       * new variables *)

      (* Keep a list of instructions to add after this *)
      let postInstrs: matchInstr list ref = ref [] in


      
      (** Remember if one of the children is a constant, or an already 
       * unified pattern variable. We'll 
       * use it to restrict the search. *)
      let childo: var option ref = ref None in

      (** Remember the distinct pattern variables unified here **)
      let unif: var list ref = ref [] in
        

      (* Call this function when we have a child, a variable that is 
       * guaranteed to have a value by now *)
      let setChild (c: var) : var = 
        if !childo == None then begin
          childo := Some c;
          c
        end else begin 
          if separateArgumentChecks then begin
            let c' = newVar s in 
            postInstrs := MatchCheck(SetEq, c, c') :: !postInstrs;
            (* And it is being unified here *)
            unif := c' :: !unif;
            c'
          end else
            c
        end
      in
      (** Call this function when you see a unification variable that was 
       * just unified, in this same rule *)
      let alreadyUnif (v: var) : var = 
        if separateArgumentChecks then begin 
          let v' = newVar s in
          postInstrs := MatchCheck(SetEq, v, v') :: !postInstrs;
          (* and it is being unified here *)
          unif := v' :: !unif;
          v'
        end else 
          v
      in

      (* Now scan again to find which one we unify in this expression, and to 
       * find the child if not already found *)
      let rec loopArgs' = function
          [] -> []
        | v :: resta -> begin 
            let v' = 
              if v > s.lastPatVar then 
                setChild v (* Guaranteed to have a value *)
              else begin 
                (* See if it was unified already *)
                if (List.mem v s.unifPattVars) then begin
                  (* It is already unified. If the unification was not done 
                   * here, then mark it as a child *)
                  if not (List.mem v !unif) then 
                    setChild v
                  else
                    alreadyUnif v
                end else begin (* It was not unified. Unify it here *)
                  unif := v :: !unif;
                  H.add s.usedExps (Var v) v;
                  s.unifPattVars <- v :: s.unifPattVars;
                  v
                end
              end
            in
            v' :: loopArgs' resta
        end
      in
      let args'' = loopArgs' args' in 
      let res = newVar s in 
      addInstr s (MatchConstr (res, head, args'', !childo, !unif));
      let rec loopPost = function
          [] -> ()
        | i :: resti -> addInstr s i; loopPost resti
      in
      loopPost !postInstrs;
      res
    end
          

      (** Compile a predicate and augment the state *)
  let rec compileAppendPred (s: state) (p: pred) : unit = 
    match p with 
      Bool true -> ()
    | Bool false -> 
        addBlock s (MOr [])

    | And pl -> (* Use the same state *)
        let rec iter f = function [] -> () | h::t -> f h; iter f t in
        iter (compileAppendPred s) pl

    | Or pl -> 
        (* Compute the new unified variables, as the intersection of all 
         * variables *)
        let intersUnif: var list option ref = ref None in 
        let intersOne (s': state) = 
          match !intersUnif with 
            None -> intersUnif := Some s'.unifPattVars
          | Some old -> 
              intersUnif := 
                Some (List.filter (fun v -> List.mem v old) s'.unifPattVars)
        in
        let blocks: matchBlock list = 
          let rec map f = function [] -> [] 
                         | h::t-> let h'=f h in h' :: map f t in 
          map 
            (fun p -> 
              let s' = branchState s in (* Do them all in the same state *)
              compileAppendPred s' p;
              let res = stateToMatchBlock s' in 
              intersOne s';
              res
            )
            pl
        in
        addBlock s (MOr blocks);
        s.unifPattVars <- (match !intersUnif with 
                              None -> s.unifPattVars
                            | Some inters -> inters) 

    | PNez(BinOp(SetEq, Fun("have", [e]), et)) 
    | PNez(BinOp(SetEq, et, Fun("have", [e]))) when et = trueExp -> 
        let v: var = compileExp s e in
        ()
          
          (** Equalities and disequalities *)
    | PNez(BinOp(((SetEq | SetNe) as bop), e1, e2)) -> 
        let v1: var = compileExp s e1 in 
        let v2: var = compileExp s e2 in 
        addInstr s (MatchCheck (bop, v1, v2))
        

    | p -> E.s (unimp "Cannot compile predicate %a\n" d_pred p)


  let matchId = ref 0 
  let compileForall (p: pred) (prf_all: P.proof) : matchProg = 

    (* Compute the deBruijn depth of the predication *)
    let rec getDepth = function
        All (_, p) -> 1 + getDepth p
      | _ -> 0
    in
    let depth = getDepth p in 

    if depth = 0 then 
      E.s (bug "compileForall: not a forall");

    (* Count how many pattern variables there are and get the body *)
    let currPattVar = ref depth in 
    let patNames = IH.create 13 in 

    (* Go get the body *)
    let rec getBody = function
        All (n, p') -> 
          decr currPattVar;
          IH.add patNames !currPattVar n;
          getBody p'

      | Pannot(ATrigger left, right) -> 
          assert (!currPattVar = 0);

          (* Compile the left into a matching block *)
          let s = initState depth in 
          compileAppendPred s left;
          incr matchId;

          let prg = 
            { mId = !matchId;
              mBlock = stateToMatchBlock s;
              mNrPattVars = depth;
              mBody = right;
              mForallPred = p;
              mForallProof = prf_all;
              mPatVarName = patNames } in
          currentProg := prg;

          (* Check that we have instantiated all pattern variables *)
          for v = 0 to depth - 1 do 
            if not (List.mem v s.unifPattVars) then begin
              currentProg := prg;
              ignore (log "Compiled %a into\n " d_pred p);
              printCurrentProg ();
              E.s (error "Trigger for %a does not unify variable %a\n"
                     d_pred p d_var v);
            end;
          done;
          prg
          
            
      | p -> 
          let trig = inferTrigger depth p in 
          getBody (Pannot(ATrigger trig, p))
    in
    let prog = getBody p in
    currentProg := prog;
    if !doProofTrace then
      ignore (E.logg "%srule %d: %a\n"
                     (proofTraceInd ())
                     prog.mId
                     d_pred p);
    if !debug then begin
      ignore (log "Compiled %a into\n " d_pred p);
      printCurrentProg ();
    end;
    prog

end

module IM = Intmap

type subst = node IM.t
      
let d_subst () (s: subst) = 
  text "{" ++ align ++
    (docList ~sep:(chr ',' ++ break) 
             (fun (k, v) -> dprintf "%a=%a" d_var k d_exp v.exp) ())
    (IM.fold (fun k v acc -> (k, v) :: acc) s []) ++ unalign ++ text "}"

let valNode (s: subst) (v: var) : node = 
  try (IM.find v s).root
  with Not_found -> 
    E.s (E.bug "MA: cannot find value of %a" d_var v)
      

let applySubstToPred (s: subst) (p : pred) = 
  substPred
    (fun depth e -> 
      match e with 
        Var v when v >= depth -> 
          Some ((valNode s (v - depth)).exp)
      | _ -> None) 
    p

    

    (** Make a proof for the body, having a proof for the predicate that 
     * includes it *)
let rec makeProofBody 
    (s: subst) 
    (depth: int) 
    (p: L.pred) (prf: P.proof) : P.proof = 
  assert (!genProofs);
  match p with 
    All(v, p') -> 
      let inst_v = valNode s depth in 
      makeProofBody s (depth - 1) p'
        (P.Alle(inst_v.exp, prf))

  | Pannot (ATrigger left, right) -> 
      (* left has no logical meaning, but we still use a proof constructor, 
       * to make it easier to check proofs *)
      P.AnnotE prf

  | _ -> prf (* We created the trigger ourselves *)
        


let varUserName (v: var) : string =
  (try IH.find !currentProg.mPatVarName v
   with Not_found -> "@@not found@@")

let d_substNamedVars () (s: subst) =
  (docList ~sep:(chr ',' ++ break)
           (fun (k, v) -> dprintf "%s=%a" (varUserName k) d_exp v.exp) ())
  (IM.fold (fun k v acc ->
              (if (k < !currentProg.mNrPattVars) then
                ((k, v) :: acc)
              else
                acc)) s [])


module Interpret = struct

  let callid = ref 0 
  let newCallbackId () = incr callid; !callid

  exception InvalidNode


  (** We keep here the bodies to be asserted *)
  let delayedBodies: (pred * P.proof) list ref = ref []

  (** This is the main matching function. It takes a current substitution, 
   * and a list of lists of blocks to interpret *)
  let rec doMatch (prg: matchProg) 
                  (s: subst) 
                  (pl: matchBlock list list) : unit = 
    match pl with 
      [] -> (* We are done *)
        if !debug then 
          ignore (log "Done matching rule %d with %a\n" 
                    prg.mId d_subst s);
        (* Now substitute the body *)
        let body' = applySubstToPred s prg.mBody in
        if !debug then 
          ignore (log "  asserting body %a\n" d_pred body');
        if !doProofTrace then
          ignore (E.logg "%s%a: instantiated rule %d with %a\n"
                         (proofTraceInd ())
                         d_pred body'
                         prg.mId
                         d_substNamedVars s);

        (* make the proof of this body *)
        let bodyprf = 
          if !genProofs then begin
            let allprf = prg.mForallProof in
            makeProofBody s 
              (prg.mNrPattVars - 1)
              prg.mForallPred 
              allprf
          end else
            noProof
        in
        delayedBodies := (body', bodyprf) :: !delayedBodies;
        (* We will reset delayedBodies on undo *)
        ()

    | [] :: rest -> doMatch prg s rest

    | (MAnd bl :: rest1) :: rest -> 
        doMatch prg s (bl :: rest1 :: rest)

    | (MOr bl :: rest1) :: rest -> 
        (* Will just match them all *)
        let idx = ref (-1) in 
        let rec iter f = function [] -> () | h::t -> f h; iter f t in
        iter (fun b -> 
          incr idx;
          doMatch prg s ((b :: rest1) :: rest)) bl

    | (MInst (pc, i) :: rest1) :: rest -> 
        let rest' = rest1 :: rest in 
        (* Now interpret the instruction *)
        doInstr prg s pc i (rest1 :: rest)

  and doInstr (prg: matchProg) 
              (s: subst)
              (pc: int) (i: matchInstr) (rest: matchBlock list list) : unit = 
    (** Get all the nodes with a given head in an equivalence class *)
    let filterEqcHead (n: node) (h: int) : node list = 
      foldEqc (fun acc n' -> if n'.head == h then n' :: acc else acc) [] n 
    in

    currentProg := prg;
    if !debug then 
      ignore (log "running %d:%d: %a with %a\n" 
                prg.mId pc d_instr i d_subst s);
    let continue (s': subst) = 
      doMatch prg s' rest
    in
    match i with  
      MatchCheck (bop, v1, v2) -> begin
        let n1 = valNode s v1 in
        let n2 = valNode s v2 in 
        match bop with 
          (SetEq | SetNe) -> 
            (* Remember if we have done this. We need to do this to prevent a 
             * callback from firing several times. *)
            let donethis = ref false in 
            let setDoneThis () = 
              donethis := true;
              registerUndoRun (fun _ -> donethis := false)
            in
            let tryIt (cont: unit -> unit) = 
              if n1.root == n2.root then begin 
                if !debug then 
                  ignore (log "  nodes are known to be equal\n");
                setDoneThis ();
                if bop = SetEq then continue s else ()
              end else begin 
                (* Perhaps they are known to be disequal *)
                if forbiddenEquality n1 n2 != None then begin  
                  if !debug then 
                    ignore (log "  nodes are known to be not equal\n");
                  setDoneThis ();
                  if bop = SetEq then () else continue s
                end else (* Undetermined *)
                  cont ()
              end
            in
            (* Make an initial try *)
            tryIt
              (fun _ ->  
                (* We do not know either way. Register a callback for 
                 * merge *)
                let callid = newCallbackId () in
                if !debug then 
                  ignore (log "  check undetermined. Registering callback with %a (cid=%d)\n" d_nodeTop n1.root callid);
                if bop = SetEq then begin
                  let mergeCallback () = 
                    if not !donethis then begin 
                      if !debug then 
                        ignore (log "called back %d:%d (cid=%d) because of merge on %a\n" 
                                  prg.mId pc callid d_nodeTop n1.root);
                      tryIt (fun _ -> ())
                    end
                  in
                  addMergeCallback n1.root mergeCallback
                end else begin 
                  let forbidCallback () = 
                    if not !donethis then begin 
                      if !debug then 
                        ignore (log "called back %d:%d (cid=%d) because of forbid on %a\n" 
                                  prg.mId pc callid d_nodeTop n1.root);
                      tryIt (fun _ -> ())
                    end
                  in
                  addForbidCallback n1.root forbidCallback
                end
              )

        | _ -> E.s (unimp "check %s" (binop_to_string bop))
      end
    | MatchLeaf (v, l) -> begin 
        (* See if we have the leaf in the EDAG already *)
        let donethis = ref false in 
        let setDoneThis () = 
          donethis := true;
          registerUndoRun (fun _ -> donethis := false)
        in
        try 
          let n = IH.find headToNode l in 
          if n.root.active then begin 
            setDoneThis ();
            continue (IM.add v n.root s)
          end else
            raise Not_found
        with Not_found -> begin
          (* Register a callback for when the node is created and stop *)
          let callid = newCallbackId () in
          if !debug then 
            ignore (log "    wait for leaf %s to be activated (cid=%d)\n"
                      (headToString l) callid);
          let activateLeafCallback (n: node) = 
            (* If we just activated it and has a member with the right 
             * head, then it's ours *)
            if not !donethis && n.active then begin 
              let goodhead = filterEqcHead n l in 
              if goodhead != [] then begin 
                if !debug then 
                  ignore (log "called back %d:%d (cid=%d) because leaf %s was activated\n"
                            prg.mId pc callid (headToString l));
                setDoneThis ();
                continue (IM.add v n.root s)
              end
            end
          in
          addActivateCallback activateLeafCallback
        end
    end
    | MatchConstr (v, h, args, parentof, unif) -> begin
        (* Set up a memo table, recording all the nodes we have already used 
         * for the "unif" variables *)
        let memo: node list list ref = ref [] in
        let d_memoElem () (nl: node list) = 
          text "{" ++ 
            (docList (fun n -> 
              dprintf "%a (root=%a)"
                d_nodeTop n 
                d_nodeTop n.root) () nl) ++  text "}"
        in
        let checkMemo (s': subst) = 
          (* Our versions of some list functions to avoid confusing the 
           * profiler *)
          let rec map f = function [] -> [] | h::t->let h' = f h in h' :: map f t in 
          let unif' = map (valNode s') unif in 
          let rec exists f = function [] -> false | h::t-> f h || exists f t in
          let rec compare2 l1 l2 = 
            match l1, l2 with 
              [], [] -> true
            | n1::t1, n2::t2 -> n1.root==n2.root && compare2 t1 t2
            | _ -> assert false
          in
          if exists 
               (fun memo' -> 
                 assert (List.length memo' = List.length unif');
                 compare2 memo' unif')
              !memo then begin
(*
                ignore (log "  duplicate this=%a\n     memo=%a\n"
                          d_memoElem unif'
                          (docList (d_memoElem ())) !memo);
*)
                false
              end
          else begin 
            pushWithUndo memo unif';
            true
          end
        in
        
        (** this is the function that gets fired for each candidate node with 
         * the right head *)
        let tryCandidate (n: node) = 
          currentProg := prg;
          assert (n.head == h && n.root.active);
          (* Look at the arguments *)
          let rec unifArgs (s: subst) 
                           (nargs: node list) (args: var list): subst = 
            match nargs, args with 
              [], [] -> s
            | (narg :: nargs), (arg :: args) -> begin
                (* See if the argument is already unified *)
                try 
                  let argn = IM.find arg s in
                  if argn.root == narg.root then 
                    unifArgs s nargs args
                  else (* Drop this node *)
                    raise InvalidNode
                with Not_found -> 
                  unifArgs (IM.add arg narg.root s) nargs args
            end
            | _ -> E.s (bug "invalid number of arguments")
          in
          try 
(*
            if List.length n.args <> List.length args then 
              ignore (warn "Node %a has arguments %a, which does not match %a\n"
                        d_nodeTop n
                        (docList (d_nodeTop ())) n.args
                        (docList (d_var ())) args);
*)
            let s' = unifArgs s n.args args in 
            if checkMemo s' then
              continue (IM.add v n.root s')
            else begin 
              if !debug then 
                ignore (log "   duplicate\n");
              () (* Duplicate, drop it *)
            end
          with InvalidNode -> begin
            if !debug then 
              ignore (log "   does not match\n");
            ()
          end
        in
            
            
        (* See if we need to look into somebody's parents *)
        match parentof with 
          None -> (* No, all nodes with the given head *)
            (** Apply it to the current parent *)
            let all_candidates: node list = IH.find_all headToNode h in
            List.iter tryCandidate all_candidates;
            (* Also register this for later *)
            let callid = newCallbackId () in
            if !debug then 
              ignore (log "    wait for node with head %s to be activated (cid=%d)\n"
                        (headToString h) callid);
            let activateHeadCallback (n: node) = 
              (* If we just activated it and has a member with the right 
               * head, then it's ours *)
              if n.root.active then begin 
                let good_head = filterEqcHead n h in 
                let rec iter f = function [] -> () | h::t -> f h; iter f t in 
                iter
                  (fun n' -> 
                    if !debug then 
                      ignore (log "called back %d:%d (cid=%d) because node %a was activated\n"
                                prg.mId pc callid d_nodeTop n');
                    tryCandidate n')
                  good_head
              end
            in
            addActivateCallback activateHeadCallback 

        | Some p -> 
            let pn = valNode s p in
            assert (pn.root == pn);
            assert (pn.active);
            (* Get the candidates from the parents of pn *)
            let all_candidates: node list = 
              foldCirclist
                (fun acc n' -> 
                  if n'.root.active && n'.head == h then 
                    n' :: acc
                  else
                    acc)
                []
                pn.parents 
            in
            let rec iter f = function [] -> () | h::t -> f h; iter f t in 
            iter 
              (fun n -> 
                if !debug then  
                  ignore (log "    try existing candidate %a\n" d_nodeTop n);
                tryCandidate n) all_candidates;

            let callid = newCallbackId () in
            if !debug then 
              ignore (log "    wait for parent of %a with head %s to be activated (cid=%d)\n"
                        d_nodeTop pn (headToString h) callid);
              
            (* Also register this for later *)
            let newParentsCallback (n: node) = 
              if n.root.active && n.head == h then begin 
                if !debug then 
                  ignore (log "called back %d:%d (cid=%d) because node %a was activated as a parent of %a\n"
                            prg.mId pc callid d_nodeTop n 
                            d_nodeTop pn);
                tryCandidate n
              end
            in
            addNewParentsCallback pn newParentsCallback
    end

  let run (prg: matchProg) = 
    doMatch prg IM.empty [ [ prg.mBlock ] ]
    
end

let internForall (p: pred) (prf: P.proof) : unit = 
  let prog = Compile.compileForall p prf in 
  (* start interpreting the program *)
  Interpret.run prog;
  ()

let witness (w: L.pred list) = w

let init () = 
  Interpret.delayedBodies := []

let snapshot () = 
  if !Interpret.delayedBodies != [] then assert false;
  registerUndoRun (fun _ -> Interpret.delayedBodies := [])

(* take note of a fact *)
let assume (msg: message) = ()


let collectMatches () : unit = 
  (* See if we have some bodies to assert *)
  let l = !Interpret.delayedBodies in 
  if l != [] then begin 
    Interpret.delayedBodies := [];
    let rec iter f = function [] -> () | h::t -> f h; iter f t in 
    if !matchingDisabled == None then 
      iter
       (fun (body, bodyprf) -> 
         newMessage
           { body = Extern body; (* We let the engine internalize it *)
             owner = matchid;
             proof = bodyprf;
             fromgoal = false;
           })
      l
  end

