(* ----------------------------------------------------------------------- *)
(* Header Definitions *)

{
module E = Errormsg
module H = Hashtbl

open Prolog_parse
open Logic

let lexerror msg = E.parse_error msg



(* Defines a lookup table of keywords *)
let keyword_table = Hashtbl.create 53
let _ = (* Mips Keywords *)
  List.iter (fun (kwd, tok) -> Hashtbl.add keyword_table kwd tok)
            [ "extern", EXTERN;
              "commit", COMMIT;
              "is", IS; 
              "if", IF;
              "endif", ENDIF;
              "true", TRUE;
              "false", FALSE;
              "then", THEN;
              "else", ELSE;
              "terminal", TERMINAL;
              "case", CASE; 
              "endcase", ENDCASE;
              "of",   OF; 
              
              "Type", TYPE;
              "Prop", PROP;
              "Constant", CONSTANT;
              "Tabled", TABLED;
              ]


let lexInt (s: string) : token =
  try
    INT (int_of_string s)
  with Failure _ -> 
    INT 0


}
     
(* ----------------------------------------------------------------------- *)
(* Identifier Definitions *)

let startvar = ['A'-'Z' '_']
let startexvar = '%'
let startatom = ['a'-'z' '$' '+' '-' '*' '=' '>' '<' '&' '@']

let hexsign = "0x"|"0X"
let hex = ['0'-'9' 'a'-'f' 'A'-'F']
let fsign = ('+'|'-')


let digit = ['0'-'9']

let whitespace = [' ' '\009' '\012' '\r']

(* Characters that can be in identifiers, except in the first position *)
(* RS: added prime *)
let tail = startvar | startatom | digit | '\''

let labelTail = startvar | startatom | digit | '.'
    
(* ----------------------------------------------------------------------- *)
(* Lexical Analysis Rules *)
    
rule token = parse
    
    (* ## white space, comment, etc ## *)
  | whitespace	       	{ token lexbuf } (* White spaces *)
  | '\n'		{  (* Newlines *)
                          E.newline (); token lexbuf
			  }
  | '\\' 'n'            { E.newline (); token lexbuf }
  | '#'                 { comment lexbuf } (* handle comments *)

 
    (* ## immediates ## - //todo: need to handle float *)
  | '-'?(digit)+        { lexInt  (Lexing.lexeme lexbuf) }
  | '-'?(hexsign)(hex)+ { lexInt (Lexing.lexeme lexbuf) }
  | '\039'[^'\039']'\039' { INT( let str = Lexing.lexeme lexbuf in
                                 (Char.code str.[1])) }

  | "/\\"                { AND }
  | "\\/"                { OR }
  | "->"                 { ARROW }
  | "EX"                 { EXISTS }
  | "ALL"                { FORALL }
  | "="                  { PBINOP SetEq }
  | "<>"                 { PBINOP SetNe }
  | ">"                  { PBINOP SetGt }
  | ">="                 { PBINOP SetGe }
  | "<"                  { PBINOP SetLt }
  | "<="                 { PBINOP SetLe }
  | ">u"                 { PBINOP SetAb }
  | ">=u"                { PBINOP SetAe }
  | "<u"                 { PBINOP SetBl }
  | "<=u"                { PBINOP SetBe }

  | "add"                { BINOP Add }
  | "sub"                { BINOP Sub }
  | "mult"               { BINOP Mult }
  | "band"               { BINOP BAnd }
  | "bor"                { BINOP BOr }
  | "bxor"               { BINOP BXor }
  | "+"                  { BINOP Plus }
  | "-"                  { BINOP Minus }
  | "*"                  { BINOP Times }
  | "srl"                { BINOP Srl }
  | "sra"                { BINOP Sra }
  | "sll"                { BINOP Sll }
  | "setpo"              { BINOP SetPo }
  | "setpe"              { BINOP SetPe }
  | "setbe"              { BINOP SetBe }
  | "setbl"              { BINOP SetBl }
  | "setae"              { BINOP SetAe }
  | "setab"              { BINOP SetAb }
  | "setle"              { BINOP SetLe }
  | "setlt"              { BINOP SetLt }
  | "setge"              { BINOP SetGe }
  | "setgt"              { BINOP SetGt }
  | "setne"              { BINOP SetNe }
  | "seteq"              { BINOP SetEq }

  | '&'(labelTail)+     {
                           let str = Lexing.lexeme lexbuf in
                           LABEL (String.sub str 1 (String.length str - 1), E.getLocation ())
                        }

  | (startatom)(tail)*  {
                           let str = Lexing.lexeme lexbuf in
                           try
                             H.find keyword_table str 
                           with Not_found -> ATOM (str, E.getLocation ())
                        } 

  | (startvar)(tail)*    
                        { 
                           let str = Lexing.lexeme lexbuf in
                           try H.find keyword_table str 
                           with Not_found -> VAR str
                        } 

  | (startexvar)(tail)*    
                        { 
                           let str = Lexing.lexeme lexbuf in
                           EXVAR (String.sub str 1 (String.length str - 1))
                        } 
                        
    (* ## control and misc ## *)
  | ','                 { COMMA } (* Commas *)
  | ":-"                { COLON_MINUS }
  | "->"                { ARROW }
  | ':'                 { COLON } (* Colon *)
  | ';'                 { SEMICOLON }
  | '|'                 { BAR }
  | '('                 { LPAREN (E.getLocation ()) } (* Left Paren *)
  | ')'                 { RPAREN } (* Right Paren *)
  | '['                 { LBRACKET } (* Left Bracket *)
  | ']'                 { RBRACKET } (* Right Bracket *)
  | '{'                 { LCURLY } (* Left curly bracket *)
  | '}'                 { RCURLY } (* Right curly bracket *)
  | '.'                 { DOT }
  | '!'                 { BANG } 
  | eof			{ EOF } (* End of File *)
  | _			{ 
                          begin
                            lexerror ("Illegal Character '"^ 
                                      (Lexing.lexeme lexbuf)^"'");
			    token lexbuf
			  end }


      (* Comments and Annotations *)
and comment = parse
    '\n'	{ E.newline (); token lexbuf  (* out of comment *) }

  | eof { EOF }
  | _           { comment lexbuf } (* eat comment *)


(* ----------------------------------------------------------------------- *)
(* Trailer Section *)
{  

(* This module must come after parser, which must come after Kettle. So, we 
 * set a pointer in Kettle to point to the parsing routine *)
let _ = 
  Prolog.thePrologParser := 
     (fun lexbuf -> Prolog_parse.main token lexbuf)

}
