open Pretty
module E = Errormsg
module H = Hashtbl
module U = Util
module IH = Inthash 

module Bitmap : 
    sig
      type t
      val make: int -> t
      val size: t -> int
      val setTo: t -> int -> bool -> unit
      val testAndSetTo: t -> int -> bool -> bool
      val test: t -> int -> bool
      val union_except: t -> t -> t -> bool
      val iter: (int -> unit) -> t -> unit
    end = Bitvector

open Logic

let checkProofs = ref false

let printMagicProofWarning = ref true

let printer = ref defaultLogicPrinter

let d_exp () e = printExp !printer e
let d_pred () p = printPred !printer p

(* In order to maintain a compact representation of proofs we represent a 
 * proof node as the name of the axiom along with a list of arguments. An 
 * exception is made for the hypothetical and parametric proof nodes. *)
type parg = 
    I of int                            (* An integer argument *)
  | S of string                         (* A string argument *)
  | E of exp                          (* An expression argument *)
  | P of proof                          (* A subproof *)
  | F of pred                         (* A predicate argument *)

and relset = 
    { relfirst: int;                      (** The hypothesis that has index 0 
                                           * in the set *)
      mutable relcomputed: bool;          (** Whether this set is filled *)
      relbm: Bitmap.t
    } 

and proof = 
    Hyp of Util.symbol                    (* A hypothesis, 
                                           introduced with Impi, or Contra *) 

  | FreeHyps of relset * proof            (** A proof with memoized set of 
                                           * free hypotheses. The first 
                                           * argument says if we have 
                                           * computed the set already. *)
  | Truei

  | Falsei of proof * proof              (** A proof of false, given a proof 
                                          * of a predicate and that of its 
                                          * negation *)

  | Andi of proof list
    (** Proves And[p1; p2; ... ] *)

  | Ande of int * proof
	(** Ande (i,pf) is a proof of p when
	   pf is a proof of And ps, where List.nth ps i = p *)

  | Ori of Logic.pred list * int * proof

  | Ore of Logic.pred * Logic.pred list * proof * (Util.symbol * proof) list
        (** [Ore(P, PDISJ, D, DL)] is a proof of P, if D is a proof of a 
         * disjunction PDISJ, and DL is a list with hypothetical proofs of P, 
         * one for each disjunct *)        


  | Alli of Logic.var * proof
  | Alle of  Logic.exp * proof

  | ExistE of proof * Logic.var * Util.symbol * proof
        (** [Existe(d1,v,h,d2)] is a proof of q, if d is a proof of 
         * Exist(x,p), v is a fresh variable, h is the fresh hypothesis name 
         * for P[v/x], and d2 is a proof of q, with hypothesis h. *)

  | Impi of Util.symbol * Logic.pred * proof (* bind var to a hypothetical 
                                            * proof of the pred. The last 
                                            * argument is a proof of the 
                                            * right hand side, possibly 
                                            * refering to var *)

  | Impe of proof * proof

  | AnnotE of proof
    (** AnnotE(d) is a proof of Q if d is a proof of Pannot(a,Q) *)

  | AnnotI of Logic.pannot * proof
    (** AnnotI(a,d) is a proof of Pannot(a, p) when d is a proof of p *)


  | Cut of proof * Util.symbol * proof
                                        (** [Cut(prf1, x, prf2)], when 
                                         * [prf1:P1], [prf2: [x:P1] P2], is a 
                                         * proof of P2 **)

                                        (* A proof of the given predicate, 
                                         * using a proof of false, based 
                                         * onthe named hypothesis that 
                                         * "negatePred" holds *)
  | Contra of Logic.pred * Util.symbol * proof
    
  | G of string * parg list             (* This is a generic proof 
                                         * constructor *)

  | Refl of Logic.exp 
  | Symm of proof
    (** Symm p proves e1=e2 if p:e2=e1 *)

  | Trans of Logic.exp * proof * proof
    (** Trans(et, p1t, p2t) proves e1=e2 if p1t:e1=et, pt2:et=e2 *)

  | Cong of Logic.exp * Logic.exp * proof list
  (** Cong \[pf\] proves Unop e1 = Unop e2 if pf proves e1 = e2;
      Cong \[pf1; pf2\] proves Binop (a1,b1) = Binop (a2,b2) if
        pf1 proves a1=a2 and pf2 proves b1=b2;
      Cong pfs proves Fun(f,as) = Fun(f,bs) if
        pfs.(n) proves as.(n) = bs.(n) 
     
      Note that each subproof of equality is really a proof 
      of (Var 0)=(Var 1) in the context in which (Var 0) and (Var 1)
      are replaced with the two things in their appropriate contexts.
   *)

  | PCong of proof * proof list
  (** If p |- PNez(e) and q |- e = e', then PCong (p, [q]) |- PNez(e') *)
  (** If p |- PFun(s,[e1,...,en]) and q1 |- e1 = e1', ..., qn |- en = en', then PCong (p, [q1,...qn]) |- PFun(s,[e1',...,en']) *)

  | WithLemma of Util.symbol (* The name of the lemma *)
                * Logic.exp list (* The expressions to instantiate the universally 
                            * quantified variables *)
                * proof list (* The proofs of the hypotheses *)

  | UnrollPDef of Logic.pred * proof (* Prove a predicate of the form 
                                * "PFun(n,args)" by replacing it with its 
                                * definition and providing a proof for the 
                                * resulting expansion *)

  | UnrollEDef of Logic.pred * string * proof
      (* Prove a predicate that uses an abbreviation expression constructor.
       * The given predicate should be the goal without the substitutions made.
       *)
  | RollEDef of Logic.pred * string * proof
      (* Like UnrollEDef, but proves the substituted version of the predicate *)

  | ArithEq of Logic.exp * Logic.exp * int * parg list
        (** [ArithEq(e,e',n,rest)] proves [e=e']. [n] is a factor >= 1. [rest] 
         * is a weighted list of proofs of equalities. Each factor is 
         * followed by its proof. Let fi be the ith factor and ei = ei' be 
         * the ith equality. Check that n*(e - e') - fi*(ei - ei') reduces 
         * using the rules of arithmetic to 0 *)

  | ArithNe of Logic.exp * Logic.exp * int * parg list
        (** [ArithNe(e,e',n,rest)] proves [e<>e']. [n] is a factor >= 1. [rest] 
         * is a weighted list of proofs of equalities. Each factor is 
         * followed by its proof. Let fi be the ith factor and ei = ei' be 
         * the ith equality. Check that n*(e - e') - fi*(ei - ei') reduces 
         * using the rules of arithmetic to something other than 0 *)

  | ArithGeq of Logic.exp * Logic.exp * int * parg list
        (** [ArithGeq(e,e',n,rest)] proves [e>=e']. [n] is a factor >= 1. The 
         * last argument is a weighted list (with positive factors) of proofs 
         * of >= inequalities. Each factor is followed by its proof. Let fi 
         * be the ith factor and ei >= ei' be the ith inequality. Check that 
         * n*(e - e') - fi*(ei - ei') reduces using the rules of arithmetic to a 
         * constant that is > - n *)

  | ArithContra of parg list
        (** Is a proof of False, when parg is a weighted list of proofs of >= 
         * inequalities (with positive factors), or equalities. Each factor 
         * is followed by its proof. Let fi be the ith factor and ei >= ei' 
         * be the ith equality or inequality. Let n be the sum of fi*(ei - 
         * ei'). It must be the case that n < 0 *)


  | ArithNeq of int * Logic.exp * Logic.exp   
    (** Check that e <> e' by verifiying that e - e' mod32 = i (i <> 0) using 
     * the rules of algebra *)

  | ArithApart of int * Logic.exp * Logic.exp 
     (** ArithApart(r, e1, e2) means that e1 - e2 = r and (-4 >= r or r >= 
      * 4).*)

  | ArithGt of int * Logic.exp * Logic.exp  
        (** Check that e >= e', by verifying that e - e' = i and i > 0. !!! 
         * This is unsound in presence of overflow. We'll have to clean this 
         * up. *)

  | ArithLt of proof
	(** If p |- e >= e', then ArithLt(p) |- e' < e *)
  | ArithLeq of proof
	(** If p |- e > e', then ArithLt(p) |- e' <= e *)

  | SelUpd0 of Logic.exp * Logic.exp * proof
        (* (SelUpd0 V M D) proves that (= (sel A (upd A' V M)) V) provided 
         * that D proves A = A' *)

  | SelUpd1 of Logic.exp * Logic.exp * proof
        (* (SelUpd1 V M D) proves that (= (sel A (upd A' V M)) (sel A M)) 
           provided that D proves A <> A' *)

  | SelUpd of Logic.exp * Logic.exp * Logic.exp * Logic.exp
        (* [SelUpd(SA,UA,UV,UM)] proves ((SA = UA) AND (sel SA (upd UA UV UM)) 
         * = UV) OR ((SA <> UA) AND (sel SA (upd UA UV UM)) = (sel SA UM)) *)

  | Magic of string * Logic.pred 
    (** These are proofs that are allowed to go through the proof checker, 
     * with a warning. Pass a string to identify this magic *)

  | MagicLemma of lemma
    (** A magic proof of a lemma *)

  | CoqLemma of string list
    (** Proof by Coq implementation, with variable types *)

  | Unimplemented of string     (* We'll expand the proof language as needed 
                                 * in the future. Use Unimplemented until 
                                 * then *)



and lemma = 
    { 
      lname: U.symbol; (* The symbol corresponding to the name of the lemma *)
      lvars: string list; (* A list of variables that are universally 
                           * quantified over. These are bound in the rest of 
                           * the lemma, using UVar *)
      lhyps: Logic.pred list; (* A list of hypotheses. *)
      lconc: Logic.pred; (* A conclusion *)
    } 


(**************** Lemmas ************)

(* We keep a global registry of lemmas *)
let allLemmas: lemma IH.t = IH.create 17

(* We keep an optional open Coq process for checking implementations *)
let coqProc = ref None
type coq_mode =
    CoqTerse
  | CoqVerbose
let coqFlag = ref CoqVerbose

let openCoq dirs modl flag =
  match !coqProc with
    None ->
      coqFlag := flag;
      let cmdline = String.concat "" (List.concat (["coqtop"] :: List.map (fun d -> [" -I "; d]) dirs @ [[" 2>/dev/null"]])) in
      let inch, outch = Unix.open_process cmdline in

      ignore (input_line inch);
      output_string outch ("Require Import " ^ modl ^ ".\nImport " ^ modl ^ "_mod.\n");
      flush outch;

      coqProc := Some (inch, outch)
  | _ -> E.s (E.error "Second call to Proof.openCoq. You may not switch Coq instances during verification.")

let registeredLemmas () : lemma list = 
  IH.fold (fun _ l acc -> l::acc) allLemmas []

let findLemma (l: lemma) : unit = 
  try 
    if l <> IH.find allLemmas l.lname then
      E.s (E.error "Using lemma %s that is different from the registered one"
             (U.symbolName l.lname))
  with Not_found -> 
    E.s (E.error "Using unknown lemma %s\n" (U.symbolName l.lname))

(* Printing for Coq *)
(* Print the Coq name for a binop *)
let printBinopCoq bop =
  text (match bop with
    Plus -> "plus"
  | Minus -> "minus"
  | Times -> "times"
  | Add -> "add"
  | Sub -> "sub"
  | Mult -> "mult"
  | Div -> "div"
  | BAnd -> "band"
  | BOr -> "bor"
  | BXor -> "bxor"
  | Sll -> "sll"
  | Srl -> "srl"
  | Sra -> "sra"
  | SetEq -> "seteq"
  | SetNe -> "setne"
  | SetGt -> "setgt"
  | SetGe -> "setge"
  | SetLt -> "setlt"
  | SetLe -> "setle"
  | SetAb -> "setab"
  | SetAe -> "setae"
  | SetBl -> "setbl"
  | SetBe -> "setbe"
  | SetPe -> "setpe"
  | SetPo -> "setpo"
  | SignExt -> "signext")

let rec printExpCoq vars e =
  match e with
    UVar n -> text "V" ++ num n
  | Var i -> List.nth vars i
  | Fun (f, []) -> text f
  | Fun ("&", [e]) -> text "(addrOf " ++ printExpCoq vars e ++ text ")"
  | Fun (f, es) ->
      let es = List.fold_left (fun es e -> es ++ text " " ++ printExpCoq vars e)
	  (text "(" ++ text f) es in
      es ++ text ")"
  | BinOp (bop, e1, e2) ->
      text "(" ++ printBinopCoq bop ++ text " " ++ printExpCoq vars e1
	++ text " " ++ printExpCoq vars e2 ++ text ")"
  | Int n ->
      if Int32.compare n Int32.zero < 0 then
	text "(" ++ num (Int32.to_int n) ++ text ")"
      else
	num (Int32.to_int n)
  | _ -> E.s (E.error "Unknown Coq expression type in Proof")

let rec printPredCoq vars p =
  match p with
    PFun (f, es) ->
      let es = List.fold_left (fun es e -> es ++ text " " ++ printExpCoq vars e)
	  (text "("
	     ++ text (String.capitalize f))
	  es in
      es ++ text ")"
  | PNez (BinOp (SetEq, e, Int n)) when n = Int32.zero ->
      printPredCoq vars (PNez (negateExp e))
  | PNez (BinOp (SetEq, e1, e2)) ->
      text "(" ++ printExpCoq vars e1 ++ text " = " ++ printExpCoq vars e2 ++ text ")"
  | PNez (BinOp (SetNe, e1, e2)) ->
      text "~(" ++ printExpCoq vars e1 ++ text " = " ++ printExpCoq vars e2 ++ text ")"
  | PNez (BinOp (SetGe, e1, e2)) ->
      text "(" ++ printExpCoq vars e1 ++ text " >= " ++ printExpCoq vars e2 ++ text ")"
  | PNez e ->
      text "~(" ++ printExpCoq vars e ++ text " = 0)"
  | Bool true -> text "True"
  | Bool false -> text "False"
  | Imply (p1, p2) -> text "(" ++ printPredCoq vars p1 ++ text " -> "
	++ printPredCoq vars p2 ++ text ")"
  | All (v, p) -> text "(forall " ++ text v ++ text ", "
	++ printPredCoq (text v :: vars) p ++ text ")"
  | Exist (v, p) -> text "(exists " ++ text v ++ text ", "
	++ printPredCoq (text v :: vars) p ++ text ")"
  | _ -> E.s (E.error "Unknown Coq predicate type in Proof: %a" d_pred p)

(** Register a lemma, along with its proof *)
let registerLemma
    (l: lemma)
    (prf: proof)   (* This is the proof that 
                            ALL lvars. (AND lhyps) => lconc 
                    *)
    : unit = 
  if IH.mem allLemmas l.lname then 
    E.s (E.error "We have already registered lemma with name %s\n" 
           (U.symbolName l.lname));
  begin match !coqProc with
    None -> () (* ignore (E.warn "Ought to check the proof of the lemma %s!" l.lname); *)
  | Some (inch, outch) ->
      begin match prf with
	MagicLemma _ -> ignore (E.warn "Magic proof for lemma %s" (U.symbolName l.lname));
      | CoqLemma vtypes ->
	  let pred = List.fold_right (fun hyp pred -> Imply (hyp, pred)) l.lhyps l.lconc in
	  let vars = match l.lvars with
	    [] -> ""
	  | _ ->
	      let len = List.length l.lvars in
	      let rec mkVars vtypes n =
		match vtypes with
		  [] -> ""
		| ty::vtypes -> " (V" ^ string_of_int n ^ ":" ^ ty ^ ")" ^ mkVars vtypes (n+1) in
	      "forall" ^ mkVars vtypes 0 ^ ", " in
	  let s = "Goal " ^ vars ^ sprint 999999 (printPredCoq [] pred) ^ ".\napply " ^ U.symbolName l.lname ^ ".\nAbort.\n" in
	  (*ignore (E.log "Commands: %s\n" s);*)

	  output_string outch s;
	  flush outch;

	  let rec untilAborted () =
	    let l = input_line inch in
	    (*ignore (E.log ":: %s\n" l);*)
	    match l with
	      "Current goal aborted" -> ()
	    | _ -> untilAborted () in
	  
	  let lines = ref "" in
	  
	  let rec findResult () =
	    let line = input_line inch in
	    (*ignore (E.log ":: %s\n" line);*)
	    match line with
	      "Proof completed." ->
		untilAborted ();
		true
	    | "Error: Unknown command of the non proof-editing mode" -> findResult ()
	    | "Current goal aborted"
	    | "User error: No focused proof (No proof-editing in progress)." ->
		(match !coqFlag with
		  CoqVerbose -> ignore (E.log "**** %s failed:\n%s\n" (U.symbolName l.lname) !lines)
		| _ -> ignore (E.log "**** %s failed.\n" (U.symbolName l.lname)));
		false
	    | _ ->
		lines := !lines ^ line ^ "\n";
		findResult () in

	  if not (findResult ()) then
	    E.s (E.error "Bad Coq proof of lemma %s\n" (U.symbolName l.lname))
      | _ -> ignore (E.error "Wrong type of proof for lemma %s" (U.symbolName l.lname))
      end
  end;

  IH.add allLemmas l.lname l


let findRegisteredLemma (n: U.symbol) : lemma = 
  IH.find allLemmas n

(***** Pretty printing *****)
let rec d_prf () = function
    Hyp v -> text (U.symbolName v)
  | FreeHyps (_, p) -> d_prf () p 
  | Truei -> text "truei"

  | Falsei (d1, d2) -> dprintf "(@[false1 %a@?%a@])" d_prf d1 d_prf d2

  | Andi pfs -> text "(andi " ++ align ++ 
	(docList ~sep:(chr ',' ++ break) (d_prf ()) () pfs) ++
	text ")" ++ unalign
  | Ande (idx,pf) -> text "(ande " ++ align ++ 
        num idx ++ break ++
	(d_prf () pf) ++ text ")" ++ unalign

  | Ori (ps, i, pf) -> dprintf "(@[ori [%a]@?%d@?%a@])"
	insert (seq (text "; ") (d_pred ()) ps)
	i d_prf pf

  | Ore (goal, disj, d, prfl) -> 
(*
      dprintf "(@[ore %a@?%a@])"
        d_prf d
        (docList ~sep:break
           (fun (h, d') -> 
             dprintf "[%s : _] @[%a@]" (U.symbolName h) d_prf d'))
        prfl
*)
      let idx = ref 0 in 
      let getDisj () = 
        let d = try List.nth disj !idx
                with _ -> E.s (E.bug "printing Ore")
        in
        incr idx;
        d
      in
      dprintf "(@[ore %a@?%a@])"
        d_prf d
        (docList ~sep:break
           (fun (h, d') -> 
             dprintf "[%s : %a] @[%a@]" 
               (U.symbolName h) d_pred (getDisj ()) d_prf d'))
        prfl

  | Impi(v,p1,d2) -> 
      dprintf "(@[impi [%s : pf %a]@?%a@])" (U.symbolName v) d_pred p1 d_prf d2
  | Impe(p1,p2) -> 
      dprintf "(@[impe %a@?%a@])" d_prf p1 d_prf p2

  | AnnotE d -> dprintf "(@[annote@?%a@])" d_prf d
  | AnnotI (a, d) -> dprintf "(@[annoti@?%a@?%a@])" d_pannot a d_prf d
      
  | Cut(p1,v,p2) -> 
      dprintf "(@[cut %a [%s:pf _] %a@])"
        d_prf p1 (U.symbolName v) d_prf p2
      
  | Alli(v,d) -> 
      dprintf "(alli @[([%s : exp]@!%a)@])" v d_prf d
  | Alle(x,d) ->
      dprintf "(alle @[([%a : exp]@!%a)@])" d_exp x d_prf d

  | ExistE(d,v,h,d2) ->
      dprintf "(ex@[istE %a@?([%s : exp]@?[%s : _]@?%a)@])" 
        d_prf d v (U.symbolName h) d_prf d2

  | Contra(a,v,d) -> begin
      let preamble = 
        match a with
          PNez(BinOp(bop, e1, e2)) -> dprintf "%s %a %a"
              (binop_to_string bop) d_exp e1 d_exp e2
        | PFun (n, args) -> 
            dprintf "pred %a" d_exp (Fun(n, args))

        | Bool b -> dprintf "_%b" b

        | _ -> E.s (E.unimp "Unsupported contradiction proof: %a. prf=%a" 
                      d_pred a d_prf d)
      in
      dprintf "(con@[tra%a @!([%s : pf _]@! %a@]))" insert preamble 
        (U.symbolName v) d_prf d
  end
 
  | G(name, args) ->
      if 0 = List.length args then
        text name
      else
        dprintf "(%s @[%a@])" name (docList ~sep:break dpa) args
  | Unimplemented n -> dprintf "(unimplemented %s)" n

  | WithLemma (ln, exps, prfs) -> 
      let lvars = 
        try (findRegisteredLemma ln).lvars
        with Not_found -> 
          Util.list_init (List.length exps) (fun _ -> "?")
      in
      dprintf "Lemma(@[%s,@!%a,@!%a@])" (U.symbolName ln)
        (docList ~sep:(chr ',' ++ line) 
           (fun (n,e) -> dprintf "%s:%a" n d_exp e)) 
        (try List.combine lvars exps with Invalid_argument s ->
	  failwith ("List length mismatch while printing use of lemma " 
                    ^ (U.symbolName ln)))
        (docList ~sep:(chr ',' ++ line) (d_prf ())) prfs

  | UnrollPDef(p,prf) -> 
      dprintf "UnrollPDef(@[%a,@?%a@])"
        d_pred p d_prf prf

  | UnrollEDef(p,s,prf) -> 
      dprintf "UnrollEDef(@[%a,@?%s,@?%a@])"
        d_pred p s d_prf prf
  | RollEDef(p,s,prf) -> 
      dprintf "RollEDef(@[%a,@?%s,@?%a@])"
        d_pred p s d_prf prf

  | ArithEq (e1, e2, n, subproofs) -> 
      dprintf "(@[aritheq%a %a@?%a@?%a@])" 
        insert (if n = 1 then nil else dprintf "(*%d)" n)
        d_exp e1 d_exp e2
        (docList ~sep:break dpa) subproofs

  | ArithNe (e1, e2, n, subproofs) -> 
      dprintf "(@[arithne%a %a@?%a@?%a@])" 
        insert (if n = 1 then nil else dprintf "(*%d)" n)
        d_exp e1 d_exp e2
        (docList ~sep:break dpa) subproofs

  | ArithGeq (e1, e2, n, subproofs) -> 
      dprintf "(@[arithgeq%a %a@?%a@?%a@])" 
        insert (if n = 1 then nil else dprintf "(*%d)" n)
        d_exp e1 d_exp e2
        (docList ~sep:break dpa) subproofs

  | ArithContra (subproofs) -> 
      dprintf "(@[arithcontra@?%a@])" 
        (docList ~sep:break dpa) subproofs

  | ArithNeq (i, e1, e2) -> 
      dprintf "(@[arithneq@?%d@?%a@?%a@])" 
        i d_exp e1 d_exp e2
  | ArithApart (r,e1,e2) ->
      dprintf "(@[arithapart@?%d@?%a@?%a@])" 
        r d_exp e1 d_exp e2
  | ArithGt (i,e1,e2) ->
      dprintf "(@[arithgt@?%d@?%a@?%a@])" 
        i d_exp e1 d_exp e2

  | ArithLeq pf ->
      dprintf "(@[arithleq@?%a@])" d_prf pf
  | ArithLt pf ->
      dprintf "(@[arithlt@?%a@])" d_prf pf

  | SelUpd0 (what, mem, pfaddr) -> 
      dprintf "(@[selupd0@?%a@?%a@?%a@])" 
        d_exp what d_exp mem d_prf pfaddr
  | SelUpd1 (what, mem, prfneq) -> 
      dprintf "(@[selupd1@?%a@?%a@?%a@])" 
        d_exp what d_exp mem d_prf prfneq
  | SelUpd (sa, ua, uv, um) -> 
      dprintf "(@[selupd@?%a@?%a@?%a@?%a@])" 
        d_exp sa d_exp ua d_exp uv d_exp um
      
  | Magic (s, p) -> dprintf "(magic_proof %s@?%a)" s d_pred p
  | MagicLemma l -> dprintf "(magic_lemma %s)" (U.symbolName l.lname)
  | CoqLemma _ -> dprintf "(coq_lemma ?)"

  | Symm d -> dprintf "(symm %a)" d_prf d
  | Refl e -> dprintf "(refl %a)" d_exp e
  | Trans (e, d1, d2) -> dprintf "(@[trans %a@?%a@?%a@])"
        d_exp e d_prf d1 d_prf d2
  | Cong (e1, e2, pfs) -> 
      dprintf "(@[cong %a@?%a@?%a@])"
        d_exp e1 d_exp e2 
        (docList ~sep:break (d_prf ())) pfs
  | PCong (pf1, pfs) -> 
      dprintf "(@[pcong %a@?@[[%a]@]@])"
        d_prf pf1 (docList ~sep:(chr ',' ++ break) (d_prf ())) pfs

and dpa = function
    I(i) -> num i
  | S s -> text s
  | E(e) -> d_exp () e
  | F(p) -> d_pred () p
  | P(d) -> d_prf () d



(**************** Definitions ******************)
(** Expression definitions. The name of the defined constructor (use a 
  * lowercase name), the list of formals, and the body, referring to the 
  * formals using deBruijn indices. *)
type edef = string * string list * Logic.exp 

(* Predicate definitions. The name of the defined constructor (use an 
 * uppercase initial for the name), the list of formals, and the body, 
 * referring to the formals using deBruijn indices. *)
type pdef = string * string list * Logic.pred 

let allEDefs: (string, edef) H.t = H.create 13
let allPDefs: (string, pdef) H.t = H.create 13

(** Register a predicate definition *)
let registerPDef (x : pdef) : unit = 
  let pname, _, _ = x in
  if H.mem allPDefs pname then 
    E.s (E.error "We have already registered the predicate constructor %s\n" 
           pname);
  H.add allPDefs pname x

(** Register an expression definition *)
let registerEDef (x : edef) : unit = 
  let ename, _, _ = x in
  if H.mem allEDefs ename then 
    E.s (E.error "We have already registered the expression constructor %s\n" 
           ename);
  H.add allEDefs ename x




(****** Type inference *****)
exception ProofCheck of doc * proof * (U.symbol * pred) list

let showAssumptions (hyps: (U.symbol * pred) list) = 
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter
    (fun (v, a) -> 
      ignore (E.log " %s: %a\n" (U.symbolName v) 
                                d_pred a))
    (List.rev hyps)

(*** This is not sound w.r.t. modulo arithmetic *)
(** Check an equality: e1 = e2 + n  *)
let checkEquality (e: exp) : bool = 
  match simplifyArith e with 
    Int 0l -> true 
  | e' -> 
      ignore (E.warn "When verifying the zeroness of %a\n we simplified it to %a\n"
                d_exp e d_exp e');
      false

let makeExistVar n = Fun ("existVar", [Fun (n, [])])


let predToAssumption (p: pred) : pred = 
  match p with 
    PFun(n, args) -> eq (Fun(n, args)) one
  | NegPFun (n, args) -> ne (Fun(n, args)) one
  | _ -> p




(** The internal inference function *)
let rec iinfer (vars: var list) 
              (hyps: (U.symbol * pred) list) 
              (d: proof)  : pred = 
  let rec i (d: proof) =
    match d with 
    | Truei -> Bool true

    | FreeHyps (rs, p) -> i p

    (* Magic *)
    | Magic (s,p) ->
	if !printMagicProofWarning then begin
          ignore (E.warn "Magic proof (%s) for %t.\n" s 
		    (fun () -> begin try
		      ignore (Str.search_forward 
				(Str.regexp "HIDEPRED") s 0);
		      text "<pred omitted>"
		    with Not_found ->
		      d_pred () p
		    end));
        end;
	p

    | WithLemma (lname, exps, prfs) -> begin
        (* Lookup the lemma *)
        let lemma =
          try findRegisteredLemma lname
          with Not_found -> E.s (E.error "Cannot find registered lemma %s"
                                   (U.symbolName lname))
        in
        (* Check that we have enough instantiations *)
        if List.length exps <> List.length lemma.lvars then 
          E.s (E.error "Lemma %s expects %d instantiations and is used here with :\n\t@[%a@]\n"
                 (U.symbolName lemma.lname) 
                 (List.length lemma.lvars)
                 (docList ~sep:(chr ',' ++break) (d_exp ())) exps);

        (* Now construct a substitution context. Each of the instantiations 
         * uses the current context *)
        let subst = newSubstFromList exps in
	let theconc : pred = 
          try unrollUVarsPredRaise subst lemma.lconc
          with HasUVar -> 
            E.s (E.bug "Not enough instantiations for conclusion of lemma %s" 
                   (U.symbolName lemma.lname))
        in

        (* Now check the proofs of the hypotheses. We could use List.iter2, 
         * but this is faster, allows us to check for different list lengths, 
         * and does not confuse the profiler. *)
        let rec checkHyps (hyp_prfs: proof list) (lhyps: pred list) = 
          match hyp_prfs, lhyps with
            [], [] -> ()
          | hyp_prf :: rest_hyp_prfs, lhyp :: rest_lhyps -> 
              let hyp' = 
                try unrollUVarsPredRaise subst lhyp 
                with HasUVar -> 
                  E.s (E.bug "Not enough instantiations for hypothesis of lemma %s" 
                         (U.symbolName lemma.lname))
              in
              let hyp'' = i hyp_prf in
              if hyp' <> hyp'' then begin
                let msg = 
                  dprintf "Hypothesis %d of lemma %s: Cannot check that we prove@!          %a@! we prove %a\n" 
                    (List.length lemma.lhyps - List.length rest_lhyps)
                    (U.symbolName lemma.lname) 
                    d_pred hyp' d_pred hyp'' 
                in 
                raise (ProofCheck (msg, hyp_prf, hyps))
              end;
              checkHyps rest_hyp_prfs rest_lhyps

          | _, _ -> 
              E.s (E.error "Lemma %s expects %d subproofs and is used here with %d:\n\t@[%a@]\n"
                     (U.symbolName lemma.lname) 
                     (List.length lemma.lhyps)
                     (List.length prfs)
                     (docList ~sep:line (d_prf ())) prfs)
        in
        checkHyps prfs lemma.lhyps;
        
        theconc
    end
	  
    | UnrollPDef (prd, prf) -> begin
        let pname, pargs, (_, pd_formals, pd_body) = 
          try
            match prd with 
	      PFun(pname, pargs) -> pname, pargs, H.find allPDefs pname
            | _ -> raise Not_found
          with Not_found -> 
            E.s (E.error "UnrollPExp used on an undefined predicate %a"
                   d_pred prd)
        in
        let subst = newSubstFromList pargs in

        if List.length pd_formals <> List.length pargs then 
          E.s (E.error "Predicate constructor %s expects %d arguments and is used here as follows:\n\t%a\n"
                 pname (List.length pd_formals)
                 d_pred prd);

        (* The predicate to be proved *)
	let prd' : pred = 
          try unrollUVarsPredRaise subst pd_body
          with HasUVar -> 
            E.s (E.bug "Not enough instantiations for conclusion of defined predicate %s" pname)
        in
        
        let prd'' = i prf in
        
        if prd' <> prd'' then begin
          let msg = dprintf "UnrollPDef %s does not prove %a\nIt proves: %a\n"
              pname d_pred prd'
	      d_pred prd'' in
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd
    end
	
    | RollEDef (prd, ename, prf) -> begin
        let _, ed_formals, ed_body = 
	  try
	    H.find allEDefs ename
          with Not_found -> 
            E.s (E.error "UnrollEDef used with an undefined constructor %s"
                   ename)
        in

	(match ed_formals with
	     [] -> ()
	   | _ -> failwith "Unimplemented: Non-nullary expression shortcuts");

	let subber depth = function
	    Fun (f, []) when f = ename -> Some ed_body
	  | _ -> None in

	let prd' = substPred subber prd in

        let prd'' = i prf in
        
        if prd <> prd'' then begin
          let msg = dprintf "UnrollEDef %s does not prove %a\nIt proves: %a\n"
                 ename d_pred prd
		 d_pred prd'' in 
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd'
    end
    | UnrollEDef (prd, ename, prf) -> begin
        let _, ed_formals, ed_body = 
	  try
	    H.find allEDefs ename
          with Not_found -> 
            E.s (E.error "UnrollEDef used with an undefined constructor %s"
                   ename)
        in

	(match ed_formals with
	     [] -> ()
	   | _ -> failwith "Unimplemented: Non-nullary expression shortcuts");

	let subber depth = function
	    Fun (f, []) when f = ename -> Some ed_body
	  | _ -> None in

	let prd' = substPred subber prd in

        let prd'' = i prf in
        
        if prd' <> prd'' then begin
          let msg = 
            dprintf "UnrollEDef %s does not prove %a\nIt proves: %a\n"
              ename d_pred prd'
	      d_pred prd'' in 
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd
    end

    (* And 
    | G("andi",  [P d1; P d2]) -> And [i d1; i d2]
    | G("andel", [P d]) -> begin
        match i d with
          And (p1:: _) -> p1
        | _ -> raise (ProofCheck (text "andel does not prove an And",
                                  d, hyps))
    end
    | G("ander", [P d]) -> begin
        match i d with
          And (_ :: p2 :: _) -> p2
        | _ -> raise (ProofCheck (text "ander does not prove an And",
                                  d, hyps))
    end
*)
    | Andi ds -> 
        (* We define our own List.map to not confuse the Profiler *)
        let rec subproofs = function
            [] -> []
          | d :: ds -> i d :: subproofs ds
        in
        And (subproofs ds)

    | Ande (idx,d) -> begin
	match i d with
	  And ds -> List.nth ds idx
	| _ -> raise (ProofCheck (text "Ande does not prove an And",
                                  d, hyps))
    end

    | Ori (ps, idx, pf) ->
        let p = i pf in 
	if idx >= 0 && idx < List.length ps && p = List.nth ps idx then
	  Or ps
	else begin 
          let msg = 
            if idx < 0 || idx >= List.length ps then 
              text "invalid index"
            else 
              dprintf "subproof proves\n %a\nexpected\n %a\n" 
                d_pred p d_pred (List.nth ps idx)
          in
	  raise (ProofCheck ((dprintf "Ori error (%a)" insert msg), d, hyps))
        end

    | Ore (goal, _, pfor, cases) -> begin
        match i pfor with 
          Or orl -> 
            if List.length orl <> List.length cases then 
              raise (ProofCheck((dprintf "Ore: disjunct has %d cases and we have %d\n" 
                                   (List.length orl) (List.length cases)),
                               d, hyps));
            List.iter2 (fun disjunct (disjhyp, case) -> 
              let hyps' = (disjhyp, disjunct) :: hyps in 
              let caseproves = iinfer vars hyps' case in 
              if caseproves <> goal then 
                raise (ProofCheck((dprintf "Ore: case proves %a not %a\n"
                                     d_pred caseproves d_pred goal),
                                 case, hyps')))
              orl cases;
            goal

        | p -> 
            raise (ProofCheck((dprintf "Ore does not have proof of Or: %a\n" 
                                 d_pred p),
                             pfor, hyps))
    end
(*        
    | G("oril", [P d; F other]) -> Or [i d; other]
    | G("orir", [F other; P d]) -> Or [other; i d]
*)
    | Alli(v,d) -> All(v, iinfer (v :: vars) hyps d)
    | Alle (x, pf) ->
	(match i pf with
	     All (v, body) ->
	       let name = Some x in
	       let subber depth = function
		   Var i ->
		     if i - depth = 0 then
		       name
		     else
		       None
		 | _ -> None in
	       let body' = substPred subber body in
	       body'
	   | _ -> raise (ProofCheck (text "Alle does not prove All",
                                     pf, hyps)))

    | ExistE (pfex, name, hyp, pf) -> begin 
        (* Make sure name is fresh *)
        if List.exists (fun v -> v = name) vars then 
          raise (ProofCheck ((dprintf "Repeated instantiation in ExistE (%s) vars=%a"
                                 name (docList text) vars), pf, hyps));
	match i pfex with
	  Exist (v, body) ->
            (* Make body[name/v] *)
            let body' = 
               substPred
                  (fun depth e -> 
                    match e with 
                      Var d when d = depth -> Some (NVar name)
                    | _ -> None)
                  body
           in
           iinfer (name :: vars) ((hyp, body') :: hyps) pf

        | p -> 
            raise (ProofCheck (text "Existe does not prove Exist",
                               d, hyps))
    end
          
       
    | Impi(v,p1,d2) -> 
        Imply(p1, iinfer vars ((v, p1) :: hyps) d2)

    | Impe(d1,d2) ->
	(match i d1 with
	     Imply (hyp, conc) ->
	       let prd = i d2 in
		 if prd <> hyp then begin
		   let msg = dprintf 
                       "Impe needs hypothesis %a\nBut given: %a\n"
			  d_pred hyp
			  d_pred prd in
		   raise (ProofCheck (msg, 
                                      d2, hyps))
		 end;
		 conc
	  | prd ->
	      let msg = dprintf "Impe given proof of non-implication: %a\n"
		     d_pred prd in 
	      raise (ProofCheck (msg, d2, hyps)))

    | AnnotE d -> begin 
        match i d with 
          Pannot (_, p) -> p
        | p' -> 
            let msg = dprintf "AnnotE given proof not Pannot: %a\n" d_pred p'
            in
            raise (ProofCheck (msg, d, hyps)) 
    end
    | AnnotI (a, d) -> Pannot(a, i d)

    | Contra(p,v,d1) -> begin
        let assum = p in 
        match iinfer vars ((v, negatePred assum) :: hyps) d1 with
          Bool false -> p
        | _ -> raise (ProofCheck (text "Contra does not prove false",
                                  d, hyps))
    end

    | Cut (prf1, v, prf2) -> 
        let p1 = i prf1 in 
        iinfer vars ((v, p1) :: hyps) prf2

    | (Hyp v) -> begin
        try 
          let (_, p) = List.find (fun (v',_) -> v = v') hyps in
          p
        with Not_found -> raise (ProofCheck (dprintf "Hyp %s not found"
                                               (U.symbolName v),
                                             d, hyps))
    end
 
    (* false elimination *)
    | G("falsee", [ F p ; P pff ]) -> begin
        match i pff with
          Bool false -> p
        |   _ -> raise (ProofCheck (nil, d, hyps))
    end

    | G("doselmap", [ E m; I idx]) -> begin
        try
          let e = findMap m idx in
          eq e (Fun("selmap", [ m; integer idx]))
        with Not_found -> begin
          let msg = dprintf "doselmap: cannot find index %d\n" idx in 
          raise (ProofCheck (msg, d, hyps))
        end
    end

    | G("doupdmap", [ E m; I idx; E newval ]) -> begin
        let m' = insertMap m idx newval in
        eq m' (Fun("updmap", [ m; integer idx; newval ]))
    end

    | G("haveid", [ E e]) -> pFun "have" [ e ]


    | ArithNeq (r, e1, e2) -> 
	if checkEquality (minus (minus e1 e2) (integer r)) && r <> 0 then 
	  ne e1 e2
	else begin 
          let msg = dprintf "ArithNeq checkEquality failed" in 
	  raise (ProofCheck (msg, d, hyps))
        end

    | ArithEq (e1, e2, n, subproofs) -> begin
        if n = 0 then begin
          let msg = dprintf "ArithEq with factor 0" in 
          raise (ProofCheck (msg, d, hyps));
        end;
        let res = residueOfArithmeticSubproofs 
            ~ineq:false
            ~infer:i
            ~err:(fun msg -> raise (ProofCheck(msg, d, hyps)))
            subproofs 
        in
        match simplifyArith (minus (times (integer n) 
                                          (minus e1 e2)) res) with 
          Int c when c = 0l -> eq e1 e2
        | rest -> 
            let msg = dprintf "residue of ArithEq is %a\n" d_exp rest in 
            raise (ProofCheck (msg, d, hyps))
    end

    | ArithNe (e1, e2, n, subproofs) -> begin
        if n = 0 then begin
          let msg = dprintf "ArithNe with factor 0" in 
          raise (ProofCheck (msg, d, hyps));
        end;
        let res = residueOfArithmeticSubproofs 
            ~ineq:false
            ~infer:i
            ~err:(fun msg -> raise (ProofCheck(msg, d, hyps)))
            subproofs 
        in
        match simplifyArith (minus (times (integer n) 
                                          (minus e1 e2)) res) with 
          Int c when c <> 0l -> ne e1 e2
        | rest -> 
            let msg = dprintf "residue of ArithNe is %a\n" d_exp rest in 
            raise (ProofCheck (msg, d, hyps))
    end


    | ArithGeq (e1, e2, n, subproofs) -> begin
        if n <= 0 then begin
          let msg = dprintf "ArithGeq with negative factor %d" n in 
          raise (ProofCheck (msg, d, hyps));
        end;
        let res = residueOfArithmeticSubproofs 
            ~ineq:true 
            ~infer:i
            ~err:(fun msg -> raise (ProofCheck(msg, d, hyps)))
            subproofs 
        in
        (* We know that the residue is >= 0 *)
        match simplifyArith (minus (times (integer n) (minus e1 e2)) res) with 
          Int c when (Int32.to_int c) + n > 0 -> geq e1 e2
        | rest -> 
            let msg = dprintf "residue of ArithGeq is %a   @!(n=%d, e1=%a,e2=%a,res=%a)\n" d_exp rest n d_exp e1 d_exp e2 d_exp res in 
            raise (ProofCheck (msg, d, hyps))
    end

    | ArithContra subproofs -> begin
        let res = residueOfArithmeticSubproofs 
            ~ineq:true
            ~infer:i
            ~err:(fun msg -> raise (ProofCheck(msg, d, hyps)))
            subproofs 
        in
        (* We know that the residue is >= 0 *)
        match simplifyArith res with 
          Int c when (Int32.to_int c) < 0 -> Bool false
        | rest -> 
            let msg = dprintf "residue of ArithContra is %a\n" d_exp rest in 
            raise (ProofCheck (msg, d, hyps))
    end

    | ArithGt (r, e1, e2) ->
        let z = minus (minus e1 e2) (integer  r) in
	if checkEquality  z && r > 0 then 
	  gt e1 e2
	else begin
          let msg = 
            dprintf "ArithGt: cannot check equality of %a\n" d_exp z in 
	  raise (ProofCheck (msg, d, hyps))
        end

    | ArithLeq pf ->
	begin match i pf with
	  PNez (BinOp (SetGt, e, e')) -> leq e' e
	| _ -> raise (ProofCheck (dprintf "ArithLeq: bad parameter",
				  d, hyps))
	end

    | ArithLt pf ->
	begin match i pf with
	  PNez (BinOp (SetGe, e, e')) -> lt e' e
	| _ -> raise (ProofCheck (dprintf "ArithLt: bad parameter",
				  d, hyps))
	end

    | ArithApart (r, e1, e2) -> 
        if checkEquality (minus (minus e1 e2) (integer r)) && 
           (r <= -4 || r >= 4) then
	  PFun("Apart", [e1; e2])
        else begin 
          let msg = dprintf "ArithApart" in 
          raise (ProofCheck (msg, d, hyps))
        end


    | SelUpd0 (what, mem, pfeq) ->  begin 
        match i pfeq with 
          PNez(BinOp(SetEq, seladdr, updaddr)) -> 
            eq (Fun("sel", [ seladdr; 
                             Fun("upd", [ updaddr; what; mem ])])) what
        | _ -> 
            let msg = dprintf "selupd0: cannot prove equality" in
            raise (ProofCheck (msg, d, hyps))
    end

    | SelUpd1(what, mem, prfneq) -> begin
        match i prfneq with 
          PNez(BinOp(SetNe, seladdr, updaddr)) -> 
            eq (Fun("sel", [seladdr; 
                            Fun("upd", [ updaddr; what; mem])]))
               (Fun("sel", [seladdr; mem ]))
        | _ -> 
            let msg = dprintf "selupd1: cannot prove disequality" in
            raise (ProofCheck (msg, d, hyps))
    end
        
    | SelUpd(sa, ua, uv, um) -> begin 
        let upd = Fun("upd", [ ua; uv; um ]) in
        let sel = Fun("sel", [ sa; upd ]) in
        let sel1 = Fun("sel", [ sa; um ]) in 
        Or [ And [ eq sa ua; eq sel uv ]; And [ ne sa ua; eq sel sel1 ] ]
    end
	  (* arithmetic axioms 
    | G ("eqCheck", [E e1; E e2]) ->
	if checkEquality (minus e1 e2) then 
	  eq e1 e2
	else begin 
          let msg = dprintf "eqCheck" in 
	  raise (ProofCheck (msg, d, hyps))
        end


    | G ("geqCheck", [E e1; E e2; I r]) ->
        let z = minus (plus e1 (integer  r)) e2 in
	if checkEquality z && 0 >= r then 
	  geq e1 e2
	else begin
          let msg = dprintf "geqCheck: cannot check equality of %a\n" d_exp z 
          in
	  raise (ProofCheck (msg, d, hyps))
        end

    | G("geq0", []) -> geq zero zero
          
    | G("eq0", []) -> eq zero zero 
*)
       
    | G("gtNeq", [ P d ]) -> begin
        match i d with 
          PNez(BinOp(SetGe, e1, BinOp(Plus, e2, o))) when o = one -> 
            ne e1 e2
        | _ -> begin
            let msg = dprintf "gtNeq" in 
            raise (ProofCheck (msg, d, hyps))
        end
    end
       
    | G("neqGt", [P d1; P d2]) -> begin
        match i d1, i d2 with
          PNez (BinOp(SetGe, e1, e2)), 
          PNez (BinOp(SetNe, e1', e2')) when e1 = e1' && e2 = e2' -> 
            geq e1 (plus e2 one)
        | p1, p2 -> 
            let msg = dprintf "neqGt: %a@!   %a@!" d_pred p1 d_pred p2 in 
            raise (ProofCheck (msg, d, hyps))
    end   
    | G("eqi", [P d1; P d2]) -> begin
	match (i d1, i d2) with
	  (PNez (BinOp(SetGe, e1,e2)), 
           PNez (BinOp(SetGe, e_2, e_1))) ->
	     if (e1 = e_1) && (e2 = e_2) then begin
               eq e1 e2
	    end else
	      raise (ProofCheck (text "eqi: 1", d, hyps))

	| _ -> raise (ProofCheck (text "eqi: 2", d, hyps))
    end

    | G("eqgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1,e2)) -> geq e1 e2
	| _ -> raise (ProofCheck (text "eqgeq", d, hyps))
    end

    | G("eqgeqr", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1,e2)) -> geq e2 e1
	| _ -> raise (ProofCheck (text "eqgeqr", d, hyps))
    end
    | G("ltgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetLt, e1,e2)) -> geq e2 (plus e1 one)
	| h -> raise (ProofCheck ((dprintf "ltgeq: hypothesis is %a" d_pred h),
                                  d, hyps))
    end
    | G("geqlt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e2, BinOp(Plus, e1, o))) when o = one ->
            lt e1 e2
	| _ -> raise (ProofCheck (text "geqlt", d, hyps))
    end
    | G("leqgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetLe, e1,e2)) -> geq e2 e1
	| _ -> raise (ProofCheck (text "leqgeq", d, hyps))
    end
    | G("geqleq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe,e1,e2)) -> leq e2 e1
	| _ -> raise (ProofCheck (text "geqleq", d, hyps))
    end
    | G("gtlt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGt,e1,e2)) -> lt e2 e1
	| _ -> raise (ProofCheck (text "gtlt", d, hyps))
    end

    | G("gtgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGt, e1,e2)) -> geq e1 (plus e2 one)
	| _ -> raise (ProofCheck (text "gtgeq", d, hyps))
    end

    | G("geqgt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e1, BinOp(Plus, e2, o))) when o = one ->
            gt e1 e2
	| _ -> raise (ProofCheck (text "geqgt", d, hyps))
    end

(*
    | G("eqeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1, e2)) -> eq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "eqeq0", d, hyps))
    end

    | G("eqgeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1, e2)) -> geq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "eqgeq0", d, hyps))
    end

    | G("geqgeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e1, e2)) -> geq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "geqgeq0", d, hyps))
    end


    | G("geqLt_FI", [ P d1; P d2]) -> begin
	match (i d1, i d2) with
	  (PNez (BinOp(SetGe, e1', e2')),
           PNez (BinOp(SetGe, e2'', BinOp(Plus, e1'', o)))) when o = one ->
	     if (e1'' = e1') && (e2' = e2'') then
	       Bool false
	     else 
	       raise (ProofCheck (text "geqLt_FI: 0", d, hyps))
                 
	| _ -> raise (ProofCheck (text "geqLt_FI: 1", d, hyps))
    end
*)


    | G ("not_FI",[P prf1; P prf2]) -> begin
        let p1 = i prf1 in 
        let p2 = i prf2 in 
        let np1 = 
          try negatePredRaise p1 
          with CannotNegate _ -> begin
            let msg = dprintf "Cannot negate %a in not_FI" d_pred p1 in 
            raise (ProofCheck (msg, d, hyps))
          end
        in
        if p2 = np1 then 
          Bool false
        else begin 
          let msg = dprintf "Proof: not_FI: ???\n" in 
          raise (ProofCheck (msg, d, hyps))
        end
    end      

    | Refl e -> eq e e

    | Symm p -> begin
	match i p with
	  PNez (BinOp(SetEq, e1,e2)) -> eq e2 e1
	| PNez (BinOp(SetNe, e1,e2)) -> ne e2 e1
	| _ -> 
            let msg = 
              dprintf "Summ: invalid use of axiom! in %a" 
                d_prf d in 
            raise (ProofCheck (msg, d, hyps))
    end

    | Trans (e2, p1, p2) -> begin
	match i p1, i p2 with
	  (PNez (BinOp(SetEq, e1, e2')),
           PNez (BinOp(SetEq, e2'', e3))) ->
	     if e2 = e2'' && e2 = e2' then 
	       eq e1 e3
	     else 
               (ignore (E.warn "eqtr: invalid use of axiom! in %a@!e2=e2''=%b@!e2=e2'=%b@!e2=%a@!e2''=%a@!e2'=%a@!" 
                          d_prf d (e2=e2'') (e2=e2') 
                          d_exp e2 d_exp e2'' d_exp e2');
                showAssumptions hyps;
                raise (ProofCheck (nil, d, hyps)))

	| _ -> 
            let msg= dprintf "Trans: subproofs are not equalities@! in %a" 
                d_prf d in
            raise (ProofCheck (msg, d, hyps))
    end

    | G("neqtr", [ P peqa; E _; P pneq; E _; P peqb ]) -> begin
        let a, a' = match i peqa with
          PNez (BinOp(SetEq, a, a')) -> a, a'
        | _ -> 
            let msg = dprintf "1st premise of neqtr is not an equality" in 
            raise (ProofCheck (msg, peqa, hyps))
        in
        let b', b = match i peqb with
          PNez (BinOp(SetEq, b', b)) -> b', b
        | _ -> 
            let msg = dprintf "3rd premise of neqtr is not an equality" in 
            raise (ProofCheck (msg, peqb, hyps)) 
        in
        match i pneq with 
          PNez (BinOp(SetNe, a'', b'')) when a' = a'' && b' = b'' ->
            ne a b
        | _ -> 
            let msg = dprintf "2nd premise of neqtr does not prove %a <> %a" 
                d_exp a' d_exp b'
            in 
            raise (ProofCheck (msg, pneq, hyps)) 
    end
        
        

    | Cong (((Fun (f, args)) as e1), ((Fun (f', args')) as e2), eqprfs) ->
        (* find the arguments *)
        let rec loop (eqpfs: proof list) : (exp list * exp list) = 
            match eqpfs with 
               [] -> [], []
             | eqprf :: rest -> begin
                 let a1, a2 = loop rest in
                 match i eqprf with 
                    PNez (BinOp(SetEq, e1, e2)) -> e1 :: a1, e2 :: a2
                  | _ -> 
                      let msg = dprintf "Subproof of \"congr\" is not =" in 
                      raise (ProofCheck (msg, d, hyps))
             end
         in
         let a1, a2 = loop eqprfs in 
         (* We assume for now that the name corresponds to a function *)
         eq e1 e2 

    | Cong (((BinOp (bop, a1, a2)) as e1), ((BinOp (bop', a1', a2')) as e2), [pf1; pf2])
      when bop = bop' ->
	begin match i pf1 with
	  PNez (BinOp (SetEq, e1, e2)) when e1 = a1 && e2 = a1' -> ()
	| _ ->
	    let msg = dprintf "Bad first subproof for BinOp congruence" in
	    raise (ProofCheck (msg, d, hyps))
	end;

	begin match i pf2 with
	  PNez (BinOp (SetEq, e1, e2)) when e1 = a2 && e2 = a2' -> ()
	| _ ->
	    let msg = dprintf "Bad second subproof for BinOp congruence" in
	    raise (ProofCheck (msg, d, hyps))
	end;

        eq e1 e2 

    | PCong (pf1, pfs) ->
        let rec map f = function [] -> [] 
                        | h::t-> let h'=f h in h' :: map f t in 
	let p1, ps = i pf1, map i pfs in
	let fail () =
	  let msg = dprintf "Mismatch for PCong@! First proves: %a@!Second proves: @[[%a]@]@!" 
              d_pred p1 (docList ~sep:(chr ',' ++ break) (d_pred ())) ps in
	  raise (ProofCheck (msg, d, hyps))
	in

	begin match p1, ps with
	  PNez e1, [ PNez (BinOp (SetEq, e1', e2)) ] when e1 = e1' ->
	    PNez e2
	| PFun (s, es), preds ->
	    let e2s =
	      try
		List.fold_right2
		  (fun e1 p acc ->
		    match p with
		      PNez (BinOp (SetEq, e1', e2)) when e1 = e1' -> e2 :: acc
		    | _ -> fail ())
		  es preds []
	      with Invalid_argument _ -> fail ()
	    in
	    PFun (s,e2s)
	| p1, ps -> fail ()
	end

    (* ------ BEGIN: MLSS proof rules ------- *)
    | G(proofRule, (P subProof) :: extraArgs) 
      when (proofRule = "subset_elim" ||
            proofRule = "ssubset_elim" ||
            proofRule = "nsubset_elim" ||
            proofRule = "nssubset_elim" ||
            proofRule = "set_extensionality_split") -> (
        (* some functions borrowed from mlss.ml ... *)
        let applyElemOf (a1: exp) (a2: exp) : exp =
          (Fun ("elemOf", [a1;a2])) in
        let predElemOf (a1: exp) (a2: exp): pred =
          (eq trueExp (applyElemOf a1 a2)) in
        let predNotElemOf (a1: exp) (a2: exp): pred =
          (ne trueExp (applyElemOf a1 a2)) in
        let applyUnion (a1: exp) (a2: exp) : exp =
          (Fun ("union", [a1;a2])) in

        (* compute the predicate proved by 'subProof' *)
        let subPred:pred = (i subProof) in

        match (proofRule, extraArgs, subPred) with
        | ("subset_elim",
           [E c],
           PNez(BinOp(SetEq, Fun("subsetOf", [a;b]), Fun("true",[])))) ->
            (eq (applyUnion a c) b)

        | ("ssubset_elim",
           [E c; E x],
           PNez(BinOp(SetEq, Fun("strictSubsetOf", [a;b]), Fun("true",[])))) ->
            (And [(eq (applyUnion a c) b);
                  (predNotElemOf x a);
                  (predElemOf x b)])

        | ("nsubset_elim",
           [E x],
           PNez(BinOp(SetNe, Fun("subsetOf", [a;b]), Fun("true",[])))) ->
            (And [(predElemOf x a);
                  (predNotElemOf x b)])

        | ("nssubset_elim",
           [E x],
           PNez(BinOp(SetNe, Fun("strictSubsetOf", [a;b]), Fun("true",[])))) ->
            (Or [(eq a b);
                 (And [(predElemOf x a);
                       (predNotElemOf x b)])])

        | ("set_extensionality_split",
           [E witnessExp],
           PNez(BinOp(SetNe, a, b))) ->
            (Or [(And [(predElemOf witnessExp a);
                       (predNotElemOf witnessExp b)]);
                 (And [(predElemOf witnessExp b);
                       (predNotElemOf witnessExp a)])])

        | _ -> (
            let msg = (dprintf "bad %s:@!%a@!" proofRule d_prf d) in
            raise (ProofCheck (msg, d, hyps))
          )
      )

    | G(proofRule, args)
      when (proofRule = "emptySet_not_empty" ||
            proofRule = "set_inclusion_cycle"
           ) -> (
        (* Complaint #1: If I say the return type is 'a, then it thinks
         * I am talking about the same 'a as in residueOfArithmeticSubproofs,
         * leading to an error there.  There is no syntactic distinction
         * between a binding introduction of a type variable (which is
         * what I intend) and a use of type variable (which is what ocaml
         * decides I mean).
         *
         * Complaint #2: Even if I say 'alpha, which does not occur
         * anywhere else in this file, it still binds it to 'pred' the
         * first time it is used, but I have two call sites that need
         * it to be 'exp', so I have to arbitrarily pick some exp to
         * "return" after calling 'fail'.
         *)
        let fail (why: string) : 'alpha =
        begin
          let msg = (dprintf "bad %s, %s:@!%a@!" proofRule why d_prf d) in
          raise (ProofCheck (msg, d, hyps))
        end in

        (* every element of 'args' is a subproof; compute the facts
         * proved by those subproofs *)
        let mapArg (a: parg) : pred =
        begin
          match a with
          | P(subproof) -> (i subproof)
          | _ -> (fail "non-proof proof arg")
        end in
        let subfacts: pred list = (List.map mapArg args) in

        match (proofRule, subfacts) with
        | ("emptySet_not_empty",
           [PNez(BinOp(SetEq, Fun("elemOf", [lhs;rhs]), Fun("true",[])));
            PNez(BinOp(SetEq, rhs', Fun("emptySet",[])))]) when rhs=rhs' ->
          begin
            Bool false
          end

        | ("set_inclusion_cycle",
           _ :: PNez(BinOp(SetEq, Fun(_, [_; terminus1]), _)) :: _) ->
          begin
            (* 'args' should be a sequence of pairs of facts, the first
             * an equality between 'other' and 'lhs', and the second
             * the inclusion of 'lhs' in 'prevNode'; 'other' becomes the
             * new 'prevNode'; the last 'prevNode' is returned *)
            let rec loop (prevNode: exp) (args: pred list) : exp =
            begin
              match args with
              | [] -> prevNode    (* return last one *)
              | PNez(BinOp(SetEq, other, lhs)) ::
                  PNez(BinOp(SetEq, Fun("elemOf", [lhs';rhs]), Fun("true",[]))) ::
                  restOfArgs when
                  (prevNode=rhs && lhs=lhs') ->
                begin
                  (loop other restOfArgs)
                end
              | p1::p2::_ -> 
                begin
                  let msg = (dprintf "bad set_inclusion_cycle element pair:@!\
                                      subfacts=%a@!prevNode=%a, p1=%a, p2=%a@!"
                                      (d_list ", " d_pred) subfacts
                                      d_exp prevNode
                                      d_pred p1
                                      d_pred p2) in
                  raise (ProofCheck (msg, d, hyps))
                end
              (* see Complaint #2 above *)
              | _ -> ignore (fail "odd # of elems"); prevNode
            end in
            let terminus2:exp = (loop terminus1 subfacts) in

            (* endpoints should be the same *)
            if (terminus1 <> terminus2) then
              (fail "mismatching termini")
            else
              Bool false
          end

        | _ ->
          begin
            let msg = (dprintf "bad %s:@!%a@!" proofRule d_prf d) in
            raise (ProofCheck (msg, d, hyps))
          end
      )
    (* ------ END: MLSS proof rules ------- *)

    | d -> E.s (E.unimp "Proof.infer:@!%a@!" d_prf d)
  in
  i d

    (** Compute the residue of a linear combination of arithmetic subproofs *)
and residueOfArithmeticSubproofs ~(ineq:bool) 
                                 ~(err: doc -> 'a) 
                                 ~(infer: proof -> pred)
                                 (subproofs: parg list) : exp = 
  let rec loop = function
      [] -> zero
    | I f :: P d :: rest -> begin
        match infer d with 
          PNez(BinOp(SetGe, e1, e2)) as p -> 
            if not ineq then 
              let msg = 
                dprintf "inequality subproofs for equality proofs not allowed:\n  proof %a\n  proves %a" d_prf d d_pred p in 
              err msg
            else if f < 0 then begin
              let msg = 
                dprintf "factor %d should be non-negative for inequality proofs" f in
              err msg
            end else begin
              let r = loop rest in 
              plus (times (integer f) (minus e1 e2)) r
            end

        | PNez(BinOp(SetEq, e1, e2)) -> 
            let r = loop rest in 
            plus (times (integer f) (minus e1 e2)) r

        | p -> 
            let msg = 
              dprintf "arithmetic subproof is not >= or =:\n  proof %a\n  proves %a\n"
                d_prf d d_pred p in 
            err msg
    end
    | _ -> 
        let msg = dprintf "unexpected arithmetic subproof" in 
        err msg
  in
  loop subproofs


let check (vars: string list) 
          (hyps: (U.symbol * pred) list) 
          (prf: proof) 
          (goal: pred) = 
  try
    let ipred = iinfer vars hyps prf in
    if ipred <> goal (* not (equalPred ipred goal) *) then begin
      let msg = dprintf "Proof proves a different predicate from the goal@!Got :%a@!Need:%a" d_pred ipred d_pred goal in 
      raise (ProofCheck (msg, prf, hyps))
    end
        
  with ProofCheck (msg, d, hyps) -> 
    let old = !printDepth in 
    printDepth := 64; (* Sometimes we have huge proofs *)
    ignore (E.warn "Invalid proof:@!%a@!of@!%a" 
              d_prf d d_pred goal);
    ignore (E.warn "Whole proof is:@!%a@!" d_prf prf);
    printDepth := old;
    showAssumptions hyps;
    E.s (E.error "Proof checking failed:\n  %a\n" insert msg)

  | E.Error -> 
    ignore (E.warn "Error checking proof:@!%a" d_prf prf);
    showAssumptions hyps;
    E.s (E.error "Proof checking failed")
      

(** This is the exported inference function *)
let infer (vars: string list) 
          (hyps: (U.symbol * pred) list) 
          (prf: proof) = 
  try
    let ipred = iinfer vars hyps prf in
    ipred

  with ProofCheck (msg, d, hyps) -> 
    ignore (E.warn "Invalid proof:@!%a@!" 
              d_prf d);
    ignore (E.warn "Whole proof is:@!%a@!" d_prf prf);
    showAssumptions hyps;
    E.s (E.error "Proof checking failed:\n  %a\n" insert msg)

  | E.Error -> 
    ignore (E.warn "Error checking proof:@!%a" d_prf prf);
    showAssumptions hyps;
    E.s (E.error "Proof checking failed")
      
(** Scan a proof and watch for free and bound hypotheses. pushBound is 
 * invoked before scanning the scope of a bound hypothesis. popBound is 
 * called after the body was scanned. If this function 
 * returns "false" then the body is not scanned, and the matching popBound is 
 * not called. *)

let scanHyps ~(pushBound: U.symbol -> bool)
             ~(popBound: U.symbol -> unit)
             ~(seenHyp: U.symbol -> unit) 
             ~(seenFreeHyps: relset -> proof -> bool)
            (d: proof) : unit = 
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  let rec scanHypsWithBound (b: U.symbol) (d: proof) : unit = 
    if pushBound b then begin
      scanHypsInt d;
      popBound b
    end

  and scanHypsInt (d: proof) = 
    match d with 
      Hyp h -> seenHyp h

    | FreeHyps (rs, p) -> begin
        if seenFreeHyps rs p then 
          scanHypsInt p
        else 
          ()
    end

      (* no subproofs *)
    | Truei 
    | ArithNeq _
    | ArithApart _
    | ArithGt _ 
    | Refl _ 
    | SelUpd0 _ 
    | Magic _ 
    | MagicLemma _ 
    | CoqLemma _ 
    | Unimplemented _ 
    | SelUpd _ 
        -> ()

      (* unary *)
    | Symm d
    | Ande (_, d) 
    | Ori (_, _, d)
    | Alli (_, d) 
    | Alle (_, d) 
    | UnrollPDef (_, d) 
    | UnrollEDef (_, _, d) 
    | RollEDef (_, _, d) 
    | ArithLt d
    | ArithLeq d
    | SelUpd1 (_, _, d)
    | AnnotE d
    | AnnotI (_, d)
      -> scanHypsInt d

      (* binary *)
    | Trans (_, d1, d2) 
    | Falsei (d1, d2)
    | Impe (d1, d2) -> scanHypsInt d1; scanHypsInt d2

      (* proof list *)
    | WithLemma (_, _, dl) 
    | Cong (_, _, dl) 
    | Andi dl -> 
        iter scanHypsInt dl

    | PCong (d1, dl) -> iter scanHypsInt (d1 :: dl)

    | G (_, args) | ArithEq (_, _, _, args) | ArithGeq (_, _, _, args) 
    | ArithNe (_, _, _, args) | ArithContra args -> 
        iter (function P d -> scanHypsInt d | _ -> ()) args
            
      (* hypothetical *)
    | Impi (b, _, d) -> scanHypsWithBound b d
    | Ore (_, _, d, dl) -> 
        scanHypsInt d; 
        iter (fun (b, d) -> scanHypsWithBound b d) dl

    | Cut (d, b, d') -> scanHypsInt d; scanHypsWithBound b d'
    | Contra (_, b, d) -> scanHypsWithBound b d
    | ExistE (d, _, h, d2) -> scanHypsInt d; scanHypsWithBound h d2 
  in
  scanHypsInt d

let getFreeHyp (h: unit IH.t) (d: proof) : unit = 
  let bound: unit IH.t = IH.create 13 in
  scanHyps ~pushBound:(fun b -> IH.add bound b (); true)
           ~popBound:(fun b -> IH.remove bound b)
           ~seenHyp:(fun b -> if not (IH.mem bound b) then IH.replace h b ())
           ~seenFreeHyps:(fun _ p -> true)
           d

let getAllHyp (h: unit IH.t) (d: proof) = 
  scanHyps ~pushBound:(fun _ -> true)
           ~popBound:(fun _ -> ())
           ~seenHyp:(fun b -> IH.replace h b ())
           ~seenFreeHyps:(fun _ p -> true)
           d


(** Prepare arithmetic proof arguments **)
let prepArithArgs ~(foreq: bool) (a: parg list) : parg list = 
  (* Keep here a list of the proofs with their factors. We compare proofs by 
   * pointer equality. *)
  let collect: (int ref * proof) list ref = ref [] in
  let addOne (f: int) (p: proof) = 
    let rec loop = function
        [] -> collect := (ref f, p) :: !collect
      | (f', p') :: rest -> if p == p' then f' := f + !f' else loop rest
    in
    loop !collect
  in
  let rec mkArgs acc = function
      [] -> acc
    | (f, p) :: rest -> 
        mkArgs (if !f = 0 then acc else I !f :: P p :: acc) rest
  in
  let rec loopPargs (fact: int) = function
      [] -> ()
    | I f :: P prf :: rest -> begin
        doOneParg (fact * f) prf; loopPargs fact rest
    end
    | _ -> assert false

       
  and doOneParg (fact: int) (prf: proof) : unit = 
    match prf with 
      ArithGeq (_, _, f', args) 
    | ArithEq (_, _, f', args) -> 
        if f' = 1 then 
          loopPargs fact args
        else begin (*
          ignore (E.warn "loopPargs for Arith proof with factor %d" f');
          *)
          addOne fact prf
        end
            
    | Refl _ -> (* drop *) ()
          
    | Trans(_, p1, p2) -> doOneParg fact p1; doOneParg fact p2
          
    | Symm p -> doOneParg (- fact) p


    | G("eqi", [ P geq; P leq ]) when not foreq -> 
        if fact > 0 then 
          doOneParg fact geq
        else
          doOneParg (- fact) leq

    | _ -> addOne fact prf

  in
  loopPargs 1 a; 
  mkArgs [] !collect



(** A function to make arithmetic ArithEq proofs, with optimization of the 
 * arguments *)
let mkArithEq (e1: exp) (e2: exp) ?(fact=1)
              (args: parg list) : proof = 
  ArithEq (e1, e2, fact, prepArithArgs ~foreq:true args)

let mkArithGeq (e1: exp) (e2: exp) ?(fact=1)
               (args: parg list) : proof = 
  ArithGeq (e1, e2, fact, prepArithArgs ~foreq:false args)

let mkArithContra (args: parg list) : proof = 
  ArithContra (prepArithArgs ~foreq:false args)


let optimArithProof (prf: proof) = 
  match prf with
    ArithEq (e1, e2, fact, args) -> 
      ArithEq(e1, e2, fact, prepArithArgs ~foreq:true args)
  | ArithGeq (e1, e2, fact, args) -> 
      ArithGeq(e1, e2, fact, prepArithArgs ~foreq:false args)
  | _ -> prf

(** Create an optimize the proof of not_FI for equalities and disequalities **)
let rec mkEqNeq_FI (e1: exp) (e2 : exp) 
                   (peq: proof) (pneq: proof) : proof = 
  (* Clean up the proof of disequality *)
  match pneq with 
    G("neqtr", [ P e1e1'; E e1'; P e1'_e2'; E e2'; P e2'e2 ]) -> 
      (* Make a new proof of equality *)
      mkEqNeq_FI e1' e2' 
        (mkArithEq e1' e2' [ I 1;  P peq; I (-1); 
                             P e1e1'; I (-1); P e2'e2 ]) e1'_e2'

  | _ -> 
      G("not_FI", [ P peq; P pneq ])
        

 (** Make a proof by contradiction *)
let mkContraProof (a: pred)
                  (h: U.symbol)
                  (pffalse: proof) (* [H:pf (not A)] pf false *)
    : proof  (* pf A *) =
(*
  ignore (E.log "mkContraProof %s:%a is false: %a\n"
            (U.symbolName h)
            d_pred a
            d_prf pffalse);
*)
  (* A few obvious optimizations. Those optimizations that do not reduce the 
   * number of relevant hypotheses are done only if !checkProofs (on the 
   * assumptions that this is the only case when you are willing to save/see 
   * the proofs) *)
  if !checkProofs then begin 
    (* Optimizations that do not change relevant proofs *)
    match a, pffalse with 
      a,
      G("not_FI", [ P p1; P (Hyp h')]) when h == h' -> p1
          
    | a,
        G("not_FI", [ P (Hyp h'); P p2]) when h == h' -> p2
            
            (** Optimization of inequality proofs *)
    | PNez (BinOp(((SetGe|SetLt|SetLe|SetGt) as bop), e1, e2)),
          ArithContra prfl -> begin
            (* Look in the list prfl for h (along with the coversion to 
             * SetGe), and extract the factor along with the rest of the list *)
            let rec extractFactor (prfl: parg list) : int option * parg list = 
              match prfl with 
                [] -> None, []
              | I f' :: P p' :: rest -> begin
                  match bop, p' with 
                    SetGe, G("ltgeq", [ P (Hyp h') ]) when h == h' -> 
                      Some f', rest
                        
                  | SetLt, Hyp h' when h == h' -> 
                      Some f', rest
                        
                  | SetLe, G("gtgeq", [ P (Hyp h') ]) when h == h' -> 
                      Some f', rest
                        
                  | SetGt, G("leqgeq", [ P (Hyp h') ]) when h == h' -> 
                      Some f', rest
                        
                  | _ -> let f'', rest'' = extractFactor rest in 
                    if rest'' != rest then 
                      f'', I f' :: P p' :: rest''
                    else
                      f'', prfl
              end
              | _ -> assert false
            in
            match extractFactor prfl with
              None, _ -> Contra (a, h, pffalse)
            | Some f, prfl' -> begin
                match bop with
                  SetGe -> mkArithGeq e1 e2 ~fact:f prfl'
                | SetLt -> G("geqlt", [ P (mkArithGeq e2 (plus e1 one)
                                             ~fact:f prfl')])
                | SetLe -> G("geqleq", [ P (mkArithGeq e2 e1
                                              ~fact:f prfl')])
                | SetGt -> G("geqgt", [ P (mkArithGeq e1 (plus e2 one)
                                             ~fact:f prfl')])
                | _ -> assert false
            end
          end  
              
    | _, _ -> Contra (a, h, pffalse)
  end else
    Contra (a, h, pffalse)

(* We want to allow the registration of lemmas, predicate definitions and 
 * expressions definitions at the top level. We take a snapshot when we call 
 * init the first time. *)
let initThunk: (unit -> unit) option ref = ref None

let snapshot () : unit -> unit = 
  U.runThunks [ U.restoreIntHash allLemmas;
                U.restoreHash allPDefs;
                U.restoreHash allEDefs ]



let init () = 
  match !initThunk with 
    None -> initThunk := Some (snapshot ())
  | Some t -> t ()

