(** The representation and checking of proofs *)
module Bitmap : 
    sig
      type t
      val make: int -> t
      val size: t -> int
      val setTo: t -> int -> bool -> unit
      val testAndSetTo: t -> int -> bool -> bool
      val test: t -> int -> bool
      val union_except: t -> t -> t -> bool
      val iter: (int -> unit) -> t -> unit
    end
 
(* This module defines the language of proofs *)
val checkProofs: bool ref (** Whether to check the proofs *)

val printMagicProofWarning: bool ref
(** Whether to print warnings for Magic proofs *)


(* In order to maintain a compact representation of proofs we represent a 
 * proof node as the name of the axiom along with a list of arguments. An 
 * exception is made for the hypothetical and parametric proof nodes. *)
type parg = 
    I of int
  | S of string
  | E of Logic.exp
  | P of proof
  | F of Logic.pred

and relset = 
    { relfirst: int;                      (** The hypothesis that has index 0 
                                           * in the set *)
      mutable relcomputed: bool;          (** Whether this set is filled *)
      relbm: Bitmap.t
    } 

and proof = 
    Hyp of Util.symbol                    (* A hypothesis, 
                                           introduced with Impi, or Contra *) 

  | FreeHyps of relset * proof            (** A proof with memoized set of 
                                           * free hypotheses. The first 
                                           * argument says if we have 
                                           * computed the set already. *)

  | Truei

  | Falsei of proof * proof              (** A proof of false, given a proof 
                                          * of a predicate and that of its 
                                          * negation *)

    (** False elimination, if we have a proof of false *)

  | Andi of proof list
    (** Proves And[p1; p2; ... ] *)

  | Ande of int * proof
	(** [Ande (i,pf)] is a proof of p when
	   pf is a proof of [And ps], where [p] is the ith element of [ps] *)

  | Ori of Logic.pred list * int * proof

  | Ore of Logic.pred * Logic.pred list * proof * (Util.symbol * proof) list
        (** [Ore(P, PDISJ, D, DL)] is a proof of P, if D is a proof of a 
         * disjunction PDISJ, and DL is a list with hypothetical proofs of P, 
         * one for each disjunct *)        

  | Alli of Logic.var * proof

  | Alle of  Logic.exp * proof
        (** [Alle(e, d)] is a proof of [e/v]P when d is a proof of All(v,p) *)

  | ExistE of proof * Logic.var * Util.symbol * proof
        (** [Existe(d1,v,h,d2)] is a proof of q, if d is a proof of 
         * Exist(x,p), v is a fresh variable, h is the fresh hypothesis name 
         * for P[v/x], and d2 is a proof of q, with hypothesis h. *)

  | Impi of Util.symbol * Logic.pred * proof (* bind var to a hypothetical 
                                            * proof of the pred. The last 
                                            * argument is a proof of the 
                                            * right hand side, possibly 
                                            * refering to var *)

  | Impe of proof * proof
    (** Impe[pf1, pf2] when pf1 : P1 => P2 and pf2: P1 *)

  
  | AnnotE of proof
    (** AnnotE(d) is a proof of Q if d is a proof of Pannot(a,Q) *)

  | AnnotI of Logic.pannot * proof
    (** AnnotI(a,d) is a proof of Pannot(a, p) when d is a proof of p *)

  | Cut of proof * Util.symbol * proof
                                        (** [Cut(prf1, x, prf2)], when 
                                         * [prf1:P1], [prf2: [x:P1] P2], is a 
                                         * proof of P2 **)

                                        (* A proof of the given predicate, 
                                         * using a proof of false, based 
                                         * on the named hypothesis that 
                                         * "negatePred" holds *)
  | Contra of Logic.pred * Util.symbol * proof
    
  | G of string * parg list             (* This is a generic proof 
                                         * constructor *)

  | Refl of Logic.exp 
  | Symm of proof
    (** Symm p proves e1=e2 if p:e2=e1, or e1<>e2 if p:e2<>e1 *)

  | Trans of Logic.exp * proof * proof
    (** Trans(et, p1t, p2t) proves e1=e2 if p1t:e1=et, pt2:et=e2 *)

  | Cong of Logic.exp * Logic.exp * proof list
  (** Cong \[pf\] proves Unop e1 = Unop e2 if pf proves e1 = e2;
      Cong \[pf1; pf2\] proves Binop (a1,b1) = Binop (a2,b2) if
        pf1 proves a1=a2 and pf2 proves b1=b2;
      Cong pfs proves Fun(f,as) = Fun(f,bs) if
        pfs.(n) proves as.(n) = bs.(n) 
     
      Note that each subproof of equality is really a proof 
      of (Var 0)=(Var 1) in the context in which (Var 0) and (Var 1)
      are replaced with the two things in their appropriate contexts.
   *)

  | PCong of proof * proof list
  (** If p |- PNez(e) and q |- e = e', then PCong (p, q) |- PNez(e') *)
  (** If p |- PFun(s,[e1,...,en]) and q1 |- e1 = e1', ..., qn |- en = en', then PCong (p, [q1,...qn]) |- PFun(s,[e1',...,en']) *)

  | WithLemma of Util.symbol (* The name of the lemma *)
                * Logic.exp list (* The expressions to instantiate the 
                                  * universally quantified variables *)
                * proof list (* The proofs of the hypotheses *)

  | UnrollPDef of Logic.pred * proof (* Prove a predicate of the form 
                                * "PFun(n,args)" by replacing it with its 
                                * definition and providing a proof for the 
                                * resulting expansion *)

  | UnrollEDef of Logic.pred * string * proof
      (* Prove a predicate that uses an abbreviation expression constructor.
       * The given predicate should be the goal without the substitutions made.
       *)
  | RollEDef of Logic.pred * string * proof
      (* Like UnrollEDef, but proves the substituted version of the predicate *)

  | ArithEq of Logic.exp * Logic.exp * int * parg list
        (** [ArithEq(e,e',n,rest)] proves [e=e']. [n] is a factor >= 1. [rest] 
         * is a weighted list of proofs of equalities. Each factor is 
         * followed by its proof. Let fi be the ith factor and ei = ei' be 
         * the ith equality. Check that n*(e - e') - fi*(ei - ei') reduces 
         * using the rules of arithmetic to 0 *)

  | ArithNe of Logic.exp * Logic.exp * int * parg list
        (** [ArithNe(e,e',n,rest)] proves [e<>e']. [n] is a factor >= 1. [rest] 
         * is a weighted list of proofs of equalities. Each factor is 
         * followed by its proof. Let fi be the ith factor and ei = ei' be 
         * the ith equality. Check that n*(e - e') - fi*(ei - ei') reduces 
         * using the rules of arithmetic to something other than 0 *)

  | ArithGeq of Logic.exp * Logic.exp * int * parg list
        (** [ArithGeq(e,e',n,rest)] proves [e>=e']. [n] is a factor >= 1. The 
         * last argument is a weighted list of proofs of >= inequalities 
         * (with positive factors), or equalities. Each factor is followed by 
         * its proof. Let fi be the ith factor and ei >= ei' be the ith 
         * inequality. Check that n*(e - e') - fi*(ei - ei') reduces using 
         * the rules of arithmetic to a constant that is > - n *)


  | ArithContra of parg list
        (** Is a proof of False, when parg is a weighted list of proofs of >= 
         * inequalities (with positive factors), or equalities. Each factor 
         * is followed by its proof. Let fi be the ith factor and ei >= ei' 
         * be the ith equality or inequality. Let n be the sum of fi*(ei - 
         * ei'). It must be the case that n < 0 *)

  | ArithNeq of int * Logic.exp * Logic.exp   
    (** Check that e <> e' by verifiying that e - e' mod32 = i (i <> 0) using 
     * the rules of algebra *)

  | ArithApart of int * Logic.exp * Logic.exp 
     (** ArithApart(r, e1, e2) means that e1 - e2 = r and (-4 >= r or r >= 
      * 4).*)

  | ArithGt of int * Logic.exp * Logic.exp  
        (** Check that e >= e', by verifying that e - e' = i and i > 0. !!! 
         * This is unsound in presence of overflow. We'll have to clean this 
         * up. *)

  | ArithLt of proof
	(** If p |- e >= e', then ArithLt(p) |- e' < e *)
  | ArithLeq of proof
	(** If p |- e > e', then ArithLt(p) |- e' <= e *)

  | SelUpd0 of  Logic.exp * Logic.exp * proof
        (* (SelUpd0 V M D) proves that (= (sel A (upd A' V M)) V) provided 
         * that D proves A = A' *)

  | SelUpd1 of Logic.exp * Logic.exp * proof
        (* (SelUpd1 V M D) proves that (= (sel A (upd A' V M)) (sel A M)) 
           provided that D proves A <> A' *)

  | SelUpd of Logic.exp * Logic.exp * Logic.exp * Logic.exp
        (* [SelUpd(SA,UA,UV,UM)] proves ((SA = UA) AND (sel SA (upd UA UV UM)) 
         * = UV) OR ((SA <> UA) AND (sel SA (upd UA UV UM)) = (sel SA UM)) *)

  | Magic of string * Logic.pred 
    (** These are proofs that are allowed to go through the proof checker, 
     * with a warning. Pass a string to identify this magic *)

  | MagicLemma of lemma
    (** A magic proof of a lemma *)

  | CoqLemma of string list
    (** Proof by Coq implementation, with variable types *)

  | Unimplemented of string     (* We'll expand the proof language as needed 
                                 * in the future. Use Unimplemented until 
                                 * then *)


and lemma = 
    { lname: Util.symbol; (* Lemma name *)
      lvars: string list; (* A list of variables that are universally 
                           * quantified over. These are bound in the rest of 
                           * the lemma, using UVar *)
      lhyps: Logic.pred list; (* A list of hypotheses. *)
      lconc: Logic.pred; (* A conclusion *)
    } 


   (* Infer the "type" of a proof. First argument is a list of variables and 
    * the second is a mapping of hypothesis variables to the predicates they 
    * prove. Raise ProofCheck exception if not a valid proof.  *)
val infer : Logic.var list -> 
            (Util.symbol * Logic.pred) list -> proof -> Logic.pred

val check : Logic.var list -> 
            (Util.symbol * Logic.pred) list -> proof -> Logic.pred -> unit



(**************** Lemmas ************)

type coq_mode =
    CoqTerse
  | CoqVerbose

(** Load a Coq process for checking lemma proofs *)
val openCoq:
    string list (* Directories to search for implementation files *)
  -> string     (* Name of implementing module *)
  -> coq_mode
  -> unit

(** Register a lemma, along with its proof *)
val registerLemma: 
    lemma -> 
    proof -> (* This is the proof that ALL lvars. (AND lhyps) => lconc *)
    unit

val registeredLemmas: unit -> lemma list

val findRegisteredLemma: Util.symbol -> lemma

(*************** Definitions **********)
(** Expression definitions. The name of the defined constructor (use a 
  * lowercase name), the list of formals, and the body, referring to the 
  * formals using deBruijn indices. *)
type edef = string * string list * Logic.exp 

(* Predicate definitions. The name of the defined constructor (use an 
 * uppercase initial for the name), the list of formals, and the body, 
 * referring to the formals using deBruijn indices. *)
type pdef = string * string list * Logic.pred 


(** Register an expression definition *)
val registerEDef: edef -> unit

(** Register a predicate definition *)
val registerPDef: pdef -> unit




 (** Make a proof by contradiction *)
val mkContraProof: Logic.pred (* A *) -> 
                   Util.symbol (* H *) -> 
                   proof (* [H:pf (not A)] pf false *)
                   -> proof (* pf A *)

(** Make a proof of false from a proof of an equality and a proof of a 
 * disequality *)
val mkEqNeq_FI: Logic.exp -> Logic.exp -> proof -> proof -> proof

val d_prf:  unit -> proof -> Pretty.doc
val dpa: parg -> Pretty.doc


(** Use these function to make ArithEq and ArithGeq proofs, if you want to 
 * optimize the hypotheses (to avoid duplicate hypotheses) *)
val mkArithEq: Logic.exp -> Logic.exp -> ?fact:int -> parg list -> proof
val mkArithGeq: Logic.exp -> Logic.exp -> ?fact:int -> parg list -> proof

val mkArithContra: parg list -> proof

val optimArithProof: proof -> proof

(** We can install a new printer for the Proof module *)
val printer: Logic.logicPrinter ref

(** These are exported temporarily *)
val allEDefs: (string, edef) Hashtbl.t
val allPDefs: (string, pdef) Hashtbl.t
val allLemmas: lemma Inthash.t (* Indexed by symbols *)

(** Initialize the proof module *)
val init: unit -> unit


(** Take a snapshot of the state of the proof module. The only state is the 
 * set of registered lemmas and constructors *)
val snapshot: unit -> unit -> unit

val findLemma: lemma -> unit

val registeredLemmas: unit -> lemma list

val makeExistVar: string -> Logic.exp



(** Scan a proof and watch for free and bound hypotheses. pushBound is 
 * invoked before scanning the scope of a bound hypothesis. popBound is 
 * called after the body was scanned. If this function 
 * returns "false" then the body is not scanned, and the matching popBound is 
 * not called. *)
val scanHyps: pushBound:(Util.symbol -> bool) ->
              popBound:(Util.symbol -> unit) ->
              seenHyp:(Util.symbol -> unit) ->
              seenFreeHyps:(relset -> proof -> bool) ->  
              proof -> unit



val getFreeHyp: unit Inthash.t -> proof -> unit
