module E = Errormsg
module L = Logic
module P = Proof
module S = Satproc
module H = Hashtbl
open Pretty
exception Failure

(* George's data structures *)
let ccid = "FAST_RAND"

let proveIt _ = E.s (E.unimp "Rand.proveIt")
type undoelem = SnapShot (* This is the last snapshot *)
let undoList: undoelem list ref = ref []
(* ---------- *)


(* Tunable Parameters *)
let debug = ref false

(* How many times you want to repeat the experiment and then take the average *)
let repetitions = 30

(* Set this true when performing measurements - It won't print a thing *)
(* However, to know the answer for the problem, set this variable to false *)
let timing = true

(* The following parameters control the error probability *)
let randPrime = 4093
let nrPoints = 40

(* Set this true to allow for cheap adjusts when a new term occurs in an axiom *)
(* Set this false, if the program doesn't seem to work correctly *)
let optimization = true

(* Total number of random values used by the program *)
let nrRandom = 8192

(* Maximum arity of any uniterpreted function symbol *)
let maxArity = 64
(* Maximum number of terms that occur in an equation *)
let maxTerms = 256
(* ------------- *)



let randBound = randPrime - 1
let nrAdjusts = ref (0) 
(* Number of optimized cheap adjusts performed *)
let nrAdjustNewTerms = ref (0)
let nrAxioms = ref (0)
let failure = ref(false)

(* Data structures *)
(* This is to represent an affine expression *) 
type expr = int * (int * term) list (* constant * (coefficient * term) list *)
and term = Var of string
  | Func of expr (* we convert all uninterpreted function symbols (UFS) to one unary UFS *)

type value = int array
(* to store the values of variables *)
let state_variables : (string, value) Hashtbl.t = Hashtbl.create 53 
(* to store the values of the unary uninterpreted function symbol *)
let state_functions : (value, value) Hashtbl.t ref = ref (Hashtbl.create 253)

(* to store weights for conversion to one unary UFS *)
let randWeights_argument = 
  if (not !debug) then Random.self_init (); 
  Array.init maxArity (fun i -> let t () = Random.int randPrime in t ()) 
let randWeight_functionSymbol :(string, int) Hashtbl.t = Hashtbl.create 53
(* -------------- *)



(* Modular Arithmetic Functions BEGIN *)
let pmod x = 
  let t = x mod randPrime
  in 
  if (t < 0) then t + randPrime else t

let currentRandCounter = ref (-1) 

let randomElements = 
  if (not !debug) then Random.self_init ();
  Array.init nrRandom (fun i -> let t () = Random.int randPrime in t ())
 
let extended_euclidean_algorithm a1 b1 = 
  let sign i = if (i > 0) then 1 else (-1) in
  let a = ref (abs a1) in
  let b = ref (abs b1) in
  let x0 = ref (sign a1) in
  let x1 = ref (0) in
  let y0 = ref (0) in
  let y1 = ref (sign b1) in
  let x = ref (0) in
  let y = ref (0) in
  while (not (!b = 0)) do
    let q = !a / !b in
    let r = !a mod !b in
    a := !b; b := r;
    x := !x0 - (!x1 * q); x0 := !x1; x1 := !x;
    y := !y0 - (!y1 * q); y0 := !y1; y1 := !y;
  done;
  (!x0,!y0)

let inverses = Array.init randPrime (fun i -> pmod (fst (extended_euclidean_algorithm i randPrime)))

let add e1 e2 = pmod (e1 + e2)
let sub e1 e2 = if (e1 >= e2) then e1 - e2 else e1 - e2 + randPrime
let mult e1 e2 = pmod (e1 * e2)  
let div e1 e2 = mult e1 (inverses.(e2)) 
let join e1 e2 r = if (e1 = e2) then e1 else pmod ((e1 * r) + (e2 * (1-r)))

let randElt () = 
  incr currentRandCounter;
  if (!currentRandCounter >= nrRandom) then begin
    ignore (E.log "RAND: INCREASE nrRandom\n");
    assert(false);
  end;
  randomElements.(!currentRandCounter)


let docElt () e = dprintf "%d" e 
let eq e1 e2 = ((pmod e1) = (pmod e2)) 
let lt e1 e2 = ((pmod e1) < (pmod e2)) 
let gt e1 e2 = ((pmod e1) > (pmod e2)) 
let leq e1 e2 = ((pmod e1) <= (pmod e2)) 
let geq e1 e2 = ((pmod e1) >= (pmod e2)) 
(* ---------------- *)      


(* some useful functions on values *)
let docValue () (v: value) : Pretty.doc =
  let vlist = ref ([]) in
  for i = nrPoints - 1 downto !nrAdjusts do
    vlist := v.(i)::!vlist;
  done;
  dprintf "[%a]" (docList (fun n -> dprintf "%d" n)) (!vlist)

let valueI i = Array.create nrPoints (pmod i)

let valueZero = Array.create nrPoints 0

let is_equal_value v1 v2 =
  let result = ref (true) in
  let i = ref (!nrAdjusts) in
  while (!i < nrPoints) do
    if (v1.(!i) != v2.(!i)) then begin
      i := nrPoints;
      result := false;
    end
    else incr i;
  done;
  !result

let is_constant_value v =
  let c = v.(!nrAdjusts) in
  let result = ref (true) in
  let i = ref (!nrAdjusts) in
  while (!i < nrPoints) do 
    if (c != v.(!i)) then begin
      i := nrPoints;
      result := false;
    end
    else incr i;
  done;
  !result

let randValue () = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- randElt () done;
  result  


let add_value v1 v2 = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- add v1.(i) v2.(i) done;
  result

let sub_value v1 v2 = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- sub v1.(i) v2.(i) done;
  result


let mult_value v1 v2 = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- mult v1.(i) v2.(i) done;
  result

let div_value v1 v2 = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- div v1.(i) v2.(i) done;
  result

let join_value v1 v2 r = 
  let result = Array.create nrPoints 0 in 
  for i = !nrAdjusts to nrPoints - 1 do result.(i) <- join v1.(i) v2.(i) r.(i) done;
  result

(* -------------- *)


let pendingDiffs = ref ([]) (* for collecting the differences that need to be adjusted because of hash-table clashes *)

(* We collect all the disEqualities encountered in assumptions to check their validity at the end 
 * when proving the goal *)
let disEqualities = ref ([]) 
let axioms = ref ([])



(****************************************************************************************)
(* The following function is called with an argument diff \neq [0,...0] *)
(* It computes the adjustment factors such that diff goes to [0,...,0] after adjust_values *)
(* adjusting factors is the tuple (i1,i2,r,[r0, ..., rn]) such that
   (i) diff.(i1) \neq diff.(i2)  
   (ii) Let p = (r * diff.(i1)) + ((1-r) * diff.(i2)). 
        r is such that p \neq diff.(i) 
   (ii) ri = p/(p-diff.(i))
*)

let compute_adjustment_factors diff =
  let currentFuncDebug = false & !debug in
  let af = Array.create nrPoints 0 in   (* the adjusting factors array *)

  (* we first find indices i1 and i2 such that diff.(i1) \neq diff.(i2) *)
  let i1 = !nrAdjusts in 
  let tmp = ref (!nrAdjusts + 1) in
  let found = ref false in 
  while (not !found) do
    if (diff.(!tmp) = diff.(i1)) then incr tmp
    else found := true;
  done;
  let i2 = !tmp in

  (* we now find an r such that p = (r * diff.(i1)) + ((1-r) * diff.(i2)) \neq diff.(i) *)
  let r = ref (randElt ()) in 
  let p = ref (0) in  
  let found = ref (false) in
  while (not !found) do 
    (* we guess r and then test whether it satisfies the required property *)
    r := randElt (); (* our new guess for r is another random number *)
    found := true;
    p := join diff.(i1) diff.(i2) !r; 
    let i = ref (!nrAdjusts) in
    while (!found) & (!i < nrPoints) do
       if (!p = diff.(!i)) then found := false
       else incr i;
    done;
  done;
 
  (* Now, we compute ri *)
  for i = 0 to nrPoints - 1 do    
    let ri = div !p (sub !p diff.(i)) in 
    af.(i) <- ri; 
  done;

  if (currentFuncDebug) then begin 
    ignore (E.log "RAND: adjustment_factors(%a) := i1 = %d, i2 = %d, r = %a, af = %a\n" 
	      docValue diff i1 i2 docElt !r 
              (docArray 
                 (fun i r -> dprintf "(%d,%a)" i docElt r)) af);
  end;

  (i1,i2,!r,af)

let dumpState () = 
  if (!debug) then begin
    ignore (E.log "RAND: State is:\n");
      Hashtbl.iter (fun v1 v2 ->
	ignore (E.log "RAND: F(%a) = %a\n" docValue v1 docValue v2)) !state_functions;
    Hashtbl.iter (fun x v ->
      ignore (E.log "RAND: %s = %a\n" x docValue v)) state_variables;
  end


(* The following are used to keep track of any new term with non zero coefficient that occurs in an 
 * axiom to facilitate cheap adjust operation *)

(* For storing new random values assigned to Function terms so as to 
 * facilitate cheap adjustment. Slightly non-trivial. For example, consider 
 * the axiom x = f(x) and suppose that x occurs for the first time. We start 
 * by giving the value r1 to x and r2 to F(r1). Then to perform cheap adjust, 
 * we set x to r2, but also create an entry F(r2) = r2 and delete the entry 
 * F(r1) = r2 *)
let newValues = Array.create maxTerms (valueZero,valueZero) 
let newValueCount = ref (0)
(* For collecting any new variables and its coefficient *)
let newVTerms :(string,int) Hashtbl.t = Hashtbl.create 53
(* For collecting any new function term and its coefficient *)
let newFTerms :(value,int) Hashtbl.t = Hashtbl.create 53
(* ------------------ *)

(* ROUTINES TO EVALUATE EXPR *)

let get_variable (s: string): value = 
  let newValue = 
    try
      Hashtbl.find state_variables s
    with Not_found -> 
      let newValue = randValue () in
      Hashtbl.add state_variables s newValue;
      newValue
  in
  newValue


let get_functionTerm (v: value): value =
  let newValue = 
    try  
      Hashtbl.find !state_functions v 
    with Not_found ->
      let newValue = randValue () in
      Hashtbl.add !state_functions v newValue;
      newValue
  in
  if (optimization) then begin
    newValues.(!newValueCount) <- (v,newValue);
    incr newValueCount;
  end;
  newValue

let rec evalExpr (i,ctl) = 
  List.fold_left (fun temp (c,t) -> 
    add_value temp (mult_value (valueI c) (evalTerm (c,t)))) (valueI i) ctl

and evalTerm (c,t) =
  match t with
    Var x -> get_variable x
  | Func e -> get_functionTerm (evalExpr e)


(* _o stands for "optimizaton". "_o" routines are used to track any new term 
 * with a non-zero coefficient and then do a cheap adjust later *)

let get_variable_o (s: string) (c: int): value = 
  let newValue = 
    try
      begin try
	let old = Hashtbl.find newVTerms s in
	let c' = old + c in
	if (c' != 0) then Hashtbl.replace newVTerms s c'
	else Hashtbl.remove newVTerms s
      with Not_found -> () 
      end;
      Hashtbl.find state_variables s
    with Not_found -> 
      let newValue = randValue () in
      Hashtbl.add state_variables s newValue;
      Hashtbl.add newVTerms s c;
      newValue
  in
  newValue

let get_functionTerm_o (v: value) (c: int): value =
  let newValue = 
    try  
      begin try
	let old = Hashtbl.find newFTerms v in
	let c' = old + c in
	if (c' != 0) then Hashtbl.replace newFTerms v c'
	else Hashtbl.remove newFTerms v
      with Not_found -> () 
      end;
      Hashtbl.find !state_functions v; 
    with Not_found ->
      let newValue = randValue () in
      Hashtbl.add !state_functions v newValue;
      Hashtbl.add newFTerms v c;
      newValue
  in
  newValues.(!newValueCount) <- (v,newValue);
  incr newValueCount;
  newValue

let evalExpr_o (i,ctl) = 
  List.fold_left (fun temp (c,t) -> 
    let v = match t with
      Var x -> get_variable_o x c
    | Func e -> get_functionTerm_o (evalExpr e) c
    in
    add_value temp (mult_value (valueI c) v)) (valueI i) ctl
(* ------------------------ *)


(* Not so neat routine to print Expr neatly *)
let rec docExpr () (i,ctl)  = 
  match ctl with
    [] -> dprintf "%d" i
  | (i1,t1)::ctl' ->
      let docTerms () ctl = (docList ~sep:(text "") (fun (i,t) -> 
	if (i > 0) then dprintf " + %a" docCTerm (i,t)
	else dprintf " - %a" docCTerm (i,t))) () ctl in
      let terms = dprintf "%a%a" docCTerm (i1,t1) docTerms ctl' in
      if (i = 0) then
	if (i1 < 0) then dprintf "- %a" insert terms
	else dprintf "%a" insert terms
      else
	if (i1 < 0) then dprintf "%d - %a" i insert terms
	else dprintf "%d + %a" i insert terms
	    
and docCTerm () (i,t) = 
  let ai = abs i in
  if (ai != 1) then dprintf "%d %a" ai docTerm t 
  else dprintf "%a" docTerm t
  
and docTerm () t = 
    match t with
      Var s -> dprintf "%s" s
    | Func e -> dprintf "F(%a)" docExpr e
(* -----------------------*)    


(* Routine to Covert exp to expr *) 
let rec exp2expr (e: L.exp): expr = 
  match e with
    L.NVar v -> (0,[(1, Var v)])
  | L.Int i -> (L.to_int i,[])
  | L.BinOp(L.Plus, e1,e2) ->
      let (i1,t1) = exp2expr e1 in
      let (i2,t2) = exp2expr  e2 in
      (i1+i2, List.append t1 t2)
  | L.BinOp(L.Minus, e1,e2) -> 
      let (i1,t1) = exp2expr e1 in
      let (i2,t2) = exp2expr e2 in
      let t2' = List.map (fun (i,t) -> (-i,t)) t2 in
      (i1-i2, List.append t1 t2')
  | L.BinOp(L.Times, e1,e2) ->
      let (i1,t1) = exp2expr e1 in
      let (i2,t2) = exp2expr  e2 in
      let t1' = List.map (fun (c,t) -> (i2*c,t)) t1 in
      let t1'' = List.filter (fun (c,t) -> (c != 0)) t1' in
      let t2' = List.map (fun (c,t) -> (i1*c,t)) t2 in
      let t2'' = List.filter (fun (c,t) -> (c != 0)) t2' in
      (i1*i2, List.append t1'' t2'') 
      (* Assuming linearity, this is correct because linearity implies that at least one of t1 or t2 is empty *)
  | L.Fun (f,el) -> 
      begin match el with
	[] -> (0,[(1, Var f)]) (* convert nullary function symbols into variables *)
      |	_ -> (* we transform all function symbols into one unary function symbol *)
	  let vl = List.map exp2expr el in
	  let weight = try
	    Hashtbl.find randWeight_functionSymbol f 
	  with Not_found ->
	    let r = randElt () in
	    Hashtbl.add randWeight_functionSymbol f r;
	    r
	  in
	  let j = ref (0) in
	  let ti = ref (weight) in
	  let tt = ref ([]) in
	  List.iter (fun (i,t) ->
	    let weight' = randWeights_argument.(!j) in
	    ti := pmod (!ti + weight'*i);
	    tt := List.append !tt (List.map (fun (i1,t1) -> (pmod (i1*weight'), t1)) t);
	    incr j ) vl;
	  (0,[(1, Func (!ti,!tt))])
      end
      | _ -> E.s (E.unimp "evalExp: Unimplemented")
(* --------------------------------- *)


(* The next functions are for performing cheap adjustment *)

(* This one checks if cheap adjustment is possible, and also sets the new term *)	    
let adjustNewTerm diff = 
  let currentFuncDebug = false & !debug in 
  let found = ref (false) in
  Hashtbl.iter (fun x c -> 
    if (!found = false) then begin
      let oldValue = Hashtbl.find state_variables x in
      let newValue = div_value (sub_value diff (mult_value oldValue (valueI c))) (valueI (-c)) in
      if (currentFuncDebug) then 
	ignore (eprintf "NewTerm = %s, oldValue = %a, newValue = %a, c = %d\n" 
		  x docValue oldValue docValue newValue c);
      Hashtbl.replace state_variables x newValue;
      found := true;
      incr nrAdjustNewTerms;
    end ) newVTerms;
  if (!found = false) then begin
   Hashtbl.iter (fun v c -> 
   if (!found = false) then begin
   let oldValue = Hashtbl.find !state_functions v in
	let newValue = div_value (sub_value diff (mult_value oldValue (valueI c))) (valueI (-c)) in
	Hashtbl.replace !state_functions v newValue;
	if (currentFuncDebug) then 
	  ignore (eprintf "NewTerm = F(%a), oldValue = %a, newValue = %a, c = %d\n" 
		    docValue v docValue oldValue docValue newValue c);
	found := true;
	incr nrAdjustNewTerms;
      end ) newFTerms;
  end;
  !found

let adjust_functionTerm (v: value) = 
  let result = 
    try
      Hashtbl.find !state_functions v
    with Not_found -> 
      let (v1,v2) = newValues.(!newValueCount) in
      Hashtbl.remove !state_functions v1;
      Hashtbl.add !state_functions v v2;
      v2
  in 
  incr newValueCount;
  result

(* This is called if adjustNewTerm succeeds to adjust other dependent function terms accordingly *)
let adjust_expr (i,ctl): unit = 
  let rec evalExpr_a (i,ctl) = 
    List.fold_left (fun temp (c,t) -> 
      let v = match t with
	Var x -> get_variable x
      | Func e -> adjust_functionTerm (evalExpr_a e)
      in
      add_value temp (mult_value (valueI c) v)) (valueI i) ctl
  in
  List.iter (fun (c,t) -> 
    match t with 
      Var x -> ()
    | Func e -> let v = adjust_functionTerm (evalExpr_a e) in ()) ctl


(* This is essentially for debugging purposes. Otherwise not used anywhere *)
let verify_axioms () = 
  List.iter (fun e -> 
    let v = evalExpr e in
    if (not (is_equal_value v valueZero)) then begin 
      ignore (eprintf "RAND: AXIOM %a=0 NOT SATISFIED\n" docExpr e);
      assert false; end) !axioms
   

(* The Real Adjustment Routines *)
let adjust_values2 (diff: int array) : unit =
  let currentFuncDebug = false & !debug in
  if (!nrAdjusts >=  nrPoints-4) then begin
    ignore (E.log "RAND: # Adjusts = %d, # Points = %d\n" !nrAdjusts nrPoints);
    ignore (E.log "RAND: INCREASE nrPoints\n");
    assert(false);
  end;
  if (currentFuncDebug) then begin
    ignore (E.log "RAND: Before adjustment, the state is:\n");
    dumpState ();
  end;
  
  if (not (is_equal_value diff valueZero)) then begin (* adjustment is needed such that diff' becomes zero *)
    if (is_constant_value diff) then begin (* adjustment cannot be done *)
      if (not timing) then ignore (E.log "RAND: Axioms are Inconsistent\n");
      raise Failure;
    end;
    let (i1,i2,r,af) = compute_adjustment_factors diff in 
    
    (* cache the adjusted values to avoid spending effort in readjusting the same value again *)
    let cache_adjustValues = Hashtbl.create 53 in

    (* a function to adjust one value *)
    let adjust_value v = 
      try
	Hashtbl.find cache_adjustValues v
      with Not_found -> 
	let v' = Array.create nrPoints (0) in
	let p = join v.(i1) v.(i2) r in 
	v'.(i1+1) <- join v.(i2) p af.(i2);
	for i = (i1+1) to (i2-1) do v'.(i+1) <- join v.(i) p af.(i) done;
	for i = (i2+1) to (nrPoints-1) do v'.(i) <- join v.(i) p af.(i) done;
	Hashtbl.add cache_adjustValues v v';
	v'
    in

    (* Adjust all the values in pendingDiffs *)
    pendingDiffs := List.map adjust_value !pendingDiffs;

    (* Adjust all the values in the state_variables *)    
    Hashtbl.iter (fun x v ->
      let v' = adjust_value v in
      Hashtbl.replace state_variables x v') state_variables;

    (* Adjust and re-hash all the values in the state_functions *) 
    let state_functions' = Hashtbl.create 253 in
    Hashtbl.iter (fun (v1) v2 ->
      let v1' = adjust_value v1 in
      let v2' = adjust_value v2 in
      begin try
	let v2o = Hashtbl.find state_functions' (v1') in
	let diff' = sub_value v2o v2' in 
        (* There is a clash in the hash-table and the difference diff' needs to be adjusted *)
	if (currentFuncDebug) then ignore (E.log "RAND: Adding %a to pendingDiffs\n" docValue diff');
	pendingDiffs := diff'::!pendingDiffs;
      with Not_found -> 
      Hashtbl.add state_functions' (v1') v2'
      end) !state_functions;
    state_functions := state_functions';
    incr nrAdjusts;    
    if (currentFuncDebug) then begin
      ignore (E.log "RAND: After adjustment, the state is: \n");
      dumpState ();
      ignore (E.log "RAND: # adjusts performed till now = %d\n" !nrAdjusts)
    end;
  end


let adjust_values (diff: int array) : unit =
  adjust_values2 diff;
  while (!pendingDiffs != []) do
    let diff' = List.hd !pendingDiffs in
    pendingDiffs := List.tl !pendingDiffs;
    adjust_values2 diff';
  done

    
let verify_disEqualities () = 
  List.iter (fun e -> 
    let v = evalExpr e in
    if (is_equal_value v valueZero) then begin
      if (not timing) then 
	ignore (eprintf "RAND: The disequality %a != 0 was found to be Inconsistent at the end\n" docExpr e);
      raise Failure;
    end
    else if (!debug) then 
      ignore (E.log "RAND: The disequality %a != 0 was found to be Consistent at the end\n" docExpr e);
    ) !disEqualities


let verify_goal e = 
  let v = evalExpr e in 
  if (is_equal_value v valueZero) then begin
    if (not timing) then ignore (E.log "RAND: PROVED GOAL\n");
  end
  else begin 
    if (not timing) then ignore (E.log "RAND: UNABLE TO PROVE THE GOAL\n");
    raise Failure;
  end


let initialize_structures () =
  failure := false;
  currentRandCounter := 0;
  nrAdjusts := 0;
  nrAdjustNewTerms := 0;
  Hashtbl.clear state_variables;
  Hashtbl.clear !state_functions

let process_assumptions () =
  List.iter (fun e -> 
    let currentFuncDebug = false & !debug in
    
    if (currentFuncDebug) then begin
      ignore (eprintf "RAND: Processing Axiom: %a = 0\n" docExpr e);
      ignore (eprintf "RAND: State before Processing:\n"); dumpState ();
    end;
    
    (* some global variables to facilitate discovery of any new term in an 
     * axiom for a cheaper adjust optimization *)
    newValueCount := 0;
    Hashtbl.clear newVTerms;
    Hashtbl.clear newFTerms;

    let v = if (optimization) then evalExpr_o e else evalExpr_o e in
    if (is_constant_value v) then begin
      if (is_equal_value v valueZero) then begin 
	if (!debug) then ignore (eprintf "RAND: Axiom %a = 0 is redundant\n" docExpr e)
      end
      else begin
	if (not timing) then ignore (eprintf "RAND: Axiom %a = 0 is inconsistent\n" docExpr e);
	failure := true;
      end
    end
    else begin
      if (optimization) then begin
	newValueCount := 0;
	if (adjustNewTerm v) then adjust_expr e
	else Stats.time "adjust_values" adjust_values v
      end
      else Stats.time "adjust_values" adjust_values v
    end;
    
    if (currentFuncDebug) then begin
      ignore (eprintf "RAND: State after Processing:\n"); dumpState ();
      ignore (eprintf "RAND: Successfully processed the axiom %a = 0\n" docExpr e);
    end ) !axioms


let assume (acc: S.message list) msg = 
  match msg.S.body with
    L.PNez(L.BinOp(L.SetEq, e1, e2)) ->
      let e = L.minus e1 e2 in 
      let e' = exp2expr e in 
      if msg.S.goal then begin
	for i = 1 to repetitions do
	  let task () = 
	    initialize_structures ();
	    try
	      process_assumptions ();
	      verify_disEqualities ();
	      verify_goal e'
	    with Failure -> failure := true;
	  in Stats.time "Total Time after parsing" task ();
	done;
	Stats.print stdout "Timings:\n";
	ignore (eprintf "# Axioms = %d, # Optimized Adjusts = %d, # Adjusts = %d, # Points = %d\n" 
		  !nrAxioms !nrAdjustNewTerms !nrAdjusts nrPoints);
	if (!failure) then raise (S.Contradiction (P.G("inconsistent", [])))
	else raise (S.Proved (P.G("eqid", [P.E e1])));
      end 
      else begin 
	axioms := List.append !axioms [e'];
	incr nrAxioms
      end;
      acc 
  | L.Bool true ->  acc
  | L.Bool false -> 
      if (not timing) then ignore (E.log "RAND: FalseAtom \n"); 
      E.s (E.error "proving false")

  | L.PNez(L.BinOp(L.SetNe, e1, e2)) -> 
      let e = L.minus e1 e2 in
      let e' = exp2expr e in
      disEqualities := (e') :: !disEqualities; (* collect disEqualities *)
      acc
  | _ -> acc
	
let initialize () =
  (* Comment the following initialization of random number generator when debugging *)  
  Random.self_init ();
  undoList := []

let entries = 
  { S.init = initialize;
    S.snapshot = (fun () -> ());
    S.undo = (fun () -> ());
    S.assume = assume;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = ccid;
    S.descr  = "randomized congruence closure";
  } 
