(* FLOATING2_ARITHMETIC: Based on ideas of Richard Fateman of maintaining 
 * both lower bounds and upper bounds *)

module E = Errormsg
module N = Num
open Pretty

let currentFileDebug = false


(* WARNING: The Floating_arithmetic.floating_arithmetic module is better than 
   this module since the former has provably more completeness than this one.
   i.e. if the former module gives a false positive, this module also does 
   but vice versa is not always true.
   Also, this module is not better than the former module w.r.t. probabilistic soundness. 
   NOTE: False Positives arise not from incompleteness of our algorithm but
   because of the small precision associated with floats in Ocaml
*)


let delta = 0.5 (* the maximum error tolerance factor - choice of 0.5 is arbitrary *)
(* For more notes on choice of the value of "delta", read the comments on "delta" 
   used in Floating_arithmetic.floating_arithmetic module. *)

(* If the difference between the lower bound and the upper bound of any
   element becomes greater than "delta", we fail *)



module Floating2_arithmetic: Arithmetic.arithmetic = struct 
  type elt = float * float (* lower bound * upper bound *)

  let docElt () (lb, ub) = dprintf "(%f,%f)" lb ub
  let eq (lb, ub) (lb', ub') = not ((ub < lb') or (ub' < lb)) 

  let lt (lb, ub) (lb', ub') = (ub < lb')
  let gt (lb, ub) (lb', ub') = (lb > ub')
  let leq f1 f2 = (lt f1 f2) or (eq f1 f2)
  let geq f1 f2 = (gt f1 f2) or (eq f1 f2)



  (* Hashing function should be such that equal floats (by the above 
     definition) should have the same hash. Now this is tricky since
     floats are now represented by a range *)
  (* ASSUMPTION: The exponent and the first 3 digits of the significant of lb and ub are same *) 
  (* Basis of Assumption: lb and ub don't differ 'that' much *)
  (* If Assumption Fails: we will loose completeness (not soundness) *)
  (* Detection of Failure of Assumption: Failure of Assumption can be checked by putting an assert *)
  let hashElt (lb, ub) = 
    let (sl,el) = frexp lb in
    let (su,eu) = frexp ub in
    let sl' = truncate (sl *. 1000.0) in
    let su' = truncate (su *. 1000.0) in

    if not ((el = eu) & (sl' = su')) then begin
      ignore (E.log "Floating2_arithmetic: Hashing fails \n");
      assert (false);
    end;
    Hashtbl.hash (el,sl') 
  


  (* generates a random number between -randBound and +randBound, both exclusive *)
  let randElt () randBound =
    let doubleRandBound = randBound * 2 in
    let randInt = (Random.int doubleRandBound) - randBound
    in
    (float randInt, float randInt)
  
  let zero = (float 0, float 0)
  let one = (float 1, float 1)
  let int2elt i = (float i, float i)

  let absolute (lb, ub) = 
    let abs_lb = abs_float lb in
    let abs_ub = abs_float ub in
    if (abs_lb < abs_ub) then (abs_lb, abs_ub)
    else (abs_ub, abs_lb)

  (* ASSUMPTION: The functions "lower_bound" and "upper_bound" work as expected *)
  (* Basis of ASSUMPTION: several hand tests with ocaml interpretor *)

  (* lower_bound f returns the largest f' such that f' < f *)
  let lower_bound f = 
    let (s,e) = frexp f in
    let s' = s -. epsilon_float in
    ldexp s' e
    
  (* upper_bound f returns the smallest f' such that f < f' *)
  let upper_bound f = 
    let (s,e) = frexp f in
    let s' = s +. epsilon_float in
    ldexp s' e

  (* Following function is called after every arithmetic operation to 
     compute lower bound of the "lower-bound result" and upper bound of the 
     "upper-bound result". It is necessary to call this function 
     to account for any rounding performed by Ocaml. 
  *) 

  let bounds lb ub = 
    if (lb > ub) then begin
      ignore (E.log "floating_arithmetic: invariant violated because of some operation \n");
      ignore (E.log "lb = %f, ub = %f \n" lb ub);
      assert (false);
    end;
    if (ub > lb +. delta) then begin
      ignore (E.log "floating_arithmetic: Too much difference, lb = %f, ub = %f \n" lb ub);
      assert (false);
    end;
    let lb' = lower_bound lb in
    let ub' = upper_bound ub in
    if (lb' > ub') then begin
      ignore (E.log "floating_arithmetic: invariant violated because of bounding \n");
      ignore (E.log "lb = %f, ub = %f \n lb' = %f, ub' = %f \n" lb ub lb' ub');
      assert (false);
    end;
    if (ub' > lb' +. delta) then begin
      ignore (E.log "floating_arithmetic: Too much difference, lb=%f, ub=%f, lb' = %f, ub' = %f \n" lb ub lb' ub');
      assert (false);
    end;
    (lb',ub')


  let add (lb1, ub1) (lb2, ub2) = 
    let lb = lb1 +. lb2 in
    let ub = ub1 +. ub2 in
    if currentFileDebug then begin
      ignore (E.log "add(%a, %a) = %a \n" docElt (lb1,ub1) docElt(lb2,ub2) docElt(lb,ub));
    end;
    bounds lb ub


  let sub (lb1, ub1) (lb2, ub2) = 
    let lb = lb1 -. ub2 in
    let ub = ub1 -. lb2 in
    if currentFileDebug then begin
      ignore (E.log "sub(%a, %a) = %a \n" docElt (lb1,ub1) docElt(lb2,ub2) docElt(lb,ub));
    end;
    bounds lb ub


  let mult (lb1, ub1) (lb2, ub2) =
    (* FOLLOWING IS NOT CORRECT: Consider the cases when lb1 and ub1 have different signs 
       let lb = lb1 *. lb2 in
       let ub = ub1 *. ub2 in
       An untried alternative is that when lb1 and ub1 have different signs, force it to be 0
       We however do the following:
     *)
    let allCombinations = [lb1 *. lb2; lb1 *. ub2; ub1 *. lb2; ub1 *. ub2] in
    let sortedCombinations = List.sort compare allCombinations in
    let lb = List.nth sortedCombinations 0 in (* the smallest element *)
    let ub = List.nth sortedCombinations 3 in (* the largest element *)

    if currentFileDebug then begin
      ignore (E.log "mult(%a, %a) = %a \n" docElt (lb1,ub1) docElt(lb2,ub2) docElt(lb,ub));
    end;
    bounds lb ub

  let div (lb1, ub1) (lb2, ub2) = 
   (* FOLLOWING IS NOT CORRECT: Consider the cases when lb1 and ub1 have different signs 
       let lb = lb1 /. lb2 in
       let ub = ub1 /. ub2 in
       We do the following: 
    *)
    let allCombinations = [lb1 /. lb2; lb1 /. ub2; ub1 /. lb2; ub1 /. ub2] in
    let sortedCombinations = List.sort compare allCombinations in
    let lb = List.nth sortedCombinations 0 in (* the smallest element *)
    let ub = List.nth sortedCombinations 3 in (* the largest element *)
    if currentFileDebug then begin
      ignore (E.log "div(%a, %a) = %a \n" docElt (lb1,ub1) docElt(lb2,ub2) docElt(lb,ub));
    end;
    bounds lb ub
      
   let join (lb1, ub1) (lb2, ub2) (lbr, ubr) = 
     bounds lb1 ub1 (* incomplete *)

end
    
