module E = Errormsg
module N = Num
open Pretty

(*
WARNING:
 
   PREFACE: Two floats are considered equal if they differ by some small amount "delta".
   Small value of "delta" => Loss of completeness
   Large value of "delta" => Loss of soundness
   Ideally, delta should be a function of maxElt as defined below but the relationship is unclear.
   For the time begin, delta has been defined as a constant 1.0e-3
   If it doesn't suit, then play around with its value.

   The current implementation uses the type "float" available in Ocaml
   
   IMPLICATION/WARNING: To avoid loss of completeness, the constants involved in the test-cases should be small and
   randBound should also be small. 

   SOLUTION: However, if we implement a "bigfloat" (i.e. higher precision floats)
   then we may get away with this restriction of constants and randBound being small   
*)

let currentFileDebug = false

module Floating_arithmetic: Arithmetic.arithmetic = struct 
  type elt = float
  let docElt () f = dprintf "%f" f


  (* maximum (absolute value) element that is produced as a result of any computation *)
  (* The error tolerance factor, delta, used in "eq" func should be a function of maxElt *) 
  let maxElt = ref (1.0) 
  let acceptable_error_fraction = 1.0e-6

  let eq f1 f2 = 
    (* let delta = min 1.0e-4 (max 1.0e-6 (!maxElt *. acceptable_error_fraction)) in *)
    let delta = 1.0e-3 in (* the error tolerance factor *)
    let pdiff = abs_float (f1 -. f2) in
    if (currentFileDebug) then ignore (E.log "\n f1 = %f, f2 = %f, pdiff = %f, delta = %f, result=%b" f1 f2 pdiff delta (pdiff < delta));
    pdiff < delta 
      
  let lt f1 f2 = (f1 < f2) & (not (eq f1 f2))
  let gt f1 f2 = (f1 > f2) & (not (eq f1 f2))
  let leq f1 f2 = (f1 < f2) or (eq f1 f2)
  let geq f1 f2 = (f1 > f2) or (eq f1 f2)

  (* Hashing function should be such that equal floats (by the above 
     definition) should have the same hash. Now this is tricky since
     floats are now represented by a range *)
  (* ASSUMPTION: The exponent and the first 3 digits of the significant of lb and ub are same *) 
  (* Basis of Assumption: lb and ub don't differ 'that' much *)
  (* If Assumption Fails: we will loose completeness (not soundness) *)
  (* Detection of Failure of Assumption: Failure of Assumption can be checked by putting an assert *)
  let hashElt f = 
    let (s,e) = frexp f in
    let s' = truncate (s *. 1000.0) in
    Hashtbl.hash (e,s')
  


  (* generates a random number between -randBound and +randBound, both exclusive *)
  let randElt () randBound =
    let doubleRandBound = randBound * 2 in
    let randInt = (Random.int doubleRandBound) - randBound
    in
    (float randInt)
  
  let zero = (float 0)
  let one = (float 1)
  let int2elt i = (float i)

  let absolute f = abs_float f

  let update_maxElt result = 
    let abs_result = absolute result in
    if (abs_result > !maxElt) then maxElt := result  

  let add f1 f2 = 
    let result = f1 +. f2 in
    update_maxElt result;
    result

  let sub f1 f2 = 
    let result = f1 -. f2 in
    update_maxElt result;
    result

  let mult f1 f2 = 
    let result = f1 *. f2 in
    update_maxElt result;
    result

  let div f1 f2 =
    let result = f1 /. f2 in
    update_maxElt result;
    result

  let join f1 f2 r =
    let r' = one -. r in
    let a = f1 *. r in
    let b = f2 *. r' in
    let result = a +. b in
    update_maxElt result;
    result
end
    
