module E = Errormsg
open Pretty

(* WARNING: Overflow will result in loss of completeness i.e. overflow may cause false positives. 
Hence, this module should be used when "randBound" and the constants in the test cases are 
less than (sqrt max_int) (= approx 30000).
Even though we do arithmetic modulo randPrime, overflow may result before modulo is performed.
For e.g. c = (a*b) mod randPrime will overflow if (a*b > max_int)
*) 

module Integer_arithmetic: Arithmetic.arithmetic = struct 
  type elt = int

  (* tunable parameters *)
  let randPrime = 1033
  (* randPrime should be less than (sqrt max_int) to avoid any overflow 
   * problems which will result in loss of completeness *)

  (* computes non-negative mod *)
  let pmod x y = 
    let t = x mod y
    in 
    if (t < 0) then t + y else t

  let docElt () e = dprintf "%d" e
  let eq e1 e2 = ((pmod e1 randPrime) = (pmod e2 randPrime))
  let lt e1 e2 = ((pmod e1 randPrime) < (pmod e2 randPrime))
  let gt e1 e2 = ((pmod e1 randPrime) > (pmod e2 randPrime))
  let leq e1 e2 = ((pmod e1 randPrime) <= (pmod e2 randPrime))
  let geq e1 e2 = ((pmod e1 randPrime) >= (pmod e2 randPrime))
  let hashElt e = Hashtbl.hash e
      
  (* generates a random integer between -randBound and +randBound, both exclusive *)
  let randElt () randBound = 
    let doubleRandBound = randBound * 2 in
    ((Random.int doubleRandBound) - randBound)
  
  let zero = 0
  let one = 1
  let int2elt i = i
  let absolute i = abs i
  let add e1 e2 = pmod (e1 + e2) randPrime
  let sub e1 e2 = pmod (e1 - e2) randPrime
  let mult e1 e2 = pmod (e1 * e2) randPrime
  let join e1 e2 r = pmod ((e1 * r) + (e2 * (1-r))) randPrime
  let div e1 e2 = 
    let gcd a1 b1 = 
      if (a1 = 0) or (b1 = 0) then 1
      else
	let a = ref (abs a1) in
	let b = ref (abs b1) in
	while (!b > 0) do
	  let tmp = !a in
	  a := !b;
	  b := tmp mod !b;
	done;
	!a
    in
     (* fun a1 b1 -> (r1, r2) such that r1*a1 + r2*a2 = 1 *)
    let extended_euclidean_algorithm a1 b1 = 
      let sign i = if (i > 0) then 1 else (-1) in
      let a = ref (abs a1) in
      let b = ref (abs b1) in
      let x0 = ref (sign a1) in
      let x1 = ref (0) in
      let y0 = ref (0) in
      let y1 = ref (sign b1) in
      let x = ref (0) in
      let y = ref (0) in
      while (not (!b = 0)) do
	let q = !a / !b in
	let r = !a mod !b in
	a := !b; b := r;
	x := !x0 - (!x1 * q); x0 := !x1; x1 := !x;
	y := !y0 - (!y1 * q); y0 := !y1; y1 := !y;
      done;
      (!x0,!y0)
    in
     (* mod_ratio is used to map rationals to small integers (modulo randPrime) since arithmetic on small integers is fast *)
     (* fun n d -> (n * d^{-1}) % randPrime, where gcd(n,d) = 1 *)
     (* By definition, d^{-1} is such that d * d^{-1} = 1 modulo randPrime *)
    let mod_ratio n d =
      let g = gcd n d in
      let (n', d') = (n/g, d/g) in (* ensuring gcd(n',d') = 1 *)
      let g = gcd d' randPrime in (* Checking if d' is not a multiple of randPrime *)
      if (g > 1) then begin (* 'Rationals modulo prime' aspect of randomization fails *)
	ignore (E.log "Integer_Arithmetic: We failed: GCD of denominator and randPrime > 1\n ");
	ignore (E.log "............ Try again. Better Luck next time! \n");
	assert (false);
      end;
      let i = fst (extended_euclidean_algorithm d' randPrime) in
      let result = pmod (n' * i) randPrime in   
      result   
    in 
    mod_ratio e1 e2
      
end
    
