(* Windowing Support for the VCGen Graphical User Interface
 * Implemented with LablTk Toolkit
 * Library functions for: Resize, Docking
 * COOLAID Project: 2003
 *)

(* --- Supporting Code --- *)
open Options
open Pretty           (* pretty printing *)
open Tk               (* open the Tk interface *)

(* --- Support Modules --- *)
module E = Errormsg   (* Error messaging module *)
module G = Engine     (* Engine definitions *)
module ST  = Stree    (* State tree *)
module H = Hashtbl    (* hashtable *)
module GL = Guilib
module U = Util

(* --- Type Declarations --- *)
type toplevel = Widget.toplevel Widget.widget (* toplevel type *)
type frame = Widget.frame Widget.widget       (* frames *)
type mousePos = { mutable cursorX: int;   (* type of mouse position *)
		  mutable cursorY: int } 

(******************* Resizing Suport ****************)

(* --- Interface for Resizable Component --- *)
(* UI Components that need to be resized should implement these *)
class virtual resizeComponent =
  object (self)
    method virtual packIn: frame -> unit (* redraw this in frame *)
    method virtual update: unit -> unit (* update method *)
  end



(******************* Hierachy Layout Manager ****************)

(* --- Type Definitions --- *)
type winDim = int
type winPercent = int
type winOrient = Horiz | Vert

type winSplit = { 
    mutable wPercent: winPercent; (* percent given to left, or top *)
    mutable wOrient: winOrient;   (* horizontal or vertical split *)
    mutable wChild1: winSplit option;  (* child split *)
    mutable wChild2: winSplit option;  (* child split *)
    mutable wFrame: frame option;       (* parent frame *)
    mutable wComponent: resizeComponent option;  (* component if leaf *)
    mutable wParent: winSplit option    (* parent *)
  }

(* --- Manipulating splits --- *)

let getPercent (w: winSplit) =
  w.wPercent

let getParentOpt (w: winSplit) =
  w.wParent

let getParent (w: winSplit) =
  match getParentOpt w with
    Some p -> p
  | _ -> E.s(E.error "no parent for split node")

let getOrient  (w: winSplit) =
  w.wOrient

let getComponentOpt (w: winSplit) =
  w.wComponent

let getComponent (w: winSplit) =
  match getComponentOpt w with
    Some c -> c
  | _ -> E.s (E.error "No component in split")

let getFrameOpt (w: winSplit) =
  w.wFrame
 
let getFrame (w: winSplit) =
  match getFrameOpt w with
    Some f -> f
  | _ -> E.s (E.error "No frame in split")

let getChild1Opt (w: winSplit) =
  w.wChild1

let getChild2Opt (w: winSplit) =
  w.wChild2

let getChild1 (w: winSplit) =
  match getChild1Opt w with
    Some c -> c
  | _ -> E.s (E.error "No Child 1 in Split")

let getChild2 (w: winSplit) =
  match getChild2Opt w with
    Some c -> c
  | _ -> E.s (E.error "No Child 2 in Split")

let isLeaf (w: winSplit) =
  match getChild1Opt w, getChild2Opt w with
    None, None -> true
  | _ -> false

let getNode (node: winSplit option) = 
  match node with
    Some t -> t
  | _ -> E.s(E.error "No root for resize top.")

(* --- Class Definitions --- *)

(* class for instantiating resizable toplevels *)
class resizeTop = 
  let root: winSplit option ref = ref None in
  let top: toplevel option ref = ref None in
  let dynamicViews: 
    (frame * resizeComponent) list ref = ref [] in

  (* default packing variables *)
  let itemsPerColumn = 3 in
  let myColumn: winSplit option ref = ref None in
  let myRow: winSplit option ref = ref None in
  let myColumnCount: int ref = ref 0 in
  let myRowCount: int ref = ref 1 in
  
  object (self)

    (* empty tree *)
    method private isEmpty (t: winSplit option) = begin
      match t with
	Some t -> false
      | _ -> true
    end
	  
    (* go through the tree and do the correct resizing *)
    method private reCompute (node: winSplit)
	(orient: winOrient) (percent: winPercent) = begin
      (* propogate upwards *)
      let rec reCompUp node orient percent = begin
	match node with
	  Some n -> 
	    if n.wOrient = orient then
	      n.wPercent <- percent;
	    reCompUp (getParentOpt n) orient percent
	| _ -> ()
      end
      in
      (* propagate downwards *)
      let rec reCompDown node orient size = begin
	match node with
	  Some n ->
	    if n.wOrient = orient then
	      n.wPercent <- percent;
	    reCompDown (getChild1Opt n) orient size;
	    reCompDown (getChild2Opt n) orient (100-size);
	| _ -> ()
      end
      in
      
      (* is this child 1*)
      let isChild1 (node: winSplit option) = 
	match (getParentOpt (getNode node)) with
	  Some p -> 
	    (getChild1Opt p) = node
	| None -> false
      in
      
      (* just need to set these cases *)
      let parent = getParent node in
      if (orient = Horiz) && (parent.wOrient = Vert) then 
	parent.wPercent <- percent
      else if (orient = Vert) && (parent.wOrient = Horiz) then
	parent.wPercent <- percent
      else begin
	(* need to do do recomp *)
	reCompUp (Some node) orient percent;
	reCompDown (Some node) orient percent;
      end
    end


    (* callback for resizing *)
    method private wrapUI parent (componentFun: 'a Widget.widget -> resizeComponent) = begin
      let parentFrame = getFrame parent in
      (* make the resizing frame *)
      let newFrame = Frame.create parentFrame in
      let side = Button.create ~height:1 ~width:1 newFrame in
      let bottom = Button.create ~height:1 ~width:1 newFrame in
      pack [side] ~side:`Right;
      pack [bottom] ~side:`Bottom;

      let ui = componentFun self#getTop in

      ui#packIn newFrame;
      let newLeaf parent ui = 
	{ wPercent = 0;
	  wOrient  = Horiz;
	  wFrame =  Some newFrame ;
	  wChild1 = None;
	  wChild2 = None;
	  wComponent = Some ui; 
	  wParent = Some parent ;
	};
      in

      (* make the leaf *)
      let leaf = newLeaf parent ui in

      (* add the callback *)
      let bindUpDown widget = 
	let myMouse = { cursorX = 0; cursorY = 0 } in
	Tk.bind
	  ~events:([`Modified ([`Double], `ButtonPress)]) 
	  ~fields:[`MouseX; `MouseY]
	  ~action:(fun (e: Tk.eventInfo) -> 
	    let (mouseX: int) = e.ev_MouseX in
	    let (mouseY: int) = e.ev_MouseY in
	    myMouse.cursorX <- mouseX;
	    myMouse.cursorY <- mouseY;)
	  widget;
	Tk.bind
	  ~events:([`Modified ([`Double], `ButtonRelease)]) 
	  ~fields:[`MouseX; `MouseY]
	  ~action:(fun (e: Tk.eventInfo) -> 
	    let (mouseX: int) = e.ev_MouseX in
	    let (mouseY: int) = e.ev_MouseY in
	    (* compute the delta *)
	    let delta = myMouse.cursorY - mouseY in
	    (* maybe need to scale this? *)
	    self#reCompute leaf Vert delta
	    )
	  widget
      in

      let bindLeftRight widget =
	let myMouse = { cursorX = 0; cursorY = 0 } in
	Tk.bind
	  ~events:([`Modified ([`Double], `ButtonPress)]) 
	  ~fields:[`MouseX; `MouseY]
	  ~action:(fun (e: Tk.eventInfo) -> 
	    let (mouseX: int) = e.ev_MouseX in
	    let (mouseY: int) = e.ev_MouseY in
	    myMouse.cursorX <- mouseX;
	    myMouse.cursorY <- mouseY;)
	  widget;
	Tk.bind
	  ~events:([`Modified ([`Double], `ButtonRelease)]) 
	  ~fields:[`MouseX; `MouseY]
	  ~action:(fun (e: Tk.eventInfo) -> 
	    let (mouseX: int) = e.ev_MouseX in
	    let (mouseY: int) = e.ev_MouseY in
	    (* compute the delta *)
	    let delta = myMouse.cursorX - mouseX in
	    (* maybe need to scale this? *)
	    self#reCompute leaf Horiz delta
	    )
	  widget 
      in

      bindUpDown bottom;
      bindLeftRight side;

      (* pack *)
      pack [newFrame] ~side:`Left ~expand:true;


      (* return the leaf *)
      leaf
    end


    (* make new column pack *)
    method private newColumnPack () = begin
      if not (self#isEmpty !root) then 
	E.s(E.error "root not empty for column pack");
	      
      let newRoot = { wPercent = 100;
		      wOrient = Horiz;
		      wFrame = Some (Frame.create self#getTop);
		      wChild1 = None;
		      wChild2 = None;
		      wComponent = None;
		      wParent = None;
		    }
      in
      let newColumn = { wPercent = 100;
			wOrient = Vert;
			wFrame = Some (Frame.create (getFrame newRoot));
			wChild1 = None;
			wChild2 = None;
			wComponent = None; 
			wParent = Some newRoot;
		      }
      in
      newRoot.wChild1 <- Some newColumn;
      root := Some newRoot;
      myRow := Some newRoot;
      myColumn := Some newColumn
    end

    (* default packing alogrithm, make columns, each column with widgets *)
    method private columnPack (componentFun: 'a Widget.widget -> resizeComponent) = begin
      (* check if is empty *)
      if self#isEmpty !root then 
	self#newColumnPack ();
	 
      let newNode parent top = 
	{ wPercent = 100;
	  wOrient = Horiz;
	  wFrame = Some (Frame.create top);
	  wChild1 = None;
	  wChild2 = None;
	  wComponent = None; 
	  wParent = Some parent
	}
      in

      let isFullColumn () =
	!myColumnCount >= itemsPerColumn
      in

      (* add the widget thing *)
      let addToNode (parent: winSplit) (node: winSplit)  = 
	match getChild2Opt node with
	  None -> node.wChild2 <- Some node
	| Some c ->
	    let grow = newNode parent (getFrame parent) in
	    grow.wChild1 <- Some c;
	    grow.wChild2 <- Some node;
      in

      (* recompute orientation and size *)
      let rec doRecompute (count: int) 
	  (percent: winPercent) (orient: winOrient) (w: winSplit option) =
	if count > 0 then begin
	  let node = getNode w in
	  node.wOrient <- orient;
	  node.wPercent <- percent;
	  doRecompute (count - 1) percent orient (getParentOpt node)
	end
      in

      (* check if rightmost column is full *)
      if isFullColumn () then begin
	let newColumn = 
	  newNode (getNode !myRow) (getFrame (getNode !myRow)) in
	addToNode (getNode !myRow) newColumn;
	myColumn := Some newColumn;
	myColumnCount := 0;
	myRowCount := !myRowCount + 1;
	(* recompute for rows *)
	let newPercent = 100 / !myRowCount in
	doRecompute !myRowCount newPercent Horiz !myRow
      end;
      
      (* add to the column *)
      let newDock = self#wrapUI (getNode !myColumn) componentFun in
      addToNode (getNode !myColumn) newDock;
      myColumn := Some newDock;
      myColumnCount := !myColumnCount + 1;

      (* recompute for columns *)
      let newPercent = 100 / !myColumnCount in
      doRecompute !myColumnCount newPercent Vert !myColumn;

      (* redraw *)
      self#redraw
    end


    (* methods to redraw the root *)
    method private redrawNode (w: winSplit) = begin
      match w.wComponent with
	Some ui ->
	  (* draw the child *)
	  ui#packIn (getFrame (getParent w))
      | _ -> ()
    end

    (* method to redraw a tree *)
    method private redrawTree (t: winSplit option) = begin
      match t with
	Some w ->
	  if isLeaf w then self#redrawNode w
	  else begin
	    (* get orientation *)
	    let orientation = 
	      match w.wOrient with
		Horiz -> `Left
	      | Vert -> `Top
	    in
	    (* configure frame percent *)
	    let configFrameGetList () =
	      let configFrame frame percent =
		match w.wOrient with
		  Horiz ->
		    Frame.configure (getFrame frame) ~width:percent
		| Vert ->
		    Frame.configure (getFrame frame) ~height:percent
	      in
	      match getChild1Opt w, getChild2Opt w with
		Some c1, Some c2 -> 
		  configFrame c1 w.wPercent;
		  configFrame c2 (100 - w.wPercent);
		  [getFrame c1; getFrame c2] 
	      | Some c1, None -> 
		  configFrame c1 w.wPercent;
		  [getFrame c1]
	      | None, Some c2 -> 
		  configFrame c2 (100 - w.wPercent);
		  [getFrame c2]
	      | _ -> []
	    in
	    (* get packing list *)
	    let list = configFrameGetList () in
	    pack list ~side:orientation;
	    
	    (* recursively do children *)
	    self#redrawTree w.wChild1;
	    self#redrawTree w.wChild2;
	  end
      | _ -> ()
    end

    (* default packing *)
    method pack (componentFun: 'a Widget.widget -> resizeComponent) = 
      self#columnPack componentFun


    (* --- Public Methods --- *)
    (* initialize *)
    method init = begin
      top := Some (Toplevel.create (GL.tkTop ()))
    end

    (* redraw everything *)
    method redraw = begin
      self#redrawTree !root
    end
	
    (* the root for packing *)
    method getTop = begin
      match !top with 
	Some t -> t
      | _ -> E.s(E.error "No TK Top found\n")
    end
	
    (* add a new window *)
    method addDock (componentFun: 'a Widget.widget -> resizeComponent) = begin
      (* first we need to wrap this to get callback and handle bars *)
      self#pack componentFun;
      (* redraw *)
      self#redraw
    end
	
    (* set the name *)
    method setTitle (str: string) = 
      Wm.title_set self#getTop str
    
	
    (* updating *)
    method updateUI (force: bool) = 
      if force then begin
	(* update the dynamic views *)
	List.iter
	  (fun (_,view) -> view#update ())
	  !dynamicViews;
	
      end;
      (* do a tk draw by yielding to gtk *)
      Tk.update ();
      ()
	
  end
    

      
