// c2v_parse_c.cc
// code for c2v_parse_c.h

#define C2V_PARSE_C_IMPL
#include "c2v_parse_c.h"    // this module
#include "lexer.h"          // Lexer for C++
#include "exc.h"            // xbase
#include "strutil.h"        // pluraln
#include "ccparse.h"        // numAmbiguousNodes

#include <string.h>         // strlen


// ------------------------- C2VParseC --------------------------
C2VParseC::C2VParseC(StringTable &s, CCLang &L)
  : strTable(s),
    lang(L),

    parseContext(s, L),
    tables(parseContext.makeTables()),
    glr(&parseContext, tables),

    errors(0)
{}

C2VParseC::~C2VParseC()
{}


SemanticValue C2VParseC::parseFragment(int preToken, SourceLoc loc, StringRef text)
{
  TRACE("c2vparsec", "parsing embedded C at " << toString(loc));

  // lexer
  Lexer lexer(strTable, lang, loc, text, strlen(text));

  // prime it with the special pre-token
  lexer.type = preToken;
  lexer.sval = NULL_SVAL;
  lexer.loc = loc;     // probably redundant

  // parse it
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    // 'glrParse' will have already printed a complaint
    errors++;

    // for now, stop on first error
    xbase(stringc << toString(loc) << ": parsing embedded C failed");
  }

  // add lexing errors to the count
  errors += lexer.errors;

  return treeTop;
}

template <class T>
static void checkAmbig(SourceLoc loc, T *&node)
{
  // apply disambiguation
  PatternDisambig disambig;
  disambig.mtraverse(node);

  // check to see if there is any remaining ambiguity
  int a = numAmbiguousNodes(node);
  if (a) {
    xbase(stringc << toString(loc)
                  << ": contains " << pluraln(a, "ambiguous node"));
  }
}


bool C2VParseC::visitC_Statement(C_Statement *obj)
{
  obj->cstmt =
    (Statement*)parseFragment(TOK_PARSE_STATEMENT, obj->loc, obj->text);
  checkAmbig(obj->loc, obj->cstmt);
  return true;         // irrelevant
}


bool C2VParseC::visitC_Expression(C_Expression *obj)
{
  obj->cexpr =
    (Expression*)parseFragment(TOK_PARSE_EXPRESSION, obj->loc, obj->text);
  checkAmbig(obj->loc, obj->cexpr);
  return true;
}

bool C2VParseC::visitC_TypeId(C_TypeId *obj)
{
  obj->ctype =
    (ASTTypeId*)parseFragment(TOK_PARSE_TYPEID, obj->loc, obj->text);
  checkAmbig(obj->loc, obj->ctype);
  return true;
}


void c2v_parse_c(C2VFile *file, StringTable &strTable, CCLang &lang)
{
  C2VParseC vis(strTable, lang);
  file->traverse(vis);
  if (vis.errors) {
    xbase(stringc << "parsing embedded C failed with "
                  << pluraln(vis.errors, "error"));
  }
}


// ------------------------- PatternDisambig -------------------------
bool PatternDisambig::visitExpression(Expression *&obj)
{
  // this function also serves as a syntax normalizer
  if (obj->isBinary(BIN_BRACKETS)) {
    E_binary *b = obj->asE_binary();
    
    // translate a[b] into *(a+b)          
    b->op = BIN_PLUS;
    obj = new E_deref(b);
  }
  
  if (obj->isE_arrow()) {
    // translate a->b into (*a).b
    E_arrow *arrow = obj->asE_arrow();
    Expression *a = arrow->obj;
    PQName *b = arrow->fieldName;
    obj = new E_fieldAcc(new E_deref(a), b);
  }

  // if no ambiguity, skip
  if (!obj->ambiguity) {
    return true;
  }
  
  if (obj->ambiguity->ambiguity) {
    throw_XUnimp("more than 2 ambiguous alternatives");
  }

  Expression *e1 = obj;
  Expression *e2 = obj->ambiguity;

  if (e1->isE_funCall() && e2->isE_constructor()) {
    // canonicalize which is which
    Expression *tmp = e1;
    e1 = e2;
    e2 = tmp;
  }

  if (e1->isE_constructor() && e2->isE_funCall()) {
    // always prefer the funcall interpretation
    e2->ambiguity = NULL;
    mtraverse(e2);
    obj = e2;        // replace parent's pointer
    return false;    // already did children, don't do again
  }

  throw_XUnimp("unhandled ambiguity");
  return true;
}


// EOF
