// doublylinked.c
// doubly-linked list, modeled after pale/Examples/doublylinked.pale

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar


// --------------- code to verify -----------------
struct Node;

typedef struct Head {
  struct Node *first;   // NULL or valid
  struct Node *last;    // NULL or last element in list
} Head;

typedef struct Node {
  int value;
  struct Node *next;    // invariant: NULL or valid
  struct Node *prev;    // thing whose 'next' points at me
  struct Head *head;    // ghost: pointer to my list's header object
  struct Node *tail;    // last element in list
} Node;

// invariant: NULL or valid
Node *garbage = NULL;


// this disables the effect of the pattern specs
#if 0
  #define PATS(...) /*nothing*/
  #define MPAT(...) /*nothing*/
#endif // 0


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // garbage
  garbage!=NULL ==> all_of {
    tag[garbage]==Node_tag;
    garbage->prev==NULL;
    garbage->head==NULL;
  };


  // Head

  // 'first' is valid if not NULL
  forall(Head *h) PATS(h->first) .
    tag[h]==Head_tag && h->first!=NULL ==> tag[h->first] == Node_tag;

  // list labels are accurate; 'first' is injective
  forall(Head *h) PATS(h->first) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->head == h;

  // 'first' points to the first element (if not NULL)
  forall(Head *h) PATS(h->first->prev) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->prev == NULL;

  // similar for 'last'
  forall(Head *h) PATS(h->last) .
    tag[h]==Head_tag && h->last!=NULL ==> tag[h->last] == Node_tag;
  forall(Head *h) PATS(h->last) .
    tag[h]==Head_tag && h->last!=NULL ==> h->last->head == h;
  forall(Head *h) PATS(h->last->next) .
    tag[h]==Head_tag && h->last!=NULL ==> h->last->next == NULL;

  // connect 'last' with 'tail'
  forall(Head *h) PATS(h->first->tail, MPAT(h->first, h->last)).
    tag[h]==Head_tag && h->last!=NULL && h->first!=NULL ==> h->first->tail == h->last;

  // should have 'first' iff have 'next'; actually, I cannot ensure this
  // because my invariants about 'first' and 'last' are too weak
  //
  // maybe they are now strong enough, with 'tail'?
  forall(Head *h).
    tag[h]==Head_tag ==> (h->first==NULL <==> h->last==NULL);

  // Node

  // declared type of 'next' is accurate
  forall(Node *n) PATS(n->next) .
    tag[n]==Node_tag && n->next!=NULL ==> tag[n->next] == Node_tag;

  // declared type of 'prev' is accurate
  forall(Node *n) PATS(n->prev) .
    tag[n]==Node_tag && n->prev!=NULL ==> tag[n->prev] == Node_tag;

  // 'tail' is well-typed and never null
  forall(Node *n) PATS(n->tail) .
    tag[n]==Node_tag ==> tag[n->tail]==Node_tag;

  // transitivity of list label
  forall(Node *n) PATS(MPAT(n->next, n->head), n->next->head) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->head == n->head;

  // injectivity of 'next'
  forall(Node *n) PATS(n->next) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->prev == n;

  // injectivity of 'prev' (seems to be needed...)
  forall(Node *n) PATS(n->prev) .
    tag[n]==Node_tag && n->prev!=NULL ==> n->prev->next == n;

  // transitivity of 'tail'
  forall(Node *n) PATS(MPAT(n->next, n->tail), n->next->tail) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->tail == n->tail;

  // termination of 'tail'
  forall(Node *n) PATS(n->next) .
    tag[n]==Node_tag && n->next==NULL ==> n->tail == n;

  // connect 'head' to 'last'
  forall(Node *n) PATS(n->head->last, MPAT(n->head, n->tail)).
    tag[n]==Node_tag && n->head!=NULL ==> n->head->last == n->tail;
    
  // connect 'prev' and 'tail' (weird...)
  forall(Node *n) /*hope for the best*/.
    tag[n]==Node_tag && n->tail!=n ==> n->tail->prev != NULL;

});}


// ----------------------- delete -----------------------
#if 0     // doesn't work yet
void delete(Head *list, Node *p)
  pre(all_of {
        tag[list]==Head_tag;      // implied by 'data' keyowrd
        tag[p]==Node_tag;         // p!=null
        p->head == list;          // list.first<next*>p
        /* ? */                   // list.last<prev*>p
        /* ? */                   // p<next*>list.last
        /* ? */                   // p<prev*>list.first
        /* ? */                   // list.first<next*>S
      })
  post(all_of {
         p->head == NULL;         // !L.first<next*>P
         forall(Node *q).         // allpos q of Node: q in S <=> (L.first<next*>q | q=P)
           q!=p ==> pre(q->head) == q->head;
       })
  changes(Head, Node, garbage)
{
  if (p == list->first) {                 // t
    list->first = p->next;
  }
  if (p == list->last) {                  // f
    list->last = p->prev;

    // update all of 'last' in one shot (it is ghost)
    Node_tail = fresh_value();
    assume(forall(Node *n). all_of {
      pre(n->head)!=list ==> n->tail == pre(n->tail);
      pre(n->head)==list ==> n->tail == p->prev;
    });
  }
  if (p->prev != NULL) {                  // f
    p->prev->next = p->next;
  }
  if (p->next != NULL) {                  // t
    p->next->prev = p->prev;
  }
  if (garbage != NULL) {                  // t
    garbage->prev = p;
    p->tail = garbage->tail;
  }
  else {
    p->tail = p;
  }
  p->next = garbage;
  p->prev = NULL;
  p->head = NULL;
  garbage = p;
}
#endif // 0


Node *search(Head *list, int value)
  pre(tag[list] == Head_tag)      // list!=null
  post(all_of {
    /* redundant! */              // L!=null
    result!=NULL ==> result->value == pre(value);     // approximates pale's postcondition...
  })
  changes()
{
  Node *t = list->first;
  while (t!=NULL && t->value != value) {

    invariant(all_of {
      tag[list] == Head_tag;      // list!=null
      pre(value) == value;        // V=value
      pre(list) == list;          // L=list
      t->head == list;            // list.first<next*>t
      // cannot state: allpos p of Node: p<next.next*>t => p.value!=value]
      
      // additional
      tag[t] == Node_tag;
    });

    t = t->next;
  }
  return t;
}


void insert(Head *list, Node *node)
  pre(all_of {
    tag[list] == Head_tag;          // list!=null 
    tag[node] == Node_tag;          // node!=null
    node->next == NULL;             // node.next=null
    node->prev == NULL;             // node.prev=null
    
    // additional
    node->head == NULL;
    node != garbage;
  })
  post(1==1)
  changes(Head, Node)
{
  if (list->first == NULL) {
    list->last = node;
  }
  else {
    list->first->prev = node;
    node->tail = list->first->tail;
  }
  node->head = list;
  node->next = list->first;
  list->first = node;
  node = NULL;
}


Head *concat(Head *list1, Head *list2)
  pre(all_of {
    tag[list1] == Head_tag;       // list1!=null
    tag[list2] == Head_tag;       // list2!=null
    list1!=list2;
  })
  post(1==1)
  changes(Head, Node)
{
  // list1->first list1->last  list2->first list2->last
  //      |          |              |          |
  //      V          V              V          V
  //     +--+       +--+           +--+       +--+
  //     |  |->...->|  |->null     |  |->...->|  |->null
  //     +--+       +--+           +--+       +--+

  if (list1->last != NULL) {
    list1->last->next = list2->first;
  }
  else {
    list1->first = list2->first;
  }
  if (list2->first != NULL) {
    list2->first->prev = list1->last;
    list1->last = list2->last;
  }
  list2->first = NULL;
  list2->last = NULL;

  // list1->first                           list1->last
  //      |                                    |
  //      V                                    V
  //     +--+       +--+           +--+       +--+
  //     |  |->...->|  |---------->|  |->...->|  |->null
  //     +--+       +--+           +--+       +--+

  // update all of 'head' in one shot (it is ghost)
  Node_head = fresh_value();
  assume(forall(Node *n). all_of {
    pre(n->head)!=list2 ==> n->head == pre(n->head);
    pre(n->head)==list2 ==> n->head == list1;
  });

  // update all of 'last' in one shot (it is ghost)
  Node_tail = fresh_value();
  assume(forall(Node *n). all_of {
    pre(n->head)!=list1 ==> n->tail == pre(n->tail);
    pre(n->head)==list1 ==> n->tail == list1->last;
  });

  return list1;
}






// EOF
