// hashtable.c
// verification example with a hashtable

// ------------------- preamble ----------------
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"minmax.vc\";");
asm("verifier: include \"cont_po.vc\";");    // CLT, CLT_between

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int data, int count);

// declarations of model functions
int sel(int mem, int p);
int ptr(int array, int index);
int min(int a, int b);
int max(int a, int b);
int fresh_value();
int CLT(int x, int y);
int CLT_between(int x, int y);

int tag;
int objct;
int objsize;
         
// holds arrays
int mem;
int meta;


// ---------------- code to verify -------------
#define AND all_of
#define OR one_of


typedef int bool;
#define false 0
#define true 1

  
// name the field that points to a node
enum InvField { IF_NEXT, IF_TABLE };

// node in a bucket
struct Node {
  int key, data;
  Node *next;

  // inverse of 'next' and 'table'
  InvField referrer;                               
  Node *prev;            // if referrer == IF_NEXT
  int myIndex;           // if referrer == IF_TABLE
};

#define isNode(n) (0 < (n) < objct && tag[n]==Node_tag)
#define NODE_TYPE(field) (field)!=NULL ==> isNode(field)

// hashtable, indexed by key's hash value
Node **table;
int tableSize;

// work around unforunate mem/ptr thing
#define tableAcc(i) mem[ptr(table,i)]


// modifications that automatically set inverses
#define setNext(n,m) {            \
  n->next = m;                    \
  if (n->next) {                  \
    n->next->referrer = IF_NEXT;  \
    n->next->prev = n;            \
  }                               \
}

#define setTableIndex(table,i,n) {     \
  table[i] = n;                        \
  if (table[i]) {                      \
    table[i]->referrer = IF_TABLE;     \
    table[i]->myIndex = i;             \
  }                                    \
}


// global invariant
int global_inv() { return vc_expression(AND {
  0 < objct;
  tag[0] == 0;

  0 < table < objct;
  tableSize == objsize[table];

  forall(int i). 0 <= i < tableSize && tableAcc(i)!=NULL ==> AND {
    isNode(tableAcc(i));
    Node_referrer[tableAcc(i)] == IF_TABLE;
    Node_myIndex[tableAcc(i)] == i;
  };
  
  forall(Node *n). isNode(n) ==> AND {
    n->next!=NULL ==> AND {
      isNode(n->next);
      n->next->referrer == IF_NEXT;
      n->next->prev == n;
    };
  };
  
});}


// compute hash function
int hashfunc(int)
  pre(true)
  post(0 <= result < tableSize)
  changes()
;


// lookup an element
int lookup(int key)
{
  int index = hashfunc(key);

  Node *n = table[index];

  while (n && n->key != key) {
    invariant(global_inv && isNode(n));
    n = n->next;
  }
  
  if (n) {
    return n->data;
  }
  else {
    return 0;   // not found
  }
}


// insert an element into table
void insert(int key, int data)
  post(AND {
    objct >= pre(objct);
    forall(int addr). addr<pre(objct) ==> AND {
      tag[addr] == pre(tag)[addr];
      objsize[addr] == pre(objsize)[addr];
    };
  })
  changes(tag, objct, objsize, Node_tag, mem, meta)
{
  int index = hashfunc(key);

  Node *n = (Node*)malloc(sizeof(Node));
  n->key = key;
  n->data = data;

  // prepend 'n' to list at 'table[index]'
  setNext(n, table[index]);           // n->next = table[index];
  setTableIndex(table, index, n);     // table[index] = n
}


// remove an element
void remove(int key)
  changes(tag, objct, objsize, Node_tag, mem, meta)
{
  int index = hashfunc(key);

  if (!table[index]) {
    return;       // not present
  }

  if (table[index]->key == key) {
    // remove first element
    Node *n = table[index];
    setTableIndex(table, index, n->next);
    free(n);
    return;
  }

  // find element that precedes one to delete
  Node *before = table[index];
  while (before->next && before->next->key != key) {
    invariant(AND {
      global_inv;
      isNode(before);
      before->next!=NULL;
    });

    before = before->next;
  }

  if (!before->next) {
    return;       // not present
  }

  // remove element after 'before'
  Node *m = before->next;
  setNext(before, m->next);     //  before->next = m->next;
  free(m);
}


// change 'tableSize'
void resize(int newSize)
  changes(tag, objsize, objct, Node_tag, mem, meta, table, tableSize)
{
  // get ahold of old table
  Node *oldTable = table;
  int oldSize = tableSize;

  // make a new, empty table
  tableSize = newSize;
  table = (Node**)malloc(tableSize * sizeof(Node*));
  assume(tag[table]!=Node_tag);
  memset(table, 0, tableSize * sizeof(Node*));

  // re-insert elements, free old nodes
  for (int i=0; i<oldSize; i++) {
    invariant(AND {
      global_inv;
      0 <= i < oldSize;
      0 < oldTable < objct;
      objsize[oldTable] == oldSize;
      forall(int i). 0 <= i < oldSize && mem[ptr(oldTable,i)]!=NULL ==>
        isNode(mem[ptr(oldTable,i)]);
    });

    Node *n = oldTable[i];
    while (n) {
      invariant(AND {
        global_inv;
        0 <= i < oldSize;
        0 < oldTable < objct;
        objsize[oldTable] == oldSize;
        forall(int i). 0 <= i < oldSize && mem[ptr(oldTable,i)]!=NULL ==>
          isNode(mem[ptr(oldTable,i)]);
        isNode(n);
      });

      insert(n->key, n->data);

      Node *tmp = n;
      n = n->next;
      if (n) {
        n->referrer = IF_TABLE;
        n->myIndex = i;
      }
      //free(tmp);
    }
  }

  //free(oldTable);
}


// EOF
