// list.c
// linked list

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar
//ORDER: #include "dense_po.model"   // DLT, DLT_between

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);


// --------------- code to verify -----------------
typedef struct Node {
  int data;
  struct Node *next;    // invariant: NULL or valid
  struct Node *prev;    // ghost: thing whose 'next' points at me
  struct Node *listId   // pointer to my list's first element
    // At the moment, adding a typing axiom for 'listId' sends the
    // prover into a matching loop, so this annotation disables the
    // typing axiom.  I think it can be fixed in a more principled
    // way but I want to get the change committed.
    __varattribute__((noTypingAxiom));

  //ORDER: int order;            // dense partial order element (for acyclicity)
} Node;

// invariant: NULL or valid
Node *head = NULL;

// invariant: NULL or points to a node whose listId is wrong (that is,
// it appears wrong from the point of view of whichever node has this
// one as its 'next')
Node *brokenListId = NULL;


// global invariants
int global_inv() { return vc_expression(all_of {
  0 < objct;                            // next address isn't NULL
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // head
  head!=NULL ==> all_of {
    0 < head < objct && tag[head]==Node_tag;
    head->prev==NULL;                   // and first element's 'prev' is NULL
  };          
  
  // brokenListId
  brokenListId!=NULL ==> all_of {
    0 < brokenListId < objct && tag[brokenListId]==Node_tag;
  };

  // list labels are never NULL
  forall(Node *n) PATS(tag[n]). tag[n]==Node_tag ==>
    n->listId != NULL;

  #if 0    // not needed anymore
  // declared type is accurate
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==>
      tag[n->next] == Node_tag;

  // transitivity of allocated-ness
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==>
      0 < n->next < objct;
  #endif // 0

  // transitivity of list label
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==>
      n->next!=brokenListId ==>
        n->next->listId == n->listId;

  // injectivity of 'next'
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==>
      n->next->prev == n;

  // acyclicty
  //ORDER: n->order < n->next->order;

  #if 0    // not needed
  // declared type of 'prev' is accurate
  forall(Node *n) PATS(n->prev).
    tag[n]==Node_tag && n->prev!=NULL ==>
      0 < n->prev < objct && tag[n->prev]==Node_tag;
  #endif // 0

  // reverse injectivity... needed b/c I added the assertion about
  // prev's type, which itself is needed for me to be able to
  // deref the prev ptr..
  forall(Node *n) PATS(n->prev).
    tag[n]==Node_tag && n->prev!=NULL ==>
      n->prev->next == n;

  // seed for list label
  forall(Node *n) PATS(MPAT(n->listId,n->prev)).
    tag[n]==Node_tag && n!=brokenListId && n->prev==NULL ==>
      n->listId == n;
});}


// append node with data==val
void append(int val)
  pre(brokenListId == NULL)
  post(brokenListId == NULL)
  changes(tag, objct, head, Node, objsize)
{
  Node *newNode = (Node * /*owner*/)malloc(sizeof(Node));
  newNode->data = val;
  newNode->next = NULL;
  newNode->prev = NULL;
  newNode->listId = newNode;
  // order not yet set

  if (!head) {
    head = newNode;                // ownership transfer
    //ORDER: head->order = DLT_between(bottom_level, top_level);

    assert(head->next==NULL && head->prev==NULL);
    return;
  }

  Node *p = head;

  while (p->next) {
    invariant {
      // precondition
      brokenListId == NULL;

      // loop guard
      p->next != NULL;

      // p points to a valid list node
      0 < p < objct && tag[p]==Node_tag;

      // so does newNode
      0 < newNode < objct && tag[newNode]==Node_tag;
      
      // newNode is a singleton list; consequently, I don't
      // have to update more than one 'listId' when I append it
      newNode->prev == NULL;
      newNode->next == NULL;

      // 'p' is in the same list as 'head'
      p->listId == head->listId;

      // but 'newNode' is in a different list
      p->listId != newNode->listId;
    }

    p = p->next;
  }

  assert(p->next == NULL);
  assert(newNode != head);

  p->next = newNode;               // ownership transfer
  newNode->prev = p;               // ghost
  newNode->listId = p->listId;

  assert(newNode->listId != 0);

  // let newNode->order be s.t. bottom_level < newNode->order < p->order
  //ORDER: newNode->order = DLT_between(bottom_level, p->order);
}


// repair a broken list id
void fixBrokenListId()
  pre(brokenListId != NULL)
  post(brokenListId == NULL)
  changes(brokenListId, Node_listId)
{
  if (!brokenListId->prev) {
    brokenListId->listId = brokenListId;
  }
  else {
    brokenListId->listId = brokenListId->prev->listId;
  }
  
  brokenListId = brokenListId->next;
  if (brokenListId) {
    fixBrokenListId();
  }
}


// remove first node with data==val, if it exists
void remove(int val)
  pre(brokenListId == NULL)
  post(brokenListId == NULL)
  changes(tag, head, Node)
{
  if (!head) {
    return;
  }

  if (head->data == val) {
    Node *temp = head;          // ownership transfer

    // break the list so 'temp' is a singleton

    head = temp->next;          // ownership transfer
    temp->next = NULL;
    if (head) {
      head->prev = NULL;

      // this one is proved by both Kettle and Simplify
      assert(forall(Node *p) PATS(p->next). !(tag[p]==Node_tag && p->next==temp));

      // this is proved by Simplify, but not by Kettle
      // 2005-02-16: now Kettle can prove it too
      assert(!exists(Node *p) PATS(p->next). tag[p]==Node_tag && p->next==temp);
      assert(!exists(Node *p) PATS(p->prev). tag[p]==Node_tag && p->prev==temp);

      // now, the list rooted at 'head' is well-formed except that
      // the nodes' listId still points at 'temp'; fix that
      brokenListId = head;
      fixBrokenListId();
    }

    // temp->next is dead

    // this fact is crucial to ensure that free() is not going
    // to break the global invariant
    assert(!exists(Node *p) PATS(p->next). tag[p]==Node_tag && p->next==temp);
    assert(!exists(Node *p) PATS(p->prev). tag[p]==Node_tag && p->prev==temp);

    free(temp);                 // ok, *temp is dead
    return;
  }

  Node *p = head;

  while (p->next &&
         p->next->data != val) {
    // p is a valid serf pointing within *head
    invariant {
      brokenListId == NULL;
      0 < p < objct && tag[p]==Node_tag;
      p->next != NULL;       // loop guard
      p->next->data != val;  // not needed
    }

    p = p->next;
  }

  if (p->next) {
    Node *temp2 = p->next;             // ownership transfer

    // p->next is dead

    // no serf points within *temp

    p->next = temp2->next;             // ownership transfer
    if (temp2->next) {
      temp2->next->prev = p;           // ghost
    }

    // temp->next is dead

    assert(temp2->prev == p);

    // *p is valid

    assert(!exists(Node *q) PATS(q->next). tag[q]==Node_tag && q->next==temp2);

    free(temp2);                       // ok, *temp is dead
  }

  // when BUG is #defined, Kettle times out trying to prove it
  // on path 1/0, whereas Simplify fails quickly (<0.3s)
  // #define BUG
  #ifdef BUG
  assert(1==0);
  #endif
}


/*
  Errors to catch:
    - failure to restore some node's validity
*/
