// scheduler.c
// example from PLDI submission

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar

// this code does not use objct
verifier_option((no_objct));


// --------------- code to verify -----------------
typedef struct Process Process;

typedef enum State { SLP, RUN } State;

struct Process {
  int prio;
  State state;
  Process *next;
  Process *prev;
};

Process *runnable;
Process *sleeping;


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  #if 0      // not needed anymore
  // declared type of 'next' is accurate
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag && p->next!=NULL ==> tag[p->next] == Process_tag;

  // declared type of 'prev' is accurate
  forall(Process *p) PATS(p->prev).
    tag[p]==Process_tag && p->prev!=NULL ==> tag[p->prev] == Process_tag;
  #endif // 0

  // A1: injectivity of 'next'
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag && p->next!=NULL ==> p->next->prev == p;

  // A2: reverse injectivity
  forall(Process *p) PATS(p->prev).
    tag[p]==Process_tag && p->prev!=NULL ==> p->prev->next == p;

  // A3: state is transitive
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag && p->next!=NULL ==> p->state == p->next->state;

  // A4: runnable list is sorted
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag && p->next!=NULL && p->state==RUN ==>
      p->prio >= p->next->prio;

  // A5: runnable dummy node priority is highest
  forall(Process *p) PATS(tag[p]).
    tag[p]==Process_tag ==> runnable->prio >= p->prio;

  // list headers are never null
  tag[runnable]==Process_tag;
  tag[sleeping]==Process_tag;
  
  // list headers have proper state
  runnable->state == RUN;
  sleeping->state == SLP;
  
  // they point to first elements
  runnable->prev == NULL;
  sleeping->prev == NULL;
});}


// precondition: at.prio >= what.prio
// precondition: at and what are runnable (RUN)
void insert(Process *at, Process *what)
  pre(all_of {
    tag[at]==Process_tag;
    tag[what]==Process_tag;
    at->state == RUN;
    what->state == RUN;
    at->prio >= what->prio;
    what->next == NULL;
    what->prev == NULL;
    what != runnable;
  })
  changes(Process_next, Process_prev)
{
  if (at->next != NULL && at->next->prio >= what->prio) {
    insert(at->next, what);
  }
  else {
    //           what
    //            |
    //            V
    //    at     +--+
    //     |     |  |
    //     V     +--+
    //    +--+         +--+
    //    |  |-------->|  |-->..
    //    +--+         +--+
    what->prev = at;
    what->next = at->next;
    if (at->next) {
      at->next->prev = what;
    }
    at->next = what;
  }
}


// precondition: x is sleeping (SLP)
// precondition: x is not first in list
// postcondition: x is runnable (RUN)
void wakeup(Process *x)
  pre(tag[x]==Process_tag && x->state==SLP && x->prev!=NULL)
  post(x->state==RUN)
  changes(Process)
{
  // remove from sleeping list
  x->prev->next = x->next;
  if (x->next) {
    x->next->prev = x->prev;
  }
  // set the state to match its new home
  x->state = RUN; 
  x->next = NULL;
  x->prev = NULL;
  // insert in runnable list
  insert(runnable, x);
}


// EOF
