// exttype.h
// extended types for verifier

#ifndef EXTTYPE_H
#define EXTTYPE_H

#include "cc_type.h"       // original Type hierarchy
#include "variable.h"      // Variable
#include "ownerserf.h"     // nullable, owner

class Expression;          // cc.ast
class FA_changes;          // cc.ast
class TF_changeGroup;      // cc.ast
class VCEnv;               // vc_env.h


// --------------------- ECompoundType -----------------------
class ECompoundType : public CompoundType {
public:      // data 
  // model variables that represent the size and tag
  Variable * nullable owner sizeVar;
  Variable * nullable owner tagVar;

  // and variables for the offsets/memories of the fields
  ObjList<Variable> fieldSlots;

  // this class's tag number, or -1 if not yet set
  int typeTag;

  // "size" of this object; essentially, # of fields
  int objectSize;

  // vclang text to emit during vcgen; computed during tcheck
  stringBuilder sb;

  // true if we're modeling the contents of the structure as
  // a flat array of fields; if false, we're modeling it as
  // one memory per field
  bool flatMap;

  // declared superclass; only single inheritance is supported in
  // the current incarnation of subtyping for C
  ECompoundType *superclass;         // (nullable serf)

  // subclasses; inverse relation of 'superclass'
  SObjList<ECompoundType> subclasses;

public:      // funcs
  ECompoundType(Keyword k, StringRef name);
  ~ECompoundType();
                                           
  // also updates 'subclasses'
  void setSuperclass(ECompoundType *super);
};

// downcast; simply assume that every instance of CompoundType is an
// ECompoundType, since the former will never be created and there
// are no other subclasses
inline ECompoundType *ext(CompoundType *ct)
  { return static_cast<ECompoundType*>(ct); }


// --------------------- EFunctionType -----------------------
class EFunctionType : public FunctionType {
public:      // data
  // pre- and postcondition predicates
  Expression * nullable owner precondition;
  Expression * nullable owner postcondition;

  // list of potentially changed global variables; this is a pointer
  // into the AST when it's not NULL
  FA_changes * nullable serf changes;
  
  // hack: 'changes' rendered as a conjunction so I can jam it into
  // every invariant in the function body
  Expression * nullable owner changesConj;

  // variable created to stand for the return value; this might
  // be referred to by the pre- or postcondition
  Variable * nullable owner result;

public:      // funcs
  EFunctionType(Type *retType);    // pre/post initially NULL
  virtual ~EFunctionType();

  // syntax printing
  virtual void extraRightmostSyntax(stringBuilder &sb) const;

  // retrieve the postcondition; fold 'changes' into it if that
  // hasn't already been done
  Expression * nullable serf getPostcondition(VCEnv &env);
};

inline EFunctionType *ext(FunctionType *ft)
  { return static_cast<EFunctionType*>(ft); }


// --------------------- EVariable --------------------------
// attributes that can be attached with __varattribute__ mechanism
enum VarAttr {
  VA_NONE          = 0x0000,

  VA_PARAMBYREF    = 0x0001,      // pass-by-reference parameter
  VA_ARGBYREF      = 0x0002,      // pass-by-reference argument
  VA_ADDRTAKEN     = 0x0004,      // address might be taken
  VA_FUNCSYM       = 0x0008,      // function symbol, i.e. E_apply not S_funCall
  VA_TRIGGER       = 0x0010,      // nominally, the "pat" function
  VA_ARRAY         = 0x0020,      // array model 1: objsize/mem/meta/ptr
  VA_PREDSYM       = 0x0040,      // predicate symbol; affects vcgen translation
  VA_PATS_MPAT     = 0x0080,      // this function takes pattern terms as arguments
  VA_NOTYPINGAXIOM = 0x0100,      // disable automatic typing axiom

  VA_ALL           = 0x01FF,      // all flags
  NUM_VATTRFLAGS   = 9            // # of '1' bits in VA_ALL
};

ENUM_BITWISE_OPS(VarAttr, VA_ALL);

extern char const * const varAttrFlagNames[NUM_VATTRFLAGS];      // 0="parambyref" ..
string toString(VarAttr attrs);               

// e.g. "array" -> VA_ARRAY; returns VA_NONE for unrecognized names
VarAttr stringToVarAttr(char const *str);


class EVariable : public Variable {
public:      // data
  VarAttr attr;

  // if this is a change_group, this is non-NULL
  TF_changeGroup *changeGroup;        // (nullable serf)

  // list of synthesized variables representing the fields of
  // this global structure instance, if that's what this is
  // (otherwise the list is empty)
  ObjList<EVariable> synthFields;

  // if I am a synthField or a global array of field elements, 
  // what field do I stand for?
  EVariable *originalStructField;     // (nullable serf)
  
  // if I am a struct field, modeled by a model variable as an
  // array of field elements, this points to that model variable
  EVariable *modelVar;                // (nullable serf)
  
  // as a model variable, what is my offset?
  int modelVarOffset;

public:      // funcs
  EVariable(SourceLoc L, StringRef n, Type *t, DeclFlags f);
  ~EVariable();

  // attach the named string as a variable attribute, or return
  // false if it's not the name of any valid attribute
  bool attachAttribute(StringRef name);

  // test an attribute flag
  bool hasVAttr(VarAttr a) const { return !!(attr&a); }

  // Variable funcs
  virtual string namePrintSuffix() const;
};

inline EVariable *ext(Variable *v)
  { return static_cast<EVariable*>(v); }
inline EVariable const *extC(Variable const *v)
  { return static_cast<EVariable const*>(v); }


// --------------------- ETypeFactory -----------------------
// type factory that yields types extended for the verifier
class ETypeFactory : public BasicTypeFactory {
public:
  // TypeFactory funcs
  virtual CompoundType *makeCompoundType
    (CompoundType::Keyword keyword, StringRef name);
  virtual FunctionType *makeFunctionType
    (Type *retType);
  virtual Variable *makeVariable
    (SourceLoc L, StringRef n, Type *t, DeclFlags f, TranslationUnit *tunit);
};


#endif // EXTTYPE_H
