// prover.cc
// code to run Simplify

#include "prover.h"    // this module
#include "str.h"       // stringc
#include "mypopen.h"   // popen_execvp
#include "unixutil.h"  // readString, writeAll
#include "trace.h"     // trace
#include "xassert.h"   // xassert
#include "exc.h"       // xbase
#include "smregexp.h"  // Regexp
#include "strutil.h"   // replace

#include <stdio.h>     // perror, FILE stuff
#include <string.h>    // strstr
#include <sys/types.h> // open
#include <sys/stat.h>  // open
#include <fcntl.h>     // open
#include <unistd.h>    // close


// main prover object for simple 'runProver' interface
Prover *mainProver = NULL;


bool runProver(rostring str)
{
  if (!mainProver) {
    mainProver = new Prover;
    mainProver->feedFile("bgpred.sx");
    mainProver->openLog("predicate.log");
  }
  
  return mainProver->prove(str);
}


// ------------------- Prover -------------------------
Prover::Prover()
  : toSimplify(-1), fromSimplify(-1),
    predicateLog(-1)
{
  // open Simplify as a child process, with pipes to communicate;
  // this calls exit(2) if it fails...
  static char *argv[] = { "./run-Simplify", /*"-nosc",*/ NULL };
  popen_execvp(&toSimplify, &fromSimplify, &fromSimplify /*stderr==stdout*/,
               argv[0], argv);

  // read initial prompt
  char buf[10];
  if (!readString(fromSimplify, buf, 10)) {    // nul-terminates
    perror("read from Simplify");
    exit(2);
  }

  if (0!=strcmp(buf, ">\t")) {
    fprintf(stderr, "unexpected initial response from Simplify:\n");
    fprintf(stderr, "%s\n", buf);
    exit(2);
  }

  // turn off prompting
  feed("(PROMPT_OFF)\n");
}

Prover::~Prover()
{
  close(toSimplify);
  close(fromSimplify);

  if (predicateLog > -1) {
    close(predicateLog);
  }
}


void Prover::openLog(rostring fname)
{
  xassert(predicateLog == -1);    // otherwise already open!

  predicateLog = creat(toCStr(fname), 0777);
  if (predicateLog < 0) {
    perror(stringc << "creat " << fname);
    exit(2);
  }

  string header = stringc
    << "; " << fname << "     -*- lisp -*-\n"
    << "; prover log file\n"
    << "\n";
  writeAll(predicateLog, toCStr(header), header.length());
}


void Prover::feedFile(rostring fname)
{
  // open file
  FILE *bgpred = fopen(toCStr(fname), "r");
  if (!bgpred) {
    perror(stringc << "open " << fname);
    exit(2);
  }

  // read the background predicate and feed it to Simplify
  enum { BUFSIZE = 4096 };
  char buf[BUFSIZE];
  int len;
  while ((len = fread(buf, 1, BUFSIZE, bgpred)) > 0) {
    if (!writeAll(toSimplify, buf, len)) {
      perror("write file to Simplify");
      exit(2);
    }
  }
  fclose(bgpred);
}


void Prover::feed(rostring str)
{
  if (!writeAll(toSimplify, toCStr(str), strlen(str))) {
    perror("write to Simplify");
    exit(2);
  }

  if (predicateLog > -1) {
    // log the string
    writeAll(predicateLog, toCStr(str), strlen(str));
  }

  // make sure Simplify hasn't replied
  if (canRead(fromSimplify)) {
    fprintf(stderr, "unexpected reply from Simplify:\n");

    // what did it say?
    char *buf = new char[1024];
    if (!readString(fromSimplify, buf, 1024)) {
      perror("read from Simplify");
      exit(2);
    }
    fprintf(stderr, "%s\n", buf);
    delete[] buf;

    exit(2);     // obviates delete[], but whatever
  }
}


bool Prover::prove(rostring str, string *counterexample)
{
  trace("prover") << str << endl;

  // hand the current predicate to Simplify
  if (!writeAll(toSimplify, toCStr(str), strlen(str))) {
    perror("write to Simplify");
    exit(2);
  }

  if (predicateLog > -1) {
    // log predicate; don't care about failure as much
    writeAll(predicateLog, toCStr(str), strlen(str));
  }

  // pattern for responses
  Regexp re(": (Invalid|Valid).");
  Regexp bad("Bad input");

  // read Simplify's response
  stringBuilder response;
  while (!re.match(response)) {
    // read next segment of response
    char buf[200];
    int count = read(fromSimplify, buf, sizeof(buf)-1);
    if (count < 0) {
      perror("read from Simplify");
      exit(2);
    }
    buf[count]=0;

    if (count==0) {
      if (!response.length()) {
        xbase("Simplify closed the pipe\n");
      }
      break;
    }
    response << buf;
    
    if (bad.match(response)) {
      fprintf(stderr, "\nSimplify says: %s\n", buf);
      exit(2);
    }
  }

  // check for the warning about triggerless quantifiers; they are 
  // always an indication that something is wrong
  if (strstr(response, "triggerless quantifier body")) {
    char const *processing = strstr(response, "processing the formula");
    char const *after = strstr(processing, ")\n\n");
    if (processing && after) {
      processing += strlen("processing the formula\n\n");
      string formula = substring(processing, (after+1)-processing);
      xbase(stringc << "triggerless: " << formula);
    }
    else {
      xbase("triggerless (couldn't find formula in Simplify output)\n");
    }
  }

  if (strstr(response, ": Valid.")) {
    return true;           // formula is proven
  }

  if (strstr(response, ": Invalid.")) {
    if (counterexample) {
      // look for counterexample
      char const *counter = strstr(response, "Counterexample:");
      char const *inval = strstr(response, ": Invalid.");
      xassert(inval);   // just found it a moment ago
      if (counter) {
        //counter += strlen("Counterexample:\n");
        while (*inval != '\n') {
          inval--;
        }
        *counterexample = substring(counter, inval-counter);
      }
    }

    return false;          // formula is not proven
  }

  // we get here if Simplify yields a response but it isn't among
  // those expected
  xbase(stringc << "Unexpected response from Simplify:\n  "
                << replace(response, "\n", "\n  ") << "\n");
  return false;            // silence warning
}


// EOF
