// edag.h
// equality dag, central prover data structure

#ifndef EDAG_H
#define EDAG_H

#include "str.h"        // string
#include "sobjlist.h"   // SObjList
#include "objlist.h"    // ObjList
#include "strhash.h"    // StringHash
#include "array.h"      // ObjArrayStack
  

// special type for symbol indices
enum Symbol { NULL_SYMBOL=0 };


// a table of function symbols
class SymbolTable {
private:     // types
  // entry in 'symToIndex'
  class SymbolEntry {
  public:    // data
    string name;          // textual name
    Symbol index;         // index for fast reference
    int arity;            // number of arguments

  public:    // funcs
    SymbolEntry(char const *n, int i, int a);
    ~SymbolEntry();
  };

private:     // data
  // map names to integer indices in the table
  StringHash symToIndex;

  // map indices to strings
  ObjArrayStack<SymbolEntry> indexToSym;

private:     // funcs
  static char const *getTableKey(void *data);

public:      // funcs
  SymbolTable();
  ~SymbolTable();

  // get the index for a symbol string; will re-use an index
  // if this symbol has already been made
  Symbol get(char const *sym, int arity);
  
  // get the string name; the returned pointer becomes invalid
  // once this table is destroyed
  char const *name(Symbol s) const;

  // get a symbol's arity
  int arity(Symbol s) const;

  // number of symbols; this is useful for iterating over
  // all symbols
  int numSymbols() const { return indexToSym.length(); }
};


// a term in the e-dag
class Term {
public:      // data
  // if this is a function-application node, then 'symbol' is the name
  // of the applied function; otherwise it is the name of the variable
  // (or equivalently, nullary (constant) function symbol)
  Symbol symbol;

  // arguments to the function application; possibly empty
  SObjList<Term> args;

  // points toward the representative of this equivalence class; this
  // chain must be followed until it is NULL to find the ultimate
  // representative
  Term *eqclass;               // nullable serf

public:      // funcs
  Term(Symbol s);
  ~Term();

  // return the representative for this term's equivalence class
  Term const *reprC() const;
  Term *repr() { return const_cast<Term*>(reprC()); }

  // true if the two nodes are in the same equivalence class
  bool equals(Term const *t) const;
};


// the e-dag itself
class EDAG {
private:     // data
  // set of all term nodes, indexed by the symbol
  ObjArrayStack< ObjList<Term> > nodes;

public:      // data
  SymbolTable &symTable;

private:     // funcs
  // get the list of nodes for a particular symbol
  ObjList<Term> *getNodeList(Symbol s);

  // helper for 'toDOT'
  string dotNodeName(Term const *t) const;

public:      // funcs
  EDAG(SymbolTable &table);
  ~EDAG();

  // find or create a representative of a term with given symbol
  // and given arguments
  Term *get(Symbol sym, SObjList<Term> const &args);
  Term *get0(Symbol sym);
  Term *get1(Symbol sym, Term *arg1);

  // find a representative; return NULL if it is not already
  // represented
  Term *find(Symbol sym, SObjList<Term> const &args);
  Term *find0(Symbol sym);
  Term *find1(Symbol sym, Term *arg1);

  // unify two given terms so they are equivalent
  void unify(Term *t1, Term *t2);

  // close under congruence
  void congruenceClosure();

  // obtain a list of all the nodes in the EDAG
  void getNodeList(ArrayStack<Term*> &dest);

  // yield the graph in DOT format
  string toDOT(char const *graphName);
};


#endif // EDAG_H
