// rprove.cc
// rewriting-based prover

#include "rprove.gr.gen.h"       // goal parser
#include "mysig.h"               // printSegfaultAddrs
#include "glr.h"                 // GLR
#include "cc_lang.h"             // CCLang
#include "lexer.h"               // Lexer
#include "owner.h"               // Owner
#include "exc.h"                 // xbase
#include "trace.h"               // TRACE_ARGS
#include "srcloc.h"              // SourceLoc


TopForm::~TopForm()
{
  if (kind == TF_AXIOM) {
    if (ax) {
      delete ax;
    }
  }
  else {
    if (eq) {
      delete eq;
    }
  }
}


TopFormSeq *parseFile(StringTable &strTable, char const *fname)
{
  // lexer
  CCLang lang;
  Lexer lexer(strTable, lang, fname);

  // parser context
  RProveParse ctx(strTable);

  // parse tables (automatically deallocated)
  Owner<ParseTables> tables(ctx.makeTables());

  // parser
  GLR glr(&ctx, tables);

  // parse the input
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    xbase(stringc << "parse error in file " << fname);
  }

  return (TopFormSeq*)treeTop;
}


int doit(int argc, char **argv)
{
  printSegfaultAddrs();
  StringTable strTable;
  SourceLocManager mgr;

  // process arguments
  char const *progname = argv[0];
  TRACE_ARGS();

  if (tracingSys("all")) {
    traceAddSys("reduce");
    traceAddSys("reduceTerm");
    traceAddSys("axiom");
    traceAddSys("mix");
  }

  if (argc < 2) {
    cout << "usage: " << progname << " [-tr flags] input.rp\n";
    return 0;
  }

  // parse
  Owner<TopFormSeq> seq; seq = parseFile(strTable, argv[1]);

  // build the rewrite engine state
  Rewriter r;
  while (seq->forms.isNotEmpty()) {
    TopForm *tf = seq->forms.removeFirst();
    if (tf->kind == TopForm::TF_AXIOM) {
      Axiom *ax = tf->ax;
      tf->ax = NULL;

      // check orientation requirements
      ax->orient();
      ax->checkPremiseOrder();
      r.axioms.append(ax);
    }
    else {
      Equality *eq = tf->eq;
      tf->eq = NULL;

      eq->orient();
      if (tf->kind == TopForm::TF_EQUALITY) {
        r.equalities.append(eq);
      }
      else {
        r.disequalities.append(eq);
      }
    }

    delete tf;
  }

  // evaluate to quiescence
  do {
    do {
      r.print("current system");
      cout << "reducing...\n";
    } while (r.reduce());

    r.print("current system");
    cout << "mixing...\n";
  } while (r.mix());

  // report
  r.print("final system");
  if (r.contradiction) {
    cout << "system is NOT satisfiable\n";
  }
  else {
    cout << "system IS satisfiable\n";
  }

  return 0;
}


int main(int argc, char **argv)
{
  int ret;
  try {
    ret = doit(argc, argv);
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
  return ret;
}
