// btree.c
// btree example from btree.cc, but written in the
// language my verifier can handle right now


// ------------------- preamble ----------------
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"minmax.vc\";");

// C functions that are modeled directly
void *malloc(int size);

// declarations of model functions
int sel(int mem, int p);
int min(int a, int b);
int max(int a, int b);

int tag;
int objct;
int objsize;


// ---------------- code to verify -------------
typedef struct Node Node;
typedef struct Interior Interior;
typedef struct Leaf Leaf;
typedef struct TreeHeader TreeHeader;


#define AND all_of
#define OR one_of


typedef int bool;
#define false 0
#define true 1

// node in the btree
struct Node {
  // parent pointer, specifically identifying whether this is
  // the left child or the right child
  Interior *parent;      // (ghost) who points at me from above
  bool isLeft;           // (ghost) true if I'm the left child of 'parent'
  bool isRoot;           // (ghost) true if I'm the very topmost node
  int level;             // (ghost) leaves are 0, interiors on them are 1, etc.
  bool isLeaf;           // type of this node: true=Leaf, false=Interior
  TreeHeader *header;    // (ghost) tree-as-a-whole designator
  int lowKey, highKey;   // (ghost) tight inclusive bounds on keys stored in the tree
};


struct Interior {
  // fake inheritance
  Node *superclass;
  Interior *parent;
  bool isLeft;
  bool isRoot;
  int level;
  bool isLeaf;
  TreeHeader *header;   
  int lowKey, highKey;

  // binary tree links; a real B+-tree has an array of links,
  // interspersed with key values
  Node *left;
  int ikey;              // interior key; separates left and right subtrees
  Node *right;           // can be NULL

  // left/right breakdown:
  //   left: everything <= ikey
  //   right: everything > ikey

  #define newInterior(ret, hdr, lvl)               \
  {                                                \
    (ret) = (Interior*)malloc(sizeof(Interior));   \
    (ret)->parent = NULL;                          \
    (ret)->isRoot = false;                         \
    (ret)->level = (lvl);                          \
    (ret)->isLeaf = false;                         \
    (ret)->header = (hdr);                         \
    /* caller should set 'left' */                 \
    (ret)->right = NULL;                           \
  }

  #define setLeft(ths, L)   \
  {                         \
    (ths)->left = (L);      \
    (L)->parent = (ths);    \
    (L)->isLeft = true;     \
  }

  #define setRight(ths, R)  \
  {                         \
    (ths)->right = (R);     \
    (R)->parent = (ths);    \
    (R)->isLeft = false;    \
  }
};


struct Leaf {
  // fake inheritance
  Node *superclass;
  Interior *parent;
  bool isLeft;
  bool isRoot;
  int level;
  bool isLeaf;
  TreeHeader *header;
  int lowKey, highKey;

  // stored data; a real B+-tree leaf node contains an array of
  // key/data pairs
  int key;
  int data;

  // linked list, acyclic; made doubly-linked with ghost var 'prev'
  // (though some implementations actually materialize 'prev')
  Leaf *next;
  Leaf *prev;            // ghost

  #define newLeaf(ret, hdr)               \
  {                                       \
    (ret) = (Leaf*)malloc(sizeof(Leaf));  \
    (ret)->parent = NULL;                 \
    (ret)->isRoot = false;                \
    (ret)->level = 0;                     \
    (ret)->isLeaf = true;                 \
    (ret)->header = (hdr);                \
    (ret)->next = NULL;                   \
    (ret)->prev = NULL;                   \
  }

  #define setNext(ths, n)  \
  {                        \
    (ths)->next = (n);     \
    if (n) {               \
      (n)->prev = (ths);   \
    }                      \
  }
  
  // no 'setPrev' since I want 'prev' to be a ghost field
};
  

struct TreeHeader {
  // root of the tree; NULL iff the tree is empty
  Node *root;
  
  // unique leaf node that has a NULL 'prev' field, or NULL
  // if the tree is empty
  Leaf *firstLeaf;
};



#define hasNodeTag(p) (tag[p]==Interior_tag || tag[p]==Leaf_tag)

// global invariant
int global_inv() { return vc_expression(AND {
  0 < objct;
  tag[0] == 0;

  // -- Node --

  // not '<==>' because during insertion I create nodes that have
  // no parent but are not destined to become the ultimate root
  forall(Node *n) PATS(n->parent). (tag[n]==Interior_tag || tag[n]==Leaf_tag) ==>
    n->isRoot ==> n->parent == NULL;

  // level is always nonnegative
  forall(Node *n) PATS(n->level). (tag[n]==Interior_tag || tag[n]==Leaf_tag) ==>
    n->level >= 0;

  // level 0 are the leaves
  forall(Node *n) PATS(n->level). (tag[n]==Interior_tag || tag[n]==Leaf_tag) ==>
    n->level==0 <==> n->isLeaf;

  // -- Interior --

  forall(Interior *i) PATS(i->isLeaf). tag[i]==Interior_tag ==>
    i->isLeaf == false;

  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    0 < i->left < objct && hasNodeTag(i->left);                           // disequality
  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->left->parent==i && i->left->isLeft==true;                          // injectivity
  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->left->header == i->header;                                         // transitivity
  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->left->isRoot == false;
  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->left->level+1 == i->level;                                         // grid

  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->lowKey <= i->left->lowKey;                                         // order

  forall(Interior *i) PATS(i->right). tag[i]==Interior_tag ==>
    i->right!=NULL ==> AND {
      0 < i->right < objct && hasNodeTag(i->right);
      i->right->parent==i && i->right->isLeft==false;                     // injectivity
      i->right->header == i->header;                                      // trans
      i->right->isRoot == false;
      i->right->level+1 == i->level;                                      // grid

      i->left->highKey <= i->ikey <= i->right->lowKey;                    // order*2
      i->right->highKey <= i->highKey;                                    // order
    };

  forall(Interior *i) PATS(i->left). tag[i]==Interior_tag ==>
    i->right==NULL ==> AND {
      i->left->highKey <= i->highKey;                                     // order
    };

  forall(Interior *i). tag[i]==Interior_tag ==>
    i->lowKey <= i->highKey;


  // -- Leaf --

  forall(Leaf *l) PATS(l->isLeaf). tag[l]==Leaf_tag ==>
    l->isLeaf==true;

  forall(Leaf *l) PATS(l->next). tag[l]==Leaf_tag ==>
    l->next!=NULL ==> AND {
      0 < l->next < objct && tag[l->next]==Leaf_tag;
      l->next->prev == l;                                                 // inj
      l->next->header == l->header;                                       // trans

      // this assures that all the leaves are in the right order
      l->highKey <= l->next->lowKey;                                      // order
    };

  forall(Leaf *l) PATS(l->prev). tag[l]==Leaf_tag ==>
    l->prev!=NULL ==> AND {
      0 < l->prev < objct && tag[l->prev]==Leaf_tag;
      l->prev->next == l;                                                 // inj
      l->prev->header == l->header;                                       // trans
    };

  // if I appear to be first in the list, then I'm the *only*
  // such node
  forall(Leaf *l) PATS(l->header). tag[l]==Leaf_tag ==>
    l->prev==NULL ==> l->header->firstLeaf==l;                            // inj

  // bounds are accurate and tight
  forall(Leaf *l) PATS(l->lowKey, l->key). tag[l]==Leaf_tag ==>
    l->lowKey == l->key;
  forall(Leaf *l) PATS(l->highKey, l->key). tag[l]==Leaf_tag ==>
    l->highKey == l->key;


  // -- TreeHeader --

  forall(TreeHeader *th) PATS(th->root, th->firstLeaf).
    tag[th]==TreeHeader_tag ==>
      (th->root == NULL && th->firstLeaf == NULL) ||
      (th->root != NULL && th->firstLeaf != NULL);                        // disequality

  forall(TreeHeader *th) PATS(th->root).
    tag[th]==TreeHeader_tag && th->root!=NULL ==> AND {
      0 < th->root < objct && hasNodeTag(th->root);
      th->root->header == th;                                             // inj
      th->root->isRoot == true;
    };

  forall(TreeHeader *th) PATS(th->firstLeaf).
    tag[th]==TreeHeader_tag && th->firstLeaf!=NULL ==> AND {
      0 < th->firstLeaf < objct && tag[th->firstLeaf]==Leaf_tag;
      th->firstLeaf->header == th;                                        // inj
      th->firstLeaf->prev == NULL;
    };

});}


// return the data associated with the given key
int find(Node *root, int key)
  pre(root==NULL || hasNodeTag(root))
{
  if (!root) {
    return -1;                        // not found
  }

  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;         // downcast
    if (leaf->key == key) {
      return leaf->data;              // found
    }
    else {
      return -1;                      // not found
    }
  }

  else {
    Interior *i = (Interior*)root;    // downcast
    if (key <= i->ikey) {
      return find(i->left, key);      // tail recursion
    }
    else {
      return find(i->right, key);     // tail recursion
    }
  }
}


#define PARAMBYREF __varattribute__((parambyref))
#define ARGBYREF __varattribute__((argbyref))

// insert a new key,data pair; if this returns non-NULL, then
// the tree was split, and 'key' and 'result' provide the
// other two values to make up the split triple: left/key/right
// as root/(*key)/result
Node *innerInsert(Node *root, int *key PARAMBYREF, int data)
  pre(AND {
    0 < root < objct && hasNodeTag(root);
                                               
    // make sure the parent selected the right child for this key
    // TODO: make syntax so this can be 'if...'
    root->parent!=NULL && tag[root->parent]==Interior_tag ==> AND {
      // *key must be within parent's range
      root->parent->lowKey <= *key <= root->parent->highKey;
                     
      root->parent->right != NULL ==> AND {
        // not an only child, parent's 'ikey' also matters
        root->isLeft==true ==> *key <= root->parent->ikey;
        root->isLeft!=true ==> root->parent->ikey <= *key;
      };
    };
  })

  post(AND {
    // the root pointer is still good (some of this will be redundant
    // once I've got level-based change isolation installed)
    root == pre(root);
    0 < root < objct && hasNodeTag(root);
    root->header == pre(root->header);
    root->level == pre(root->level);

    // if result is not NULL, it's similar to 'root' in many respects
    result==NULL || AND {
      0 < result < objct && hasNodeTag(result);
      result != root;
      result->parent == NULL;
      result->isRoot == false;
      result->header == root->header;
      result->level == root->level;

      // yielded nodes are divided by final *key value, and have
      // their ranges expanded at most to include original *key
      root->lowKey >= pre(min(*key, root->lowKey));
      root->highKey <= *key <= result->lowKey;
      result->highKey <= pre(max(*key, root->highKey));
    };

    // nothing in levels higher than 'root' has been modified
    forall(Interior *i).
      0 < i < pre(objct) &&
      pre(i->level) > pre(root->level)
    ==> AND {
      pre(i->left) == i->left;
      pre(i->right) == i->right;
      pre(i->lowKey) == i->lowKey;
      pre(i->ikey) == i->ikey;
      pre(i->highKey) == i->highKey;
    };

    // nothing deallocated
    forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];

    // objct monotonic
    pre(objct) <= objct;
  })
  changes(Node_tag, Interior_tag, Leaf_tag, objct, tag, objsize)
{
  int origKey = *key;

  if (root->isLeaf) {
    // always split

    Leaf *leftHalfLeaf = (Leaf*)root;

    Leaf *rightHalfLeaf;
    newLeaf(rightHalfLeaf, root->header);

    // I need that my neighbors' key edges bound the inserted key;
    // this is true, but appears somewhat difficult to prove
    assume(leftHalfLeaf->prev!=NULL ==>
      leftHalfLeaf->prev->highKey <= *key);
    assume(leftHalfLeaf->next!=NULL ==>
      *key <= leftHalfLeaf->next->lowKey);

    Leaf *origNext = leftHalfLeaf->next;
    assert(rightHalfLeaf != origNext);
    //assume(leftHalfLeaf != origNext);      // this is irrelevant now

    // put the existing and new keys into the proper leaves
    if (leftHalfLeaf->key <= *key) {
      // *key goes into the right half
      rightHalfLeaf->key = *key;
      rightHalfLeaf->data = data;
      rightHalfLeaf->lowKey = *key;
      rightHalfLeaf->highKey = *key;
    }
    else {
      // *key goes into the left half, and existing key goes into right
      rightHalfLeaf->key = leftHalfLeaf->key;
      rightHalfLeaf->data = leftHalfLeaf->data;
      rightHalfLeaf->lowKey = leftHalfLeaf->lowKey;
      rightHalfLeaf->highKey = leftHalfLeaf->highKey;

      leftHalfLeaf->key = *key;
      leftHalfLeaf->data = data;
      leftHalfLeaf->lowKey = *key;
      leftHalfLeaf->highKey = *key;

      // that RHS key is what separates the yielded nodes
      *key = rightHalfLeaf->key;
    }

    // splice 'rightHalfLeaf' into the leaf list
    setNext(rightHalfLeaf, leftHalfLeaf->next);
    setNext(leftHalfLeaf, rightHalfLeaf);

    // return the new leaf; 'key' has already been set
    return rightHalfLeaf;
  }

  else {
    Interior *intr = (Interior*)root;
    Node *rightHalfNode;

    if (*key < intr->lowKey) {
      intr->lowKey = *key;
    }
    else if (*key > intr->highKey) {
      intr->highKey = *key;
    }
    assert(intr->lowKey <= *key <= intr->highKey);

    if (!intr->right) {
      // only one child
      Node *passedRoot = intr->left;
      rightHalfNode = innerInsert(intr->left, key, data);
      if (!rightHalfNode) {
        return NULL;          // no split
      }
      else {
        assert(intr->left == passedRoot);    // change isolation

        // node below me split, but split stops here because I
        // have room to accomodate it
        intr->ikey = *key;
        setRight(intr, rightHalfNode);

        return NULL;          // split does not propagate
      }
    }

    else {
      // two children, must pick one
      Node *chosen = (*key <= intr->ikey? intr->left : intr->right);
      rightHalfNode = innerInsert(chosen, key, data);
      if (!rightHalfNode) {
        return NULL;          // no split
      }
      else {
        // node below me split, I have to split also; I now have three
        // data pointers and two keys; strategy will be to yield the
        // existing node, plus a singleton node holding the extreme
        // half, as the split pair
          
        // will hold a pointer to the new node
        Interior *single;

        if (chosen == intr->left) {
          assert(intr->right != NULL);      // change isolation

          // I have:
          //
          //               *key               intr->ikey
          //              /  B \             /    D     \
          //    intr->left      rightHalfNode            intr->right
          //        A                 C                      E

          int B = *key;

          // I want:                              yield    yield
          //                                        |        |
          //                                        |        |
          //               intr->ikey             *key       |
          //              /    B     \           /  D \      |
          //    intr->left            intr->right      single->left
          //        A                     C                  E

          newInterior(single, root->header, intr->level);
          setLeft(single, intr->right);      // E
          *key = intr->ikey;                 // D
          setRight(intr, rightHalfNode);     // C
          intr->ikey = B;                    // B
          // nothing to do                   // A

          // set single's bounds (one child)
          single->lowKey = single->left->lowKey;
          single->highKey = single->left->highKey;

          // tighten my bounds
          intr->highKey = intr->right->highKey;

          return single;
        }

        else {
          assert(chosen == intr->right);     // change isolation

          // I have:
          //
          //               intr->ikey             *key
          //              /    B     \           /  D \
          //    intr->left            intr->right      rightHalfNode
          //        A                     C                  E

          // simply wrap up E as a new node, and yield D (which requires
          // no additional action); A,B,C are already in good shape

          newInterior(single, root->header, intr->level);
          setLeft(single, rightHalfNode);

          // set single's bounds (one child)
          single->lowKey = single->left->lowKey;
          single->highKey = single->left->highKey;

          // tighten my bounds
          intr->highKey = intr->right->highKey;

          return single;
        }
      }
    }
  }
}


// outer shell of insert routine
void insert(TreeHeader *header, int key ARGBYREF, int data)
  pre(0 < header < objct && tag[header]==TreeHeader_tag)
  post(AND {
    forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];
    pre(objct) <= objct;
  })
  changes(Node_tag, Interior_tag, Leaf_tag, TreeHeader_tag, objct, tag, objsize)
{
  if (!header->root) {
    Leaf *leaf;
    newLeaf(leaf, header);
    leaf->isRoot = true;
    leaf->key = key;
    leaf->data = data;
    leaf->lowKey = key;
    leaf->highKey = key;

    header->root = leaf;
    header->firstLeaf = leaf;
  }

  else {
    Node *rightHalf = innerInsert(header->root, &key, data);
    if (rightHalf) {
      // make new root node
      Interior *intr;
      newInterior(intr, header, header->root->level + 1);

      setLeft(intr, header->root);
      intr->ikey = key;
      setRight(intr, rightHalf);
                                         
      // change who is the ultimate root
      intr->isRoot = true;
      header->root->isRoot = false;
      header->root = intr;
      
      // set bounds (two children)
      intr->lowKey = intr->left->lowKey;
      intr->highKey = intr->right->highKey;
    }
  }
}


#if 0
// delete 'key' from subtree rooted at 'root'; return value:
//   -1: this subtree is now empty, delete it
//   0: everything fine, nothing more to do
//   otherwise: new rightmost key for this subtree
int innerDelete(Node *root, int key)
{
  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;

    if (leaf->key == key) {
      // this now is empty
      return -1;
    }
    else {
      // this key isn't present, nothing more to do
      return 0;
    }
  }

  else {
    Interior *intr = (Interior*)root;

    if (!intr->right) {
      // only one child
      key = innerDelete(intr->left, key);
      if (key == 0) {
        return 0;       // done
      }
      if (key == -1) {
        free(intr->left);

        // HERE: need a (ghost) global to mark the node I just
        // deleted, since it's going to be an exception to my
        // invariant


    if (key <= intr->ikey

#endif // 0
