// unify.cc
// code for unify.h

// based on Peter Norvig's article
// "Correcting A Widespread Error in Unification Algorithms"
// http://www.norvig.com/unify-bug.pdf

#include "unify.h"      // this module
#include "term.h"       // Term
#include "array.h"      // GrowArray


Unify::Unify()
  : subst(),
    matchMode(false)
{}

Unify::~Unify()
{}


bool Unify::unify(Term const *x, Term const *y)
{
  if (x->args.isNotEmpty() && y->args.isNotEmpty()) {
    // two function applications
    if (x->name != y->name) {
      // unequal names, can't unify
      return false;
    }

    // unify the argument lists
    ObjListIter<Term> xIter(x->args);
    ObjListIter<Term> yIter(y->args);
    for (; !xIter.isDone() && !yIter.isDone(); xIter.adv(), yIter.adv()) {
      if (!unify(xIter.data(), yIter.data())) {
        return false;     // arguments fail to unify
      }
    }

    // success if the lists had same length
    return xIter.isDone() && yIter.isDone();
  }

  if (x->args.isEmpty() &&
      y->args.isEmpty() &&
      x->name == y->name) {
    // identical names, unconditionally unified
    return true;
  }

  // at this point, we know that 'x' and 'y' are not 'equals',
  // in the Lisp structural equality sense

  if (isVar(x)) {
    return unifyVariable(x->name, y);
  }
  if (!matchMode && isVar(y)) {
    return unifyVariable(y->name, x);
  }

  // terms are not 'equal', and neither is a unification variable,
  // so they do not unify

  return false;
}


// Norvig calls 't' "val", but that name is too similar to 'var' IMO
bool Unify::unifyVariable(StringRef var, Term const *t)
{
  if (!matchMode &&
      isVar(t) &&
      t->name == var) {
    // same unification variable
    return true;
  }

  Term const *binding = subst.get(var);
  if (binding /*bound(var)*/) {
    return unify(binding, t);
  }

  // this is the check that Norvig's article is about
  if (!matchMode && isVar(t)) {
    binding = subst.get(t->name);
    if (binding /*bound(t->name)*/) {
      return unifyVariable(var, lookup(t->name));
    }
  }

  if (!matchMode && occurs(var, t)) {
    // since 'var' occurs in 't', there is no finite substitution term
    // that would unify the original terms (e.g. unify "f(x)" and "x"
    // is satisfied if x = "f(f(f(...)))" but no finite term works)
    return false;
  }

  // extend the current substitution
  extendSubst(var, t);

  return true;
}


bool Unify::occurs(StringRef var, Term const *t)
{
  xassert(!matchMode);

  if (isVar(t)) {
    if (t->name == var) {
      return true;     // found the variable
    }

    Term const *binding = subst.get(t->name);
    if (binding) {     // bound?
      return occurs(var, binding);
    }

    return false;
  }

  // examine arguments for occurrences
  FOREACH_OBJLIST(Term, t->args, iter) {
    if (occurs(var, iter.data())) {
      return true;
    }
  }

  // no occurrences
  return false;
}


bool Unify::isVar(Term const *t) const
{
  return t->isVar();
}


bool Unify::bound(StringRef var) const
{
  return subst.get(var) != NULL;
}


Term const *Unify::lookup(StringRef var) const
{
  return subst.get(var);
}


void Unify::extendSubst(StringRef var, Term const *t)
{
  subst.add(var, t);
}


void Unify::reset()
{
  subst.empty();
}


string Unify::termSexp(Term const *t) const
{
  stringBuilder sb;
  itermSexp(sb, t);
  return sb;
}

void Unify::itermSexp(stringBuilder &sb, Term const *t) const
{
  if (isVar(t) && bound(t->name)) {
    // replace a bound variable with the thing to which it is bound;
    // this is how this function is different from simply 't->asString()'
    itermSexp(sb, lookup(t->name));
  }
  else if (t->args.isEmpty()) {
    // just a name
    sb << t->name;
  }
  else {
    // function symbol
    sb << "(" << t->name;
    FOREACH_OBJLIST(Term, t->args, iter) {
      sb << " ";
      itermSexp(sb, iter.data());
    }
    sb << ")";
  }
}


Term *Unify::substitute(Term const *t) const
{
  if (isVar(t) && bound(t->name)) {
    // replace a bound variable with the thing to which it is bound;
    // but recursively substitute
    return substitute(lookup(t->name));
  }

  // clone the term
  Term *ret = new Term(t->name);
  FOREACH_OBJLIST(Term, t->args, iter) {
    ret->args.prepend(substitute(iter.data()));
  }
  ret->args.reverse();

  return ret;
}


bool unifies(Term const *x, Term const *y)
{
  Unify u;
  return u.unify(x, y);
}


// --------------------------- test code -----------------------------
#ifdef TEST_UNIFY

#include <stdlib.h>     // qsort

void printUnify(Term *x, Term *y)
{
  cout << "unifying " << x->asString() << " and " << y->asString() << endl;

  Unify u;
  if (u.unify(x, y)) {
    PtrMap<char const, Term const>::Iter iter(u.subst);
    for (; !iter.isDone(); iter.adv()) {
      cout << "  " << iter.key() << " = " << iter.value()->asString() << "\n";
    }
    string xStr = u.termSexp(x);
    string yStr = u.termSexp(y);
    cout << "  xStr: " << xStr << endl;
    cout << "  yStr: " << yStr << endl;
    cout << "  success" << endl;
  }
  else {
    cout << "  failure" << endl;
  }
}


static int qsort_strcmp(void const *s, void const *t)
{
  char const * const *ss = (char const * const *)s;
  char const * const *tt = (char const * const *)t;
  return strcmp(*ss, *tt);
}


// if 'subst' is NULL, then 'x' and 'y' are not expected to unify
void testUnify(Term *x, Term *y, char const *subst)
{
  cout << "unifying " << x->asString() << " and " << y->asString() << endl;

  Unify u;
  bool success = u.unify(x, y);

  // put the actual unifying bindings into 'actual'
  stringBuilder actual;
  if (success) {
    // get (a pointer to) all the names in an array, so we
    // can sort it to get a canonical binding list
    GrowArray<StringRef> names(u.subst.getNumEntries());
    PtrMap<char const, Term const>::Iter iter(u.subst);
    int ct = 0;
    for (; !iter.isDone(); iter.adv()) {
      names[ct++] = iter.key();
    }
    xassert(ct == names.size());

    // sort by name
    qsort(names.getArrayNC(), ct, sizeof(char const*), qsort_strcmp);

    // create the binding list in sorted order; I consider this important
    // because otherwise the order is dependent on the internal details
    // of StringSObjDict, and I don't want my testing code to be dependent
    // on such an implementation detail
    for (int i=0; i<ct; i++) {
      if (i > 0) {
        actual << ", ";
      }
      actual << names[i] << " = " << u.lookup(names[i])->asString();
    }
  }

  // did we get what we expected?
  if (success != !!subst ||                    // unexpected (non)unification
      success && !actual.equals(subst)) {      // wrong substitution
    cout << "FAILURE to obtain expected result\n";
    
    cout << "  expected: ";
    if (subst) {
      cout << (*subst? subst : "(unify with empty substitution");
    }
    else {
      cout << "(failure to unify)";
    }
    cout << "\n";
    
    cout << "    actual: ";
    if (success) {
      if (actual.length()) {
        cout << actual;
      }
      else {
        cout << "(unify with empty substitution)";
      }
    }
    else {
      cout << "(failure to unify)";
    }
    cout << "\n";

    exit(4);
  }

  if (success) {
    // check that toSexp under substitutions yield identical strings
    string xStr = u.termSexp(x);
    string yStr = u.termSexp(y);
    if (!xStr.equals(yStr)) {
      cout << "  FAILED to find same termSexps:\n";
      cout << "  xStr: " << xStr << endl;
      cout << "  yStr: " << yStr << endl;
      exit(4);
    }
  }

  cout << "  " << (success? "ok" : "failed as expected") << "\n";
}


void testUnifyStr(char const *x, char const *y, char const *subst)
{
  return testUnify(Term::parse(x), Term::parse(y), subst);
}


int main()
{
  StringTable str;
  Term::parseStrTable = &str;

  Term *x = Term::parse("X");
  Term *a = Term::parse("a");
  Term *fab = Term::parse("(f a b)");
  Term *gXYa = Term::parse("(g X Y a)");
  Term *gYXX = Term::parse("(g Y X X)");

  printUnify(x, a);
  printUnify(x, fab);
  printUnify(a, fab);
  printUnify(gXYa, gYXX);      // example from Norvig

  testUnify(x, a, "X = a");
  testUnify(x, fab, "X = (f a b)");
  testUnify(a, fab, NULL);
  testUnify(gXYa, gYXX, "X = Y, Y = a");

  testUnifyStr("a", "b", NULL);
  testUnifyStr("(f X Y)", "(f (g 1) (h c (d 2)))", "X = (g 1), Y = (h c (d 2))");
  testUnifyStr("(f X X)", "(f (g 1) (h c (d 2)))", NULL);

  testUnifyStr("(f X (j 2))", "(f (k 3) Y)",
               "X = (k 3), Y = (j 2)");

  // another tricky Norvig example
  testUnifyStr("(q (p X Y) (p Y X))", "(q Z Z)",
               "X = Y, Z = (p X Y)");

  testUnifyStr("(q (p X Y 3) (p Y X 4))", "(q Z Z)",
               NULL);

  return 0;
}

#endif // TEST_UNIFY
