// eval.cc
// code for eval() functions declared in vclang.ast

#include "vclang.ast.gen.h"       // VCLang AST
#include "vc_env.h"               // VCEnv
#include "strutil.h"              // parseQuotedString
#include "breaker.h"              // breaker
#include "smregexp.h"             // regexpMatch
#include "nonport.h"              // getMilliseconds


// ---------------------- VCSequence ----------------------
void VCSequence::eval(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Statement, stmts, iter) {
    iter.data()->eval(env);
  }
}


// ---------------------- Statement ----------------------
void S_include::eval(VCEnv &env)
{
  if (!env.included.isMapped(file.str)) {
    cout << "including: " << file.str << endl;

    // parse the included file
    VCSequence *seq = env.parseVCLang(parseQuotedString(file.str));

    // add it to the map; this causes it to eventually
    // be deallocated
    env.included.add(file.str, seq);

    // eval it
    env << "; ------ start include: " << file.str << " ------\n";
    seq->eval(env);
    env << "; ------ end include: " << file.str << " ------\n";
  }
}

void S_define::eval(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Macro, macros, iter) {
    iter.data()->eval(env);
  }
}


void S_var::eval(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Binding, names->names, iter) {
    Binding &b = *(iter.data());
    if (b.hasDefault()) {
      env.err(b.name.loc, "variables can't have default values");
    }

    // add initial (unconstrained) binding: name + "0"
    env << "; " << b.name.str << ": " << b.name.str << "0\n";
    Variable *v = new Variable(Variable::K_PROGRAM, b.name);
    v->index = 0;
    env.add(v);
  }
}


void S_freshen::eval(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Binding, names->names, iter) {
    Binding &b = *(iter.data());
    if (b.hasDefault()) {
      env.err(b.name.loc, "freshened variables can't have default values");
    }

    // find binding
    Variable *v = env.lookup(b.name.str);
    
    // the above call is sort of a bug, because I should have called
    // 'lookupForModification'; however I'm now going to take advantage
    // of the bug, by using freshen and intending that its effects are
    // not confined to the block in which it lives.. (HACK!)

    if (!v || v->kind != Variable::K_PROGRAM) {
      env.err(b.name.loc, stringc << "undeclared variable: " << b.name.str);
    }

    // bind to fresh variable
    v->index++;
    env << "; " << b.name.str << ": " << b.name.str << v->index << "\n";
  }
}


void S_introduce::eval(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Binding, names->names, iter) {
    LocString &s = iter.data()->name;
    env.introduce(s);
  }
}


// candidate for strutil module
string strmult(char const *cell, int n)
{
  stringBuilder sb;
  while (n-- > 0) {
    sb << cell;
  }
  return sb;
}


void S_block::eval(VCEnv &env)
{
  string n = parseQuotedString(name.str);

  if (regexpMatch(n, "^skip:")) {
    env << "; skipping block: " << n << "\n";
    return;
  }

  string indent = strmult("--", env.scopeDepth());

  env << "; ------ start block: " << n << " ------\n";
  VCScope *scope = env.enterScope(n.c_str());

  bool skipping = env.skippingThisBlock();
  if (!skipping) {
    cout << indent << " " << n << " {\n";
    //cout << "------ " << n << " ------\n";
  }

  body->eval(env);
                                                        
  // do this before 'exitScope', since that deletes 'scope'!
  long elapsed = getMilliseconds() - scope->creationTime;

  env.exitScope();
  env << "; ------ end block: " << n << " ------\n\n\n";

  if (!skipping) {
    cout << indent << " } (" << elapsed << " ms)\n";
  }
}


void S_assume::eval(VCEnv &env)
{
  env.assume(pred);
}


void S_prove::eval(VCEnv &env)
{
  env.proveExpr(label.str, pred);
}

void S_notprove::eval(VCEnv &env)
{
  env.prove(parseQuotedString(label.str), env.evalToSexp(pred, true /*pred*/), 
            false /*expect*/);
}


void S_lemma::eval(VCEnv &env)
{
  env.proveExpr(label.str, pred);
  env.assume(pred);
}


void S_assign::eval(VCEnv &env)
{
  Variable *v = env.lookupForModification(name.str);
  if (!v || v->kind != Variable::K_PROGRAM) {
    env.err(name.loc, stringc << "not a variable: " << name.str);
  }

  env << "; " << name.str << ": " << name.str << (v->index + 1) << "\n";
  env << "(BG_PUSH (EQ " << name.str << (v->index + 1) << "\n  ";
  env << env.evalToSexp(value, false /*pred*/);
  env << "))\n";
  env.scope().pushes++;

  // update index
  v->index++;
}


void S_checkConsistency::eval(VCEnv &env)
{
  env.prove("consistency check", "FALSE", false /*expect*/);
}


void S_defpred::eval(VCEnv &env)
{
  // add a scope for the self-bindings of the predicate params
  env.enterScope();

  env << "(DEFPRED (" << name.str;

  FOREACH_ASTLIST_NC(Binding, params->names, iter) {
    Binding &b = *(iter.data());
    if (b.hasDefault()) {
      env.err(b.name.loc, "defpred params can't have default values");
    }

    env << " " << b.name.str;

    // add a self-binding
    env.add(new Variable(Variable::K_SELF, b.name));
  }

  env << ")";

  if (body) {
    env << " " << env.evalToSexp(body, true /*pred*/);
  }

  env << ")\n\n";

  env.exitScope();
                      
  // now that the predicate is defined for Simplify, I no longer
  // interpret it
  env.introduce(name);
}


void S_if::eval(VCEnv &env)
{
  env << "; ------ if: " << env.toSexp(pred, true /*pred*/) << " ------\n";
  env.enterScope("if");

  // assume the guard
  env.assume(pred);

  // evaluate the body
  body->eval(env);

  env.exitScope();
  env << "; ------ end if ------\n\n\n";
}


// ---------------------- Macro ----------------------
void Macro::eval(VCEnv &env)
{
  if (params) {
    // emit this to help correlate the output with the input
    env << "; macro defined: " << name.str << "\n";

    Variable *v = new Variable(Variable::K_MACRO, name);
    v->macro = this;
    env.add(v);
  }
  else {
    env << "; constant defined: " << name.str << "\n";
                                                
    // not really a parameter, but I want to evaluate it now, so
    // I'll call it a K_PARAM anyway
    Variable *v = new Variable(Variable::K_PARAM, name);
    v->param = body->expand(env);
    env.add(v);
  }
}


// ---------------------- Expression ----------------------
void Expression::eval(VCEnv &env, bool pred)
{
  Expression *tmp = expand(env);
  tmp->tosx(env, pred);
  delete tmp;
}


// ---------------------- Expression::expand ----------------------
Expression *E_var::expand(VCEnv &env)
{
  Variable *v = env.lookup(name.str);
  if (!v) {
    env.err(name.loc, stringc << "undeclared: " << name.str);
  }

  switch (v->kind) {
    default: xfailure("bad kind code");

    // variable?
    case Variable::K_PROGRAM:
      return new E_var(new LocString(name.loc, env.str(stringc << name.str << v->index)));

    // macro?
    case Variable::K_MACRO:
      env.err(name.loc, stringc << "macro " << name.str << " requires arguments");

    // constant?
    case Variable::K_CONST:
      return v->expr->expand(env);

    // self-map?
    case Variable::K_SELF:
      return clone();

    // param?
    case Variable::K_PARAM:
      return v->param->clone();
  }
}


Expression *E_apply::expand(VCEnv &env)
{
  Variable *v = env.lookup(name.str);
  if (!v) {
    env.err(name.loc, stringc << "undeclared: " << name.str);
  }

  switch (v->kind) {
    default: xfailure("bad kind code");

    // variable or constant?
    case Variable::K_PROGRAM:
    case Variable::K_CONST:
    case Variable::K_PARAM:
      env.err(name.loc, stringc << name.str << " can't be used like a function");

    // uninterp. fn symbol?
    case Variable::K_SELF: {
      E_apply *ret = new E_apply(name.clone(), NULL);
      FOREACH_ASTLIST_NC(Expression, args, iter) {
        ret->args.append(iter.data()->expand(env));
      }
      return ret;
    }

    // macro?
    case Variable::K_MACRO: {
      // open a scope, so I can add new bindings
      env.enterScope();

      // count params and args so we know how many default
      // arguments to supply
      int numParams = v->macro->params->names.count();
      int numArgs = args.count();
      
      // list for collecting bindings; do *not* add them to the
      // environment until all of the arguments have been
      // macro-expanded!
      ObjList<Variable> bindings;

      // walk parameter list to supply default arguments
      ASTListIterNC<Binding> paramIter(v->macro->params->names);
      while (numArgs < numParams &&
             !paramIter.isDone()) {
        Binding &b = *(paramIter.data());
        if (!b.hasDefault()) {
          env.err(name.loc, stringc << "too few arguments in call to " << name.str);
        }

        // bind the parameter to the evaluation of the default argument
        Variable *bind = new Variable(Variable::K_PARAM, b.name);
        E_var defVar(b.defValue.clone());
        bind->param = defVar.expand(env);
        bindings.prepend(bind);

        paramIter.adv();
        numParams--;
      }

      if (numArgs < numParams) {
        env.err(name.loc, stringc << "too few arguments in call to " << name.str);
      }
      if (numParams < numArgs) {
        env.err(name.loc, stringc << "too many arguments in call to " << name.str);
      }

      // simultaneously walk the parameter and argument lists, creating
      // remaining (non-defaulted) bindings
      ASTListIterNC<Expression> argIter(args);
      for (; !paramIter.isDone() && !argIter.isDone();
             paramIter.adv(), argIter.adv()) {
        // bind the parameter to the argument
        Binding &b = *(paramIter.data());
        Variable *bind = new Variable(Variable::K_PARAM, b.name);
        bind->param = argIter.data()->expand(env);
        bindings.prepend(bind);
      }
      xassert(paramIter.isDone() && argIter.isDone());

      // now add the bindings
      while (bindings.isNotEmpty()) {
        env.add(bindings.removeFirst());
      }

      // expand the body in the environment with the new bindings
      Expression *ret = v->macro->body->expand(env);

      env.exitScope();
      return ret;
    }
  }
}


Expression *E_int::expand(VCEnv &env)
{
  return clone();
}


Expression *E_bool::expand(VCEnv &env)
{
  return clone();
}


Expression *E_not::expand(VCEnv &env)
{
  return new E_not(e->expand(env));
}


Expression *E_neg::expand(VCEnv &env)
{
  return new E_neg(e->expand(env));
}


Expression *E_binary::expand(VCEnv &env)
{
  return new E_binary(e1->expand(env), op, e2->expand(env));
}


Expression *E_relational::expand(VCEnv &env)
{
  E_relational *ret = new E_relational(e->expand(env), NULL);

  FOREACH_ASTLIST_NC(RelElt, links, iter) {
    RelElt *re = iter.data();
    ret->links.append(new RelElt(re->op, re->e->expand(env)));
  }
  
  return ret;
}


Expression *E_quantifier::expand(VCEnv &env)
{
  // while expanding the quantifier, enter a scope so we
  // can make some self-bindings (i.e. don't expand parameters
  // that happen to have outer bindings)
  env.enterScope();

  FOREACH_ASTLIST_NC(Binding, vars->names, iter) {
    Binding &b = *(iter.data());
    if (b.hasDefault()) {
      env.err(b.name.loc, "quantified variables can't have default values");
    }

    // add a self-binding, to hide any existing definition
    // for this name
    Variable *v = new Variable(Variable::K_SELF, b.name);
    env.add(v);
  }

  E_quantifier *ret = new E_quantifier(isForall, vars->clone(),
    patterns? patterns->expand(env) : NULL, body->expand(env));

  env.exitScope();

  return ret;
}


Expression *E_let::expand(VCEnv &env)
{
  // binding for bound variable
  env.enterScope();

  // expand and add the binding
  Variable *v = new Variable(Variable::K_PARAM, name);
  v->param = binding->expand(env);
  env.add(v);

  // eval in new environment
  Expression *ret = body->expand(env);

  env.exitScope();
  return ret;
}


// ---------------------- Expression::tosx ----------------------
void Expression::tosx(VCEnv &env, bool pred)
{
  // const-eval?
  int result;
  if (constEval(result)) {
    if (!pred) {
      env << result;
    }
    else {
      if (result) {
        env << "TRUE";
      }
      else {
        env << "FALSE";
      }
    }
  }
  else {
    itosx(env, pred);
  }
}


void E_var::itosx(VCEnv &env, bool pred)
{
  env << name.str;
}


void E_apply::itosx(VCEnv &env, bool pred)
{
  env << "(" << name.str;
  FOREACH_ASTLIST_NC(Expression, args, iter) {
    env << " ";
    iter.data()->tosx(env);
  }
  env << ")";
}


void E_int::itosx(VCEnv &env, bool pred)
{                  
  // not actually used, due to const-eval
  env << i;
}


void E_bool::itosx(VCEnv &env, bool pred)
{
  // not actually used, due to const-eval
  env << (b? "TRUE" : "FALSE");
}


void E_not::itosx(VCEnv &env, bool pred)
{
  if (pred) {
    env << "(NOT ";        
  }
  else {
    env << "(- 1 ";     // 0 -> 1, 1 -> 0
  }
  e->tosx(env, pred);
  env << ")";
}


void E_neg::itosx(VCEnv &env, bool pred)
{
  env << "(- 0 ";     // subtract from 0 so '-' is always a binary operator (workaround for bug in Simplify)
  e->tosx(env);
  env << ")";
}


void E_binary::itosx(VCEnv &env, bool pred)
{
  char const *o;    // operator name as Simplify (& my axioms) knows it
  bool innerPred = false;
  switch (op) {
    default: xfailure("bad operator");
    case BIN_EQUAL:       o = "EQ"; break;
    case BIN_NOTEQUAL:    o = "NEQ"; break;
    case BIN_MULT:        o = "*"; break;
    case BIN_DIV:         o = "div"; break;
    case BIN_MOD:         o = "mod"; break;
    case BIN_PLUS:        o = "+"; break;
    case BIN_MINUS:       o = "-"; break;
    case BIN_LSHIFT:      o = "lshift"; break;
    case BIN_RSHIFT:      o = "rshift"; break;
    case BIN_BITAND:      o = "bitand"; break;
    case BIN_BITXOR:      o = "bitxor"; break;
    case BIN_BITOR:       o = "bitor"; break;
    case BIN_AND:         o = "AND";     innerPred=true; break;
    case BIN_OR:          o = "OR";      innerPred=true; break;
    case BIN_IMPLIES:     o = "IMPLIES"; innerPred=true; break;
    case BIN_EQUIVALENT:  o = "IFF";     innerPred=true; break;
  }

  env << "(" << o << " ";
  e1->tosx(env, innerPred);
  env << " ";
  e2->tosx(env, innerPred);
  env << ")";
}


void E_relational::itosx(VCEnv &env, bool pred)
{
  // need to conjoin?
  if (links.count() > 1) {
    env << "(AND ";
  }

  // expression with which we're relating a given link's expression
  Expression *prev = e;

  int ct=0;
  FOREACH_ASTLIST_NC(RelElt, links, iter) {
    RelElt *re = iter.data();

    if (ct++ > 0) {
      env << " ";
    }

    // emit each pair relation
    env << "(" << toString(re->op) << " ";
    prev->tosx(env);
    env << " ";
    re->e->tosx(env);
    env << ")";

    prev = re->e;
  }

  if (links.count() > 1) {
    env << ")";
  }
}


void E_quantifier::itosx(VCEnv &env, bool pred)
{
  env << "\n  (" << (isForall? "FORALL" : "EXISTS") << " (";

  int ct=0;
  FOREACH_ASTLIST_NC(Binding, vars->names, iter) {
    if (ct++ > 0) {
      env << " ";
    }
    env << iter.data()->name.str;
  }

  env << ") ";

  if (patterns) {
    // note that the 'patterns' expression must be constructed
    // with the PATS and MPAT (if needed) already in it
    patterns->tosx(env);
    env << " ";
  }

  body->tosx(env, pred);

  env << ")";
}


void E_let::itosx(VCEnv &env, bool pred)
{
  // this shouldn't ever appear in the sx output
  env << "<let expression in sx output!>";
}


// --------------------- Expression::constEval ---------------------
bool Expression::constEval(int &result) const
{
  ASTSWITCHC(Expression, this) {
    // E_var: no  
    
    // E_apply: no

    ASTCASEC(E_int, i)
      result = i->i;
      return true;

    ASTNEXTC(E_bool, b)
      result = b->b? 1 : 0;
      return true;

    ASTNEXTC(E_not, n)
      if (!n->e->constEval(result)) return false;
      result = !result;
      return true;

    ASTNEXTC(E_neg, n)
      if (!n->e->constEval(result)) return false;
      result = -result;
      return true;

    ASTNEXTC(E_binary, b)
      int v1, v2;
      if (!b->e1->constEval(v1) ||
          !b->e2->constEval(v2)) return false;

      if (v2==0 && (b->op == BIN_DIV || b->op == BIN_MOD)) {
        cout << "division by zero in constant expression\n";
        return false;
      }

      switch (b->op) {
        default: xfailure("bad binary opcode");

        case BIN_EQUAL:     result = (v1 == v2);  return true;
        case BIN_NOTEQUAL:  result = (v1 != v2);  return true;
        case BIN_LESS:      result = (v1 < v2);  return true;
        case BIN_GREATER:   result = (v1 > v2);  return true;
        case BIN_LESSEQ:    result = (v1 <= v2);  return true;
        case BIN_GREATEREQ: result = (v1 >= v2);  return true;

        case BIN_MULT:      result = (v1 * v2);  return true;
        case BIN_DIV:       result = (v1 / v2);  return true;
        case BIN_MOD:       result = (v1 % v2);  return true;
        case BIN_PLUS:      result = (v1 + v2);  return true;
        case BIN_MINUS:     result = (v1 - v2);  return true;
        case BIN_LSHIFT:    result = (v1 << v2);  return true;
        case BIN_RSHIFT:    result = (v1 >> v2);  return true;
        case BIN_BITAND:    result = (v1 & v2);  return true;
        case BIN_BITXOR:    result = (v1 ^ v2);  return true;
        case BIN_BITOR:     result = (v1 | v2);  return true;
        case BIN_AND:       result = (v1 && v2);  return true;
        case BIN_OR:        result = (v1 || v2);  return true;

        case BIN_IMPLIES:    result = (!v1 || v2);  return true;
        case BIN_EQUIVALENT: result = (v1 == v2);  return true;
      }

    ASTNEXTC(E_relational, r)
      int v1, v2;            // two adjacent values in the chain
      result = 1;            // true(1) if all pairs so far have been true
      if (!r->e->constEval(v2)) return false;

      FOREACH_ASTLIST(RelElt, r->links, iter) {
        v1 = v2;
        if (!iter.data()->e->constEval(v2)) return false;

        if (result) {
          switch (iter.data()->op) {
            default: xfailure("bad relational opcode");
            case BIN_EQUAL:     result = (v1 == v2);  return true;
            case BIN_NOTEQUAL:  result = (v1 != v2);  return true;
            case BIN_LESS:      result = (v1 < v2);   return true;
            case BIN_GREATER:   result = (v1 > v2);   return true;
            case BIN_LESSEQ:    result = (v1 <= v2);  return true;
            case BIN_GREATEREQ: result = (v1 >= v2);  return true;
          }
        }
      }
      
      return true;

    // E_quantifier: no
    
    // E_let: no

    ASTDEFAULTC
      return false;

    ASTENDCASEC
  }
}



// EOF
