// vc_env.cc
// code for vc_env.h

#include "vc_env.h"          // this module
#include "lexer.h"           // Lexer
#include "vclang.gr.gen.h"   // VCParse
#include "parsetables.h"     // ParseTables
#include "glr.h"             // GLR
#include "exc.h"             // xBase
#include "prover.h"          // prover.h
#include "nonport.h"         // getMilliseconds
#include "strutil.h"         // parseQuotedString
#include "trace.h"           // tracingSys

#include <stdlib.h>          // strtol, exit


// ----------------------- Variable -------------------------
Variable::~Variable()
{
  if (param) {
    delete param;
  }
}


// ------------------------ VCScope -------------------------
VCScope::VCScope(char const * /*nullable*/ n)
  : name(n? n : ""),
    pushes(0),
    vars(),
    creationTime(getMilliseconds())
{}

VCScope::~VCScope()
{}


// ------------------------- VCEnv --------------------------
VCEnv::VCEnv()
  : str(),
    lang(),
    selStr(LITERAL_LOCSTRING(str("sel"))),
    updStr(LITERAL_LOCSTRING(str("upd"))),
    ptrStr(LITERAL_LOCSTRING(str("ptr"))),
    updsliceStr(LITERAL_LOCSTRING(str("updslice"))),
    predicateStr(LITERAL_LOCSTRING(str("\"predicate\""))),
    memStr(LITERAL_LOCSTRING(str("mem"))),
    metaStr(LITERAL_LOCSTRING(str("meta"))),
    canReachStr(LITERAL_LOCSTRING(str("canReach"))),
    canReachWOStr(LITERAL_LOCSTRING(str("canReachWO"))),
    reachFieldStr(LITERAL_LOCSTRING(str("reach_field"))),
    updrangeStr(LITERAL_LOCSTRING(str("updrange"))),
    tables(NULL),
    prover(new Prover),
    proofCount(1),
    included(),
    scopes(),
    outputs(),
    failedProofs(0)
{
  enterScope();

  lang.ANSI_Cplusplus();     // want 'true' and 'false' keywords

  prover->openLog("simp.log");
  
  // simplify's internal function symbols
  introduce(LITERAL_LOCSTRING(str("PATS")));
  introduce(LITERAL_LOCSTRING(str("MPAT")));
}

VCEnv::~VCEnv()
{
  if (tables) {
    delete tables;
  }

  delete prover;

  while (!outputs.isEmpty()) {
    delete outputs.pop();
  }
}


VCSequence *VCEnv::parseVCLang(rostring fname)
{
  // lexer
  Lexer lexer(str, lang, toCStr(fname));

  // parser context class
  VCParse vcparse(*this);

  // parse tables (I delay making these until we're ready to parse
  // because I need a VCParse object, but I want one of those for
  // each recursive parse, so don't want to waste one in advance)
  if (!tables) {
    tables = vcparse.makeTables();
  }
  GLR glr(&vcparse, tables);

  // run the parser
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    xbase(stringc << "parse error in file " << fname);
  }

  return (VCSequence*)treeTop;
}


ASTList<Expression> *VCEnv::list2(Expression *e1, Expression *e2)
{
  ASTList<Expression> *ret = new ASTList<Expression>(e1);
  ret->append(e2);
  return ret;
}

ASTList<Expression> *VCEnv::list3(Expression *e1, Expression *e2, Expression *e3)
{
  ASTList<Expression> *ret = list2(e1, e2);
  ret->append(e3);
  return ret;
}

ASTList<Expression> *VCEnv::list4(Expression *e1, Expression *e2, Expression *e3, Expression *e4)
{
  ASTList<Expression> *ret = list3(e1, e2, e3);
  ret->append(e4);
  return ret;
}

ASTList<Expression> *VCEnv::list5(Expression *e1, Expression *e2, Expression *e3, Expression *e4, Expression *e5)
{
  ASTList<Expression> *ret = list4(e1, e2, e3, e4);
  ret->append(e5);
  return ret;
}

Expression *VCEnv::sel(Expression *mem, Expression *p)
  { return new E_apply(selStr.clone(), list2(mem, p)); }
Expression *VCEnv::upd(Expression *mem, Expression *p, Expression *v)
  { return new E_apply(updStr.clone(), list3(mem, p, v)); }
Expression *VCEnv::ptr(Expression *addr, Expression *field)
  { return new E_apply(ptrStr.clone(), list2(addr, field)); }
Expression *VCEnv::updslice(Expression *mem, Expression *addr, Expression *v)
  { return new E_apply(updsliceStr.clone(), list3(mem, addr, v)); }
Expression *VCEnv::updrange(Expression *mem, Expression *addr, Expression *low, Expression *high, Expression *v)
  { return new E_apply(updrangeStr.clone(), list5(mem, addr, low, high, v)); }


Expression *VCEnv::canReach(Expression *src, LocString *field, Expression *dest)
{
  return new E_apply(canReachStr.clone(),
                     list4(
                       new E_var(memStr.clone()),
                       new E_var(field),
                       src,
                       dest
                     ));
}

Expression *VCEnv::canReachWO(Expression *src, LocString *field, Expression *dest,
                              Expression *wo)
{
  return new E_apply(canReachWOStr.clone(),
                     list5(
                       new E_var(memStr.clone()),
                       new E_var(field),
                       src,
                       dest,
                       wo
                     ));
}


Expression *VCEnv::makeRel(Expression *e1, BinaryOp op, Expression *e2)
{
  // wrap this in an E_relational if necessary
  if (!e1->isE_relational()) {
    e1 = new E_relational(e1, NULL);
  }

  // add another link
  e1->asE_relational()->links.append(new RelElt(op, e2));
  return e1;
}

int VCEnv::parseInt(StringRef s)
{
  return (int)strtol(s, NULL /*endptr*/, 0 /*radix*/);
}


VCEnv& VCEnv::operator<< (char const *str)
{
  if (outputs.isEmpty()) {
    prover->feed(str);
  }
  else {
    *(outputs.top()) << str;
  }
  return *this;
}


VCEnv& VCEnv::operator<< (int i)
{
  return operator<< (stringc << i);
}


string VCEnv::evalToSexp(Expression *e, bool pred)
{
  // push a new output stream to catch the expr's sexp
  outputs.push(new stringBuilder);

  // eval it
  e->eval(*this, pred);

  // get the string to return
  string ret = (*outputs.top());
  delete outputs.pop();
  return ret;
}

string VCEnv::toSexp(Expression *e, bool pred)
{
  // push a new output stream to catch the expr's sexp
  outputs.push(new stringBuilder);

  // sexp it
  e->tosx(*this, pred);

  // get the string to return
  string ret = (*outputs.top());
  delete outputs.pop();
  return ret;
}


VCScope &VCEnv::scope()
{
  return *(scopes.first());
}


bool isCallScopeName(rostring str)
{
  return 0==strcmp(toCStr(str), "call") ||
         0==memcmp(toCStr(str), "call:", 5);
}


VCScope &VCEnv::nonCallScope()
{
  // ugly-ass hack: skip scopes called "call"
  int n = 0;
  while (isCallScopeName(scopes.nth(n)->name)) {
    n++;
  }
  return *(scopes.nth(n));
}


Variable *VCEnv::lookup(StringRef name)
{
  VCScope *dummy;
  return lookup(name, dummy);
}

Variable *VCEnv::lookup(StringRef name, VCScope *&scope)
{
  // walk up through scopes looking for this variable
  FOREACH_OBJLIST_NC(VCScope, scopes, iter) {
    Variable *v = iter.data()->vars.queryif(name);
    if (v) {
      scope = iter.data();     // the scope in which the variable was found
      return v;
    }
  }
  return NULL;                 // not found
}


Variable *VCEnv::lookupForModification(StringRef name)
{
  VCScope *s;
  Variable *v = lookup(name, s);
  if (!v || v->kind != Variable::K_PROGRAM) {
    // didn't find a modifiable variable
    return v;
  }

  // if this variable wasn't declared in this scope, then make
  // a shadow declaration so the original binding will be restored
  // when the block exists
  VCScope *inner = scopes.first();
  if (s != inner) {
    Variable *copy = new Variable(v->kind, v->name);
    copy->index = v->index;
    inner->vars.add(copy->name.str, copy);
    return copy;
  }
  else {
    return v;    // already innermost
  }
}


void VCEnv::failIfLocallyDefined(LocString const &s)
{
  if (scope().vars.isMapped(s.str)) {
    err(s.loc, stringc << "multiply-declared: " << s.str);
  }
}

void VCEnv::add(Variable * /*owner*/ v)
{
  failIfLocallyDefined(v->name);
  scope().vars.add(v->name.str, v);
}

void VCEnv::introduce(LocString const &s)
{
  // add self-binding
  *this << "; introduce: " << s.str << "\n";
  Variable *v = new Variable(Variable::K_SELF, s);
  add(v);
}


VCScope *VCEnv::enterScope(char const * /*nullable*/ name)
{
  VCScope *newScope = new VCScope(name);

  #if 0    // delete me
  // add all the existing program variables to the new scope, so that
  // when it goes away all mappings will be restored
  FOREACH_OBJLIST_NC(VCScope, scopes, iter) {
    VCScope *s = iter.data();

    // look through all variables of 's'
    for (StringObjDict<Variable>::Iter iter2(s->vars);
         !iter2.isDone();
         iter2.next()) {
      Variable const *v = iter2.value();

      // add copies of all the program variables; if the variable
      // is already present then it means I already have a copy from
      // an inner scope
      if (v->kind == Variable::K_PROGRAM &&
          !newScope->vars.isMapped(v->name.str)) {
        Variable *copy = new Variable(v->kind, v->name);
        copy->index = v->index;
        newScope->vars.add(copy->name.str, copy);
      }
    }
  }
  #endif // 0

  scopes.prepend(newScope);
  
  return newScope;
}

void VCEnv::exitScope()
{
  // pop any added assumptions
  int i = scope().pushes;
  while (i > 0) {
    *this << "(BG_POP)\n";
    i--;
  }

  scopes.deleteAt(0);
}


void VCEnv::err(SourceLoc loc, rostring msg)
{
  xbase(stringc << toString(loc) << ": " << msg);
}


void VCEnv::assume(Expression *pred)
{
  *this << "(BG_PUSH "
        << evalToSexp(pred, true /*pred*/)   // make a string first, to minimize IPC overhead
        << ")\n\n";
  nonCallScope().pushes++;
}


void VCEnv::proveExpr(rostring label, Expression *pred)
{
  VCEnv &env = *this;

  if (!prove(parseQuotedString(label), evalToSexp(pred, true /*pred*/), 
             true /*expect*/)) {
    // failed to prove.. let's break down the goal.  first step
    // is to expand it without conversion to sexp
    Expression * /*owner*/ expanded = pred->expand(env);
    findInvalidConjunct(expanded, 0 /*depth*/);
    delete expanded;

    if (!tracingSys("contin")) {
      exit(2);    // bail on failed proof
    }
  }
}


bool VCEnv::findInvalidConjunct(Expression *pred, int depth)
{
  // recursive step
  if (pred->isE_binary() && pred->asE_binary()->op == BIN_AND) {
    E_binary *b = pred->asE_binary();
    return findInvalidConjunct(b->e1, depth+1) ||
           findInvalidConjunct(b->e2, depth+1);
  }

  // base case
  if (prove(NULL /*be silent*/, toSexp(pred, true /*pred*/), false /*expect*/)) {
    // found it; but only print if this is different from the
    // entire failing predicate
    if (depth > 0) {
      cout << "smaller: " << toSexp(pred, true /*pred*/) << "\n\n";
    }
    return true;
  }
  else {
    return false;     // keep looking
  }
}


bool VCEnv::skippingThisBlock()
{
  if (blockOfInterest.length() == 0) {
    return false;    // not skipping anything
  }
  
  // skip any scopes called "if", so that we evaluate inside
  // "if" blocks inside the block of interest
  int index = 0;
  VCScope *s = scopes.first();
  while (s->name.equals("if") ||
         isCallScopeName(s->name)) {        // also skip call blocks
    index++;
    s = scopes.nth(index);
  }
                                          
  // we're skipping it if the name doesn't match
  return !blockOfInterest.equals(s->name);      
}


bool VCEnv::prove(rostring name, rostring pred, bool expect)
{
  if (name[0]) {
    if (skippingThisBlock()) {
      // we're only trying to prove one block, but this isn't it
      *this << "; skipped proof: " << name << "\n";
      return true;    // pretend it went the way we expect
    }

    *this << "; " << (expect? "proof obligation" : "expect fail proof")
          << ": " << name << "\n";

    // print a label to see while Simplify works
    char label[20];
    sprintf(label, "%svalid: %3d: ", (expect? "   " : "not"), proofCount++);
    cout << label << name << "... ";
    cout.flush();
  }

  // render the predicate into a string
  string sx = stringc << pred << "\n";

  // try to prove it
  long start = getMilliseconds();
  bool proved = prover->prove(sx);
  long end = getMilliseconds();

  // if the proof had a label, then print some info about what happened
  if (name[0]) {
    if (!proved) {
      cout << "not ";
    }
    cout << "proved (" << (end-start) << " ms)";

    if (expect != proved) {
      cout << " (UNEXPECTED)\n";
      cout << "failed: " << sx;
      failedProofs++;
    }
    else {
      cout << "\n";
    }
    
    // put some blank lines into the log after the predicate
    *this << "\n\n";
  }
  
  return expect == proved;
}


// EOF
