// vc_env.h
// environment for interpreting VC language

#ifndef VC_ENV_H
#define VC_ENV_H

#include "cc_lang.h"         // CCLang
#include "strtable.h"        // StringTable
#include "vclang.ast.gen.h"  // VCLang ast
#include "strobjdict.h"      // StrObjDict
#include "str.h"             // stringBuilder
#include "array.h"           // ArrayStack

class ParseTables;           // parsetables.h
class Prover;                // prover.h


// information about a variable in a VC
class Variable {
public:
  enum Kind {
    K_PROGRAM,         // updateable, numerically-indexed variable
    K_MACRO,           // macro that accepts at least one parameter
    K_CONST,           // 0-ary macro
    K_SELF,            // maps to its own name
    K_PARAM,           // already-expanded argument expression
    NUM_KINDS
  } kind;

  LocString name;      // name of this variable

  union {
    int index;         // for K_PROGRAM
    Macro *macro;      // for K_MACRO
    Expression *expr;  // for K_CONST
  };

  Expression *param;   // (owner) for K_PARAM

public:
  Variable(Kind k, LocString const &n)
    : kind(k), name(n), macro(NULL), param(NULL) {}
  ~Variable();
};


// set of local maps that temporarily shadow outer maps
class VCScope {
public:      // data
  // name of this scope, if any ("" if no name)
  string name;

  // number of bg_pushes that have been done in the current block
  int pushes;

  // variable map
  StringObjDict<Variable> vars;
  
  // getMilliseconds() when scope created
  long creationTime;

public:      // funcs
  VCScope(char const *name = NULL);
  ~VCScope();
};


class VCEnv {
public:      // data
  // string table for storing identifiers
  StringTable str;

  // parsing language options; only used incidentally because
  // I borrowed the C lexer to make the VClang lexer
  CCLang lang;

  // static locstrings for use in special forms
  LocString selStr, updStr, ptrStr, updsliceStr, predicateStr, 
            memStr, metaStr, canReachStr, canReachWOStr,
            reachFieldStr, updrangeStr;

  // parse tables; shared by all active parsers
  ParseTables *tables;               // (owner)

  // interface prover
  Prover *prover;                    // (owner)
  int proofCount;

  // set of files that have been included; this is so we
  // can avoid multiply-including something
  StringObjDict<VCSequence> included;

  // stack of scopes
  ObjList<VCScope> scopes;

  // stack of temporary string dests; when this is empty, we're
  // writing directly to prover; the pushes/pops of this stack are
  // correlated with the scopes, but it's not 1-1, hence the
  // separate stack
  ArrayStack<stringBuilder*> outputs;

  // name of block to prove; if this is not "", then all proofs
  // are skipped except those for which the innermost block name
  // matches
  string blockOfInterest;

  // failed proofs; when nonzero, we say the VC as a whole failed
  int failedProofs;

private:     // funcs
  Variable *lookup(StringRef name, VCScope *&scope);
  bool findInvalidConjunct(Expression *pred, int depth);

public:      // funcs
  VCEnv();
  ~VCEnv();

  // parse a filename into an AST
  VCSequence *parseVCLang(rostring fname);

  // functions invoked during parsing, to build parts of the AST
  ASTList<Expression> *list2(Expression *e1, Expression *e2);
  ASTList<Expression> *list3(Expression *e1, Expression *e2, Expression *e3);
  ASTList<Expression> *list4(Expression *e1, Expression *e2, Expression *e3, Expression *e4);
  ASTList<Expression> *list5(Expression *e1, Expression *e2, Expression *e3, Expression *e4, Expression *e5);
  Expression *sel(Expression *mem, Expression *p);
  Expression *upd(Expression *mem, Expression *p, Expression *v);
  Expression *ptr(Expression *addr, Expression *field);
  Expression *updslice(Expression *mem, Expression *addr, Expression *v);
  Expression *updrange(Expression *mem, Expression *addr, Expression *low, Expression *high, Expression *v);
  Expression *canReach(Expression *src, LocString *field, Expression *dest);
  Expression *canReachWO(Expression *src, LocString *field, Expression *dest, Expression *wo);

  Expression *makeRel(Expression *e1, BinaryOp op, Expression *e2);
  int parseInt(StringRef s);

  // how to add to the s-exp we're currently building
  VCEnv& operator<< (char const *str);
  VCEnv& operator<< (rostring str) { return operator<< (str.c_str()); }
  VCEnv& operator<< (int i);

  // eval the expression, but grab the output and yield it as
  // a string instead of letting it go to prover
  string evalToSexp(Expression *e, bool pred);

  // just the tosx part, not the expand part
  string toSexp(Expression *e, bool pred);

  // current scope
  VCScope &scope();
  VCScope &nonCallScope();

  // consult scoped map; return NULL if no mapping
  Variable *lookup(StringRef name);
  Variable *lookupForModification(StringRef name);

  // manipulate scopes
  VCScope *enterScope(char const *name = NULL);
  void exitScope();
  int scopeDepth() const { return scopes.count(); }

  // add mappings
  void add(Variable * /*owner*/ v);
  void failIfLocallyDefined(LocString const &s);
  
  // add uninterpreted name
  void introduce(LocString const &s);

  // report error; throw exception
  void err(SourceLoc loc, rostring msg);

  // add a predicate to the prover's assumptions
  void assume(Expression *pred);
  
  // prove a predicate
  void proveExpr(rostring label, Expression *pred);

  // attempt to prove a predicate, rendered as an s-exp without
  // a terminating "\n"; 'expect' is true if we expect it to succeed;
  // return true if it went the way we expect
  bool prove(rostring name, rostring pred, bool expect);
                                 
  // true if we're in a block that isn't the block of interest, and
  // the latter is active (not "")
  bool skippingThisBlock();
};

#endif // VC_ENV_H
