/* vml.lex            see license.txt for copyright and terms of use
 * flex description of scanner for VML source
 * adapted from elsa/cc.lex
 */
 
/* ------------------ C definitions ------------------- */
%{

#include "vml_lexer.h"       // VMLLexer

// this works around a problem with cygwin & fileno
#define YY_NEVER_INTERACTIVE 1

%}


/* -------------------- flex options ------------------ */
/* no wrapping is needed; setting this means we don't have to link with libfl.a */
%option noyywrap

/* don't use the default-echo rules */
%option nodefault

/* I don't call unput */
%option nounput

/* generate a c++ lexer */
%option c++

/* use the "fast" algorithm with no table compression */
%option full

/* utilize character equivalence classes */
%option ecs

/* the scanner is never interactive */
%option never-interactive

/* and I will define the class (vml_lexer.h) */
%option yyclass="VMLLexer"


/* ------------------- definitions -------------------- */
/* newline */
NL            "\n"

/* anything but newline */
NOTNL         .

/* any of 256 source characters */
ANY           ({NOTNL}|{NL})

/* some special characters */
BACKSL        "\\"
CR            "\r"

/* beginnging of line (must be start of a pattern) */
BOL           ^

/* end of line (would like EOF to qualify also, but flex doesn't allow it */
EOL           {NL}

/* letter or underscore */
LETTER        [A-Za-z_]

/* letter or underscore or digit */
ALNUM         [A-Za-z_0-9]

/* decimal digit */
DIGIT         [0-9]

/* sequence of decimal digits */
DIGITS        ({DIGIT}+)

/* sign of a number */
SIGN          ("+"|"-")

/* integer suffix */
/* added 'LL' option for GNU long long compatibility.. */
ELL_SUFFIX    [lL]([lL]?)
INT_SUFFIX    ([uU]{ELL_SUFFIX}?|{ELL_SUFFIX}[uU]?)

/* floating-point suffix letter */
FLOAT_SUFFIX  [flFL]

/* normal string character: any but quote, newline, or backslash */
STRCHAR       [^\"\n\\]

/* (start of) an escape sequence */
/* 2005-01-17: treat \\\r\n as an escape sequence to allow backslash
 * coninuation in DOS text files */
ESCAPE        (({BACKSL}{ANY})|({BACKSL}{CR}{NL}))

/* double quote */
QUOTE         [\"]

/* normal character literal character: any but single-quote, newline, or backslash */
CCCHAR        [^\'\n\\]

/* single quote */
TICK          [\']

/* space or tab */
SPTAB         [ \t]

/* preprocessor "character" -- any but escaped newline */
PPCHAR        ([^\\\n]|{BACKSL}{NOTNL})


/* ------------- token definition rules --------------- */
%%      /* EXTENSION RULES GO HERE */

  /* operators, punctuators and keywords: tokens with one spelling */
"__pat"              return tok(TOK___PAT);
"AND"                return tok(TOK_ANDKEY);
"axiom"              return tok(TOK_AXIOM);
"assert"             return tok(TOK_ASSERT);
"assume"             return tok(TOK_ASSUME);
"const"              return tok(TOK_CONST);
"declare_partial_order"  return tok(TOK_DECLARE_PARTIAL_ORDER);
"else"               return tok(TOK_ELSE);
"exists"             return tok(TOK_EXISTS);
"false"              return tok(TOK_FALSE);
"forall"             return tok(TOK_FORALL);
"funcsym"            return tok(TOK_FUNCSYM);
"goto"               return tok(TOK_GOTO);
"if"                 return tok(TOK_IF);
"invariant"          return tok(TOK_INVARIANT);
"OR"                 return tok(TOK_ORKEY);
"post"               return tok(TOK_POST);
"pre"                return tok(TOK_PRE);
"procedure"          return tok(TOK_PROCEDURE);
"pure_assert"        return tok(TOK_PURE_ASSERT);
"reads"              return tok(TOK_READS);
"standalone_invariant"  return tok(TOK_STANDALONE_INVARIANT);
"true"               return tok(TOK_TRUE);
"var"                return tok(TOK_VAR);
"while"              return tok(TOK_WHILE);
"writes"             return tok(TOK_WRITES);

"{"                  return tok(TOK_LBRACE);
"}"                  return tok(TOK_RBRACE);
";"                  return tok(TOK_SEMICOLON);
":"                  return tok(TOK_COLON);
":="                 return tok(TOK_COLONEQ);
"("                  return tok(TOK_LPAREN);
")"                  return tok(TOK_RPAREN);
"."                  return tok(TOK_DOT);
"!"                  return tok(TOK_BANG);
"&&"                 return tok(TOK_ANDAND);
"||"                 return tok(TOK_OROR);
"==>"                return tok(TOK_IMPLIES);
"<==>"               return tok(TOK_IFF);
"=="                 return tok(TOK_EQUAL);
"!="                 return tok(TOK_NOTEQUAL);
"<"                  return tok(TOK_LESS);
"<="                 return tok(TOK_LESSEQ);
">"                  return tok(TOK_GREATER);
">="                 return tok(TOK_GREATEREQ);
"+"                  return tok(TOK_PLUS);
"-"                  return tok(TOK_MINUS);
"*"                  return tok(TOK_TIMES);
"["                  return tok(TOK_LBRACKET);
"]"                  return tok(TOK_RBRACKET);
","                  return tok(TOK_COMMA);
"->"                 return tok(TOK_ARROW);
"="                  return tok(TOK_SINGLEEQ);
"?"                  return tok(TOK_QUESTION);

  /* this rule is to avoid backing up in the lexer */
"<==" {
  yyless(2);     /* put back all but 2 */
  return tok(TOK_LESSEQ);
}

  /* identifier: e.g. foo */
{LETTER}{ALNUM}* {
  updLoc();
  sval = (SemanticValue)addString(yytext, yyleng);
  return TOK_IDENTIFIER;
}

  /* integer literal */
[0-9]+ {
  updLoc();
  sval = (SemanticValue)atoi(yytext);
  return TOK_INT_LIT;
}

  /* string literal */
"L"?{QUOTE}({STRCHAR}|{ESCAPE})*{QUOTE} {
  return svalTok(TOK_STRING_LIT);
}

  /* string literal missing final quote */
"L"?{QUOTE}({STRCHAR}|{ESCAPE})*{EOL}   {
  err("string literal missing final `\"'");
  return svalTok(TOK_STRING_LIT);          // error recovery
}

  /* unterminated string literal */
"L"?{QUOTE}({STRCHAR}|{ESCAPE})*{BACKSL}? {
  err("unterminated string literal");
  yyterminate();
}

  /* whitespace */
[ \t\n\f\v\r]+  {
  whitespace();
}

  /* C++ comment */
  /* we don't match the \n because that way this works at EOF */
"//"{NOTNL}*    {
  whitespace();
}

  /* C comment */
"/""*"([^*]|"*"*[^*/])*"*"+"/"     {
  // the pattern is a little complicated because the naive one,
  //   "/""*"([^*]|"*"[^/])*"*/"
  // fails to match e.g. "/***/"
  whitespace();
}

  /* unterminated C comment */
"/""*"([^*]|"*"*[^*/])*"*"*        {
  err("unterminated /""*...*""/ comment");
  yyterminate();
}

  /* illegal */
.  {
  updLoc();
  err(stringc << "illegal character: `" << yytext[0] 
              << "' (code " << (int)yytext[0] << ")");
}

<<EOF>> {
  srcFile->doneAdding();
  yyterminate();
}

