// vml_env.h
// type checking environment for VML

#ifndef VML_ENV_H
#define VML_ENV_H

#include "strtable.h"       // StringTable, StringRef
#include "srcloc.h"         // SourceLoc
#include "array.h"          // ObjArrayStack
#include "strsobjdict.h"    // StringSObjDict
#include "sobjlist.h"       // SObjList

class VDeclaration;         // vml.ast
class VProcedure;           // vml.ast
class VStatement;           // vml.ast
class VS_label;             // vml.ast


class VMLEnv {
private:     // data
  // stack of scopes; the scope at length()-1 is the innermost
  ObjArrayStack< StringSObjDict<VDeclaration> > scopes;

  // information about all procedures (for which there is no scoping)
  StringSObjDict<VProcedure> procs;

  // info about labels in the current procedure; must be cleared
  // between procedure
  StringSObjDict<VS_label> labels;

  // stack of sets of pending-nexts, for building the CFG
  ObjArrayStack< SObjList<VStatement> > nexts;

  // count of errors reported
  int errors;

public:      // data
  // string table used during lexing/parsing
  StringTable &str;

  // nonzero when we're inside an E_pre
  int preState;

  // number of quantifiers we're inside
  int quantifiers;

  // when true, emit the CFG as a Dot graph
  bool cfgGraph;

private:     // funcs
  // current scope
  StringSObjDict<VDeclaration> *scope()
    { return scopes.top(); }

public:      // funcs
  VMLEnv(StringTable &table);
  ~VMLEnv();
  

  // add a new variable to the current scope; return false if
  // its name collides with another variable in the same scope
  bool addVar(StringRef name, VDeclaration *decl);

  // look up a variable; searches all active scopes; returns NULL
  // if no variable by that name is found
  VDeclaration *lookup(StringRef name);
  VDeclaration *lookup(rostring name) { return lookup(str(name.c_str())); }


  // open a new, empty scope
  void enterScope();

  // close innermost scope
  void exitScope();


  // add/lookup for procedure; reports error on collision
  void addProcedure(VProcedure *p);
  VProcedure *lookupProcedure(StringRef name);   // NULL if not found


  // add/lookup for labels; reports error on collision
  void addLabel(VS_label *sl);
  void clearLabelInfo();
  VS_label *lookupLabel(StringRef name);         // NULL if not found

  
  // various manipulations of the pending-nexts sets
  void resolvePendingNexts(VStatement *target);
  void addPendingNext(VStatement *source);
  void pushPendingNexts();
  void popAndMergeNexts();
  void clearPendingNexts();


  // error reporting
  void error(SourceLoc loc, char const *msg);
  void error(char const *msg);
  int numErrors() { return errors; }
};


#endif // VML_ENV_H
