// vml_fflow.h
// fact flow for VML

// "Fact flow" is an analysis that strengthens invariants using a
// dataflow-like procedure in which assumptions flow from sources like
// preconditions, assertions, conditionals and assignments down to
// invariants.  Flow is impeded by assignments that change the value
// of one of the variables mentioned in an assumption.
//
// At the end of the analysis, any facts that flow to a given
// invariant point (along all possible incoming edges) are added to
// that invariant, strengthening it.  The analysis ensures that if
// the original predicate was provable then the new one is too.

// Conceptually, the algorithm computes the greatest fixpoint for
// dataflow equations in which:
//   1. The start node's facts are the same as the function precondition.
//   2. Every other node's facts are the intersection of facts from
//      preceeding nodes, minus the node's kill, plus its gen.
// The fixpoint is found by initializing all nodes' facts to the
// universe U of all facts, applying rule 1, then applying rule 2
// iteratively until quiescence.

// This implementation is based on the older "factflow" module, which
// was written to work directly on the C++ AST.

#ifndef VML_FFLOW_H
#define VML_FFLOW_H

#ifdef VML_FFLOW_IMPLEMENTATION // for .cc file only

#include "vml.ast.gen.h"        // VML AST
#include "ohashtbl.h"           // OwnerHashTable
#include "arrayqueue.h"         // ArrayQueue
#include "sobjset.h"            // SObjSet


// predicate set
typedef SObjList<Pred const> PredSet;


// fflow information about a point in the control flow graph;
// specifically, the point just before 'stmt'
class StmtFacts {
public:      // data
  // this StmtFacts object contains information about the control
  // point just before this statement; or, if 'stmt' is NULL, this
  // object has information about the end of the procedure
  VStatement *stmt;             // (serf)

  // set of facts inferred for this node; none of these facts is
  // P_binary(BP_AND, ...) since the entire set is implicitly
  // a conjunction; the Pred objects are owned by the AST
  PredSet facts;

  // set of predicates generated by this node; this is computed
  // once at the beginning, and these predicates are deleted at
  // the end
  ObjList<Pred> genPreds;

  // for the special case of VS_if, this is its predicate with
  // a 'not' applied to the outside
  Pred * /*nullable owner*/ negIfPred;

private:     // funcs
  void computeGenPreds(StringTable &str);

public:      // funcs
  StmtFacts(VStatement *s, StringTable &str);
  ~StmtFacts();

  Pred const *getNegIfPred();
};


// manager for the entire algorithm
class VMLFFlow {
private:     // data
  // string table
  StringTable &str;

  // map: stmt -> StmtFacts
  //
  // NOTE: if a statement is not mapped, then we regard that
  // statement's 'facts' set as the entire universe of predicates
  OwnerHashTable<StmtFacts> map;

  // worklist of statements yet to process
  ArrayQueue<VStatement*> worklist;

private:     // funcs
  // see comments in the .cc file
  static void const *getKey(StmtFacts *data);
  void reset();
  void rp_dfs(SObjSet<VStatement*> &seen, VStatement *stmt);
  void reversePostorder(VProcedure *func);
  void getSuccessors(ArrayStack<VStatement*> &dest, VStatement *stmt);
  void enqueueToWorklist(VStatement *stmt);
  void addFact(PredSet &dest, Pred const *f);
  bool intersectFacts(PredSet &dest, PredSet &src);
  void killReferencesTo(PredSet &facts, StringRef var);
  void computeOutFacts(PredSet &outFacts, StmtFacts *stmtFacts);
  void addEdgeFacts(PredSet &facts, StmtFacts *startFacts, VStatement *end);
  void strengthenInvariants();

public:      // funcs
  VMLFFlow(StringTable &str);
  ~VMLFFlow();

  // analyze a procedure and strengthen its invariants
  void analyze(VProcedure *func);
};

#endif // VML_FFLOW_IMPLEMENTATION


// stand-alone analysis function
class VProcedure;
void vmlFFlowAnalyze(VProcedure *func, StringTable &str);


#endif // VML_FFLOW_H
