// vml_tcheck.cc
// type checker for VML
//
// Since, at the moment, VML does not have typed variables, the
// function of the type checker is just:
//   - bind uses of variables to their declarations
//   - check certain well-formedness properties, e.g. with RelElt
//
// However, since adding types is a possible future direction, I
// have coded the checker as a recursive walk, not a visitor.

#include "vml.ast.gen.h"      // VML AST
#include "vml_env.h"          // VMLEnv
#include "vml_paths.h"        // isRoot, countPaths
#include "stringset.h"        // StringSet
#include "trace.h"            // trace
#include "vml_build.h"        // p_equal, etc.

#include <string.h>           // strchr


// fwds
void tcheckExprList(VMLEnv &env, ASTList<VExpr> &exprs);


// ------------------ VMLFile ------------------
void VMLFile::tcheck(VMLEnv &env)
{
  // pass 1: build map: name -> VProcedure
  {
    FOREACH_ASTLIST_NC(VTopForm, forms, iter) {
      VTopForm *tf = iter.data();
      if (tf->isVTF_proc()) {
        VProcedure *p = tf->asVTF_proc()->p;
        env.addProcedure(p);
      }
    }
  }

  // pass 2: rest of tcheck 
  FOREACH_ASTLIST_NC(VTopForm, forms, iter) {
    iter.data()->tcheck(env);
  }
}


// ------------------- VTopForm -------------------
void VTF_decl::tcheck(VMLEnv &env)
{
  d->tcheck(env, VK_GLOBAL);
}

void VTF_inv::tcheck(VMLEnv &env)
{
  p->tcheck(env);
}

void VTF_axiom::tcheck(VMLEnv &env)
{
  p->tcheck(env);
}

void VTF_proc::tcheck(VMLEnv &env)
{
  p->tcheck(env);
}

void VTF_comment::tcheck(VMLEnv &env)
{}


// ---------------- Identifier -------------------
bool Identifier::tcheck(VMLEnv &env)
{
  var = env.lookup(name);
  if (!var) {
    env.error(loc, stringc << "undeclared: " << name);
    return false;
  }              
  else {
    return true;
  }
}

void tcheckIdList(VMLEnv &env, ASTList<Identifier> &ids)
{
  FOREACH_ASTLIST_NC(Identifier, ids, iter) {
    Identifier *id = iter.data();
    if (id->tcheck(env)) {
      if (!id->var->isVD_var()) {
        env.error(id->loc, stringc << id->name << " must be a variable");
      }
    }
  }
}


void Identifier::printExtras(ostream &os, int indent) const
{
  if (var && var->id != this) {
    ind(os, indent) << "var: refers to " << toString(var->id->loc) << "\n";
  }
}


// ------------------ VDeclaration -----------------
void VDeclaration::tcheck(VMLEnv &env, VarKind vk)
{
  // subclass-specific checks

  if (isVD_var()) {
    VD_var *v = asVD_var();
    v->vkind = vk;

    if (vk != VK_LOCAL) {
      // for now, all non-locals use themselves as their alpha; locals
      // have their alpha set below, by Func1Pass
      v->alpha = id->name;
    }

    if (v->init) {
      v->init->tcheck(env);
    }
  }

  else if (isVD_const()) {
    asVD_const()->value->tcheck(env);
  }

  else {
    xassert(isVD_funcsym());
    // nothing to check
  }

  // generic declaration processing

  if (env.addVar(id->name, this)) {
    id->var = this;
    return;
  }        

  if (vk == VK_RETVAL) {
    // collision must have been with another parameter
    VD_var *prior = env.lookup(id->name)->asVD_var();
    if (prior->vkind == VK_PARAM) {
      // this repeated name indicates that the parameter is both
      // and "in" and an "out"; we change the existing one, and
      // ignore 'this' one
      prior->vkind = VK_IOPARAM;
      return;
    }
  }

  env.error(id->loc, stringc << "multiply-declared: " << id->name);
}


void tcheckDeclList(VMLEnv &env, ASTList<VD_var> &vars, VarKind vk)
{
  FOREACH_ASTLIST_NC(VD_var, vars, iter) {
    iter.data()->tcheck(env, vk);
  }
}


void VD_var::printExtras(ostream &os, int indent) const
{
  if (vkind != VK_NONE) {
    ind(os, indent) << "vkind: " << toString(vkind) << "\n";
  }
  if (alpha) {
    ind(os, indent) << "alpha: " << alpha << "\n";
  }
}


// ---------------------- VProcedure -----------------------
// first pass visitor:
//   - finds labels
//   - finds path roots
//   - finds local variables
//   - computes alpha names for locals
class ProcPass1 : public VMLVisitor {
public:
  VMLEnv &env;
  VProcedure *proc;
  StringSet localNames;

public:
  ProcPass1(VMLEnv &e, VProcedure *p)
    : env(e),
      proc(p),
      localNames()        // initially empty
  {}
  virtual ~ProcPass1() {} // I hate that stupid warning...

  virtual bool visitVStatement(VStatement *s);

  bool collides(rostring name)
  {
    return env.lookup(name)!=NULL ||      // collides with global
           localNames.contains(name);     // collides with local
  }
};

string maybeAddUnderscore(char const *src)
{
  if (strchr(src, '_')) {
    return stringc << src << "_";
  }
  else {
    return string(src);
  }
}

bool ProcPass1::visitVStatement(VStatement *s)
{
  if (s->isVS_label()) {
    env.addLabel(s->asVS_label());
  }

  if (s->isVS_decl()) {
    VS_decl *d = s->asVS_decl();
    if (d->d->isVD_var()) {
      VD_var *v = d->d->asVD_var();
      //xassert(v->vkind == VK_LOCAL);    // will be true later, but it hasn't been tcheck'd

      // alpha conversion: come up with a name that will not collide
      // with any global, nor any local that we've already seen in
      // this procedure
      string alpha;
      if (!collides(alpha)) {
        // nominal alpha is the original name
        //
        // now, a little trick: if the user's name contained any
        // underscores, then add one more; this ensures that no user
        // or alpha name matches "_[^_]+", so they cannot collide with
        // names manufactured by vcgen
        alpha = maybeAddUnderscore(v->id->name);
      }
      else {
        // collides with something; start adding numbers
        int n = 2;
        do {
          alpha = maybeAddUnderscore(stringc << v->id->name << n);
          n++;
        } while (collides(alpha));
      }

      v->alpha = env.str(alpha);
      localNames.add(v->alpha);

      proc->locals.append(v);
    }
  }

  if (isRoot(s)) {
    proc->roots.append(s);
  }

  return true;
}


class FindWrittenGlobals : public VMLVisitor {
public:      // data
  SObjList<VD_var> &writtenGlobals;
  ASTList<Identifier> const &writes;
  
public:      // funcs
  FindWrittenGlobals(SObjList<VD_var> &dest, ASTList<Identifier> const &w)
    : writtenGlobals(dest), writes(w) {}
  
  virtual bool visitVStatement(VStatement *obj);
};

bool FindWrittenGlobals::visitVStatement(VStatement *obj)
{
  if (!obj->isVS_assign()) {
    return true;      // examine children
  }
  VS_assign *as = obj->asVS_assign();

  // is 'target' a global variable not in 'writes'?
  VD_var *var = as->target->var->asVD_var();
  if (var->vkind == VK_GLOBAL) {
    FOREACH_ASTLIST(Identifier, writes, iter) {
      if (iter.data()->name == var->id->name) {
        // it does appear in 'writes', so ignore it
        return true;
      }
    }

    // not mentioned in 'writes'; add it to 'writtenGlobals', if not
    // already present
    writtenGlobals.prependUnique(var);
  }
  
  return true;
}


string cfgLabel(VStatement *s);     // vml_env.cc

void VProcedure::tcheck(VMLEnv &env)
{
  env.enterScope();

  // check the reads/writes before the params/returns are added
  // to the environment, to ensure they refer to globals only
  tcheckIdList(env, reads);
  tcheckIdList(env, writes);

  tcheckDeclList(env, params, VK_PARAM);

  // do this now, before looking at 'returns', so that the
  // precondition can't talk about them
  if (pre) { pre->tcheck(env); }

  tcheckDeclList(env, returns, VK_RETVAL);

  // TODO: what if this mentions a parameter that has been modified,
  // without qualifying it with 'pre'?  that would be bad
  if (post) { post->tcheck(env); }

  if (body) {
    // do a pre-scan to get all the goto labels and path roots
    ProcPass1 vis(env, this);
    roots.append(body);            // first root
    body->traverse(vis);           // labels + remaining roots

    if (env.cfgGraph) {
      cout << "digraph \"" << id->name << "\" {\n";
      cout << "  \"start\" -> " << cfgLabel(body) << "\n";
    }

    // scan body normally
    body->tcheck(env);

    // make a list of globals that are modified and *not* mentioned
    // in the 'writes' annotation
    SObjList<VD_var> writtenGlobals;
    {
      FindWrittenGlobals fwg(writtenGlobals, writes);
      traverse(fwg);

      // print results to a trace string
      stringBuilder sb;
      sb << id->name << "(): written globals not in 'writes':";
      if (writtenGlobals.isEmpty()) {
        sb << " (none)";
      }
      SFOREACH_OBJLIST(VD_var, writtenGlobals, iter) {
        sb << " " << iter.data()->id->name;
      }
      trace("writes") << sb << endl;
    }

    // add the assertion 'foo = pre(foo)' at end of procedure for
    // every variable 'foo' in 'writtenGlobals'
    if (writtenGlobals.isNotEmpty()) {
      // build a conjunction of equalities
      Pred *p = NULL;
      SFOREACH_OBJLIST(VD_var, writtenGlobals, globIter) {
        VD_var const *var = globIter.data();

        Pred *q = p_equal(new VE_var(var->id),
                          new VE_pre(new VE_var(var->id)));
        if (!p) {
          p = q;
        }
        else {
          p = p_and(p, q);
        }
      }
      p->tcheck(env);

      // add assertion at end
      VStatement *final = new VS_pred(id->loc, PS_ASSERT, p);
      final->tcheck(env);
      body->asVS_compound()->body.append(final);
    }

    env.clearLabelInfo();

    env.clearPendingNexts();

    if (env.cfgGraph) {
      cout << "}\n";
    }

    // do this after computing all the 'next's, since it
    // depends on them
    countPaths(env, this);
  }

  env.exitScope();
}


// ---------------------- VStatement ---------------------
void VStatement::tcheck(VMLEnv &env)
{
  // any pending-nexts go to me
  env.resolvePendingNexts(this);
  
  // my 'next' goes to whoever comes next
  env.addPendingNext(this);
  
  // run per-statement checker
  itcheck(env);
}

void VStatement::printExtras(ostream &os, int indent) const
{
  if (next) {
    ind(os, indent) << "next: refers to " << next->kindName()
                    << " at " << toString(next->loc) << "\n";
  }
  if (paths) {
    ind(os, indent) << "paths: " << paths << "\n";
  }
}


void VS_comment::itcheck(VMLEnv &env)
{}

void VS_decl::itcheck(VMLEnv &env)
{
  d->tcheck(env, VK_LOCAL);
}

void VS_pred::itcheck(VMLEnv &env)
{
  p->tcheck(env);
}

void VS_assign::itcheck(VMLEnv &env)
{
  if (target->tcheck(env)) {
    if (!target->var->isVD_var()) {
      env.error(target->loc, stringc << target->name << " must be a variable");
    }
  }

  source->tcheck(env);
}

void VS_procCall::itcheck(VMLEnv &env)
{
  tcheckIdList(env, targets);

  // find the procedure to which the name refers
  proc = env.lookupProcedure(procName->name);
  if (!proc) {
    env.error(procName->loc, stringc << "undefined procedure: " << procName->name);
  }

  tcheckExprList(env, args);

  // check that the # of in and out params is right
  if (proc) {
    if (proc->returns.count() != targets.count()) {
      env.error(procName->loc, stringc
        << procName->name << " returns "
        << proc->returns.count() << " values, but was given "
        << targets.count() << " targets");
    }

    if (proc->params.count() != args.count()) {
      env.error(procName->loc, stringc
        << procName->name << " expects "
        << proc->params.count() << " parameters, but was given "
        << args.count() << " arguments");
    }
  }
}

void VS_procCall::printExtras(ostream &os, int indent) const
{
  VStatement::printExtras(os, indent);
  if (proc) {
    ind(os, indent) << "proc: refers to " << toString(proc->id->loc) << "\n";
  }
}

void VS_if::itcheck(VMLEnv &env)
{
  env.pushPendingNexts();

  p->tcheck(env);
  sThen->tcheck(env);

  env.popAndMergeNexts();     // pending-next is mine plus sThen's
  env.pushPendingNexts();

  if (sElse) { sElse->tcheck(env); }

  env.popAndMergeNexts();     // mine, sThen's, sElse's
}

void VS_loop::itcheck(VMLEnv &env)
{
  body->tcheck(env);

  // end of the body points back to me
  env.resolvePendingNexts(this);
}

void VS_label::itcheck(VMLEnv &env)
{
  // note: the list of labels is already recorded (by ProcPass1)
}

void VS_goto::itcheck(VMLEnv &env)
{
  label = env.lookupLabel(id->name);
  if (!label) {
    env.error(id->loc, stringc << "no such label: " << id->name);
  }
  else {
    // my next goes to 'label'
    env.resolvePendingNexts(label);
  }
}

void VS_goto::printExtras(ostream &os, int indent) const
{
  VStatement::printExtras(os, indent);
  if (label) {
    ind(os, indent) << "label: refers to " << toString(label->id->loc) << "\n";
  }
}

void VS_compound::itcheck(VMLEnv &env)
{
  FOREACH_ASTLIST_NC(VStatement, body, iter) {
    iter.data()->tcheck(env);
  }
}

#ifdef VML_PAT_INCLUDED
void VS_patvar::itcheck(VMLEnv &bp)
{
  // nothing to check
}
#endif // VML_PAT_INCLUDED


// ------------------ Pred ----------------------
void P_lit::tcheck(VMLEnv &env)
{}

void P_not::tcheck(VMLEnv &env)
{
  p->tcheck(env);
}

void P_quant::tcheck(VMLEnv &env)
{
  env.enterScope();
  env.quantifiers++;

  tcheckDeclList(env, vars, VK_QUANT);
  if (patterns) {
    patterns->tcheck(env);
  }
  body->tcheck(env);

  env.quantifiers--;
  env.exitScope();
}

void P_binary::tcheck(VMLEnv &env)
{
  p1->tcheck(env);
  p2->tcheck(env);
}
   

void P_compare::tcheck(VMLEnv &env)
{
  e1->tcheck(env);
  e2->tcheck(env);
}

void P_declare_partial_order::tcheck(VMLEnv &env)
{
  // nothing to check
}

#ifdef VML_PAT_INCLUDED
void P_patvar::tcheck(VMLEnv &bp)
{
  // nothing to check
}
#endif // VML_PAT_INCLUDED


// -------------------------- VExpr ----------------------
void VE_lit::tcheck(VMLEnv &env)
{}

void VE_var::tcheck(VMLEnv &env)
{
  preState = (env.preState > 0);

  if (id->tcheck(env)) {
    if (id->var->isVD_var() ||
        id->var->isVD_const()) {
      // ok
    }
    else {
      env.error(id->loc, stringc << id->name << " must be a variable or constant");
    }
  }
}

#if 0    // delete me
void VE_var::printExtras(ostream &os, int indent) const
{
  ind(os, indent) << "preState: " << preState << "\n";
}               
#endif // 0

void VE_neg::tcheck(VMLEnv &env)
{
  e->tcheck(env);
}

void tcheckExprList(VMLEnv &env, ASTList<VExpr> &exprs)
{
  FOREACH_ASTLIST_NC(VExpr, exprs, iter) {
    iter.data()->tcheck(env);
  }
}

void VE_apply::tcheck(VMLEnv &env)
{
  if (func->tcheck(env)) {
    if (!func->var->isVD_funcsym()) {
      env.error(func->loc, stringc << func->name << " must be a function symbol");
    }
  }

  // TODO: check that the # of args corresponds to the # of params!
  // (Note that I would have to add an exception for MPAT and PATS,
  // or else handle them better in general.)

  tcheckExprList(env, args);
}

void VE_arith::tcheck(VMLEnv &env)
{
  e1->tcheck(env);
  e2->tcheck(env);
}

void VE_pre::tcheck(VMLEnv &env)
{
  env.preState++;
  e->tcheck(env);
  env.preState--;
}

void VE_cond::tcheck(VMLEnv &env)
{
  p->tcheck(env);
  a->tcheck(env);
  b->tcheck(env);
}

void VE_pat::tcheck(VMLEnv &env)
{
  if (env.quantifiers <= 0) {
    env.error("trigger patterns can only be used inside quantifiers");
  }
  e->tcheck(env);
}

#ifdef VML_PAT_INCLUDED
void VE_patvar::tcheck(VMLEnv &bp)
{
  // nothing to check
}
#endif // VML_PAT_INCLUDED


// EOF
