// vml_tocnf.cc
// code for vml_tocnf.h

// Unfortunately, the efficiency of this module is hampered by 
// a naive allocation strategy for Conj and Disj objects, and
// a failure to recognize equivalent literals.  At some point
// I'd like to come back and address these shortcomings.


#define VML_TOCNF_IMPLEMENTATION
#include "vml_tocnf.h"        // this module
#include "vml.ast.gen.h"      // Pred
#include "trace.h"            // tracingSys

#include <iostream.h>         // cout


// --------------------- statistics -----------------------
#ifndef DO_STATS
  #define DO_STATS 1
#endif

#if DO_STATS
  #define STATS(stmt) stmt
#else
  #define STATS(stmt) /*nothing*/
#endif

STATS( static int newConjCt = 0; )
STATS( static int newDisjCt = 0; )
STATS( static int newListNodeCt = 0; )



// ------------------------- toCNF ------------------------
Pred * owner toCNF(Pred const *input, bool negate)
{
  STATS( int origNewConjCt = newConjCt; )
  STATS( int origNewDisjCt = newDisjCt; )
  STATS( int origNewListNodeCt = newListNodeCt; )
  STATS( static bool emitStats = tracingSys("cnfStats"); )

  VML2CNF vml2cnf;
  vml2cnf.makeLiteralTable(input);
  Conj * owner cnf = vml2cnf.toCNF(input, negate);
  Pred * owner ret = vml2cnf.materialize(cnf);
  delete cnf;

  #if DO_STATS
    if (emitStats) {
      cout << "conj: " << (newConjCt-origNewConjCt) << "\n"
           << "disj: " << (newDisjCt-origNewDisjCt) << "\n"
           << "node: " << (newListNodeCt-origNewListNodeCt) << "\n"
           << "lits: " << vml2cnf.numLiterals()-1 << "\n"  // other than true/false
           ;
    }
  #endif

  return ret;
}


// ------------------------- Disj -------------------------
Disj::Disj(int numLiteralCodes)
  : bits(numLiteralCodes)
{}


Disj::~Disj()
{}


// ------------------------- Conj -------------------------
Conj::Conj()
  : set(),       // initially empty
    consistent(true)
{}

Conj::~Conj()
{}


void Conj::addConjunct(Disj *d)
{
  if (consistent) {
    // any identical conjuncts?
    FOREACH_OBJLIST(Disj, set, iter) {
      if (iter.data()->equals(*d)) {
        delete d;     // already present
        return;
      }
    }
    
    // no, add it
    set.prepend(d);
    STATS( newListNodeCt++; )
  }
  else {
    // no point in further constraining an already inconsistent set
    delete d;
  }
}


void Conj::conjoinAndDelete(Conj *c)
{ 
  if (consistent && c->consistent) {
    // transfer each conjunct; slow O(n^2) algorithm.. the way to
    // improve this is to maintain Conj::set in a sorted order.
    while (c->set.isNotEmpty()) {
      Disj *d = c->set.removeFirst();
      addConjunct(d);
    }
  }
  else {
    makeInconsistent();
  }


  delete c;
}


void Conj::disjoinAndDelete(Conj *c)
{                  
  if (!c->consistent) {
    // do nothing
  }
  else if (!consistent) {
    // take c's stuff
    xassert(set.isEmpty());
    set.concat(c->set);
    consistent = true;
  }
  else {
    // for each pair of conjuncts, disjoin them and add to 'product'
    ObjList<Disj> product;
    FOREACH_OBJLIST(Disj, this->set, iter1) {
      FOREACH_OBJLIST(Disj, c->set, iter2) {
        Disj *d = new Disj(*(iter1.data()));
        STATS( newDisjCt++; )
        d->unionWith(*(iter2.data()));
        
        // Argh.  Because I do not recognize when literals (that are
        // presented using separate Pred objects) are equivalent, I
        // don't think I ever get to recognize tautologies during
        // translation... that is, I cannot find an input that causes
        // 'd->tautological()' to be true at this point.

        if (d->tautological()) {
          // no point in keeping a conjunct that is always true
          cout << "*** got a tautology!\n";
          delete d;                                            
        }
        else {
          product.prepend(d);
          STATS( newListNodeCt++; )
        }
      }
    }

    // throw away the sources
    set.deleteAll();

    // save 'product'
    set.concat(product);
  }

  delete c;
}


// ----------------------- VML2CNF -----------------------
VML2CNF::VML2CNF()
  : literalTable(),
    literalCodeMap(),
    tableLocked(false)
{
  // code 1 is always 'true' and code 0 is always 'false'
  literalTable.push(NULL);
}

VML2CNF::~VML2CNF()
{}


LiteralCode VML2CNF::encode(Pred const *p, bool negate)
{
  if (p->isP_not()) {
    return encode(p->asP_notC()->p, !negate);
  }

  if (p->isP_lit()) {
    int ret = p->asP_litC()->b? 1 : 0;
    return negate? 1-ret : ret;
  }

  int index = (int)literalCodeMap.get((void*)p);
  if (!index) {
    // must add a code for this
    xassert(!tableLocked);
    index = literalTable.length();
    literalTable.push(p);
    literalCodeMap.add((void*)p, (void*)index);
  }

  return negate? index*2+1 : index*2;
}


void VML2CNF::makeLiteralTable(Pred const *input)
{
  makeLT_helper(input);
  tableLocked = true;
}

void VML2CNF::makeLT_helper(Pred const *input)
{
  ASTSWITCHC(Pred, input) {
    ASTCASEC1(P_lit)
      break;    // 0 and 1 already assigned

    ASTNEXTC(P_not, n)
      makeLT_helper(n->p);

    ASTNEXTC(P_quant, q)
      encode(q, false /*negate*/);

    ASTNEXTC(P_binary, b)
      makeLT_helper(b->p1);
      makeLT_helper(b->p2);

    ASTNEXTC(P_compare, c)
      // TODO: Now that P_compare is broken down, do something smarter here.
      encode(c, false /*negate*/);

    ASTDEFAULTC
      xfailure("bad predicate code");
      
    ASTENDCASEC
  }
}


Disj *VML2CNF::makeDisj()
{           
  STATS( newDisjCt++; )
  return new Disj(literalTable.length() * 2);
}

Conj *VML2CNF::literalToConj(LiteralCode code)
{
  Conj *conj = new Conj;
  STATS( newConjCt++; )
  
  if (code == 0) {         // "false"
    conj->makeInconsistent();
  }
  else if (code == 1) {    // "true"
    // leave the conjunction empty
  }
  else {
    Disj *disj = makeDisj();
    disj->addDisjunct(code);
    conj->addConjunct(disj);
    
    // I do not test whether adding the disjunct makes disjunction
    // tautological because a single disjunct is never tautological
  }

  return conj;
}


// this function is the heart of the algorithm
Conj * owner VML2CNF::toCNF(Pred const *input, bool negate)
{
  ASTSWITCHC(Pred, input) {
    ASTCASEC(P_lit, l)
      return literalToConj(encode(l, negate));

    ASTNEXTC(P_not, n)
      return toCNF(n->p, !negate);

    ASTNEXTC(P_quant, q)
      return literalToConj(encode(q, negate));

    ASTNEXTC(P_binary, b) {
      if (b->con == BP_AND || b->con == BP_OR || b->con == BP_IMPLIES) {
        // nominal combination mode: AND
        BPropCon finalOp = BP_AND;
        bool negateP1 = false;
        bool negateP2 = false;

        // take b->con into account
        if (b->con == BP_OR) {
          finalOp = BP_OR;
        }
        else if (b->con == BP_IMPLIES) {
          finalOp = BP_OR;
          negateP1 = true;
        }
        
        // take 'negate' into account (DeMorgan)
        if (negate) {
          finalOp = (finalOp==BP_AND)? BP_OR : BP_AND;
          negateP1 = !negateP1;
          negateP2 = !negateP2;
        }
        
        // recursively convert
        Conj *c1 = toCNF(b->p1, negateP1);
        Conj *c2 = toCNF(b->p2, negateP2);

        // combine
        if (finalOp == BP_AND) {
          c1->conjoinAndDelete(c2);
        }
        else {
          c1->disjoinAndDelete(c2);
        }
        return c1;
      }

      else if (b->con == BP_IFF) {
        // convert each argument to CNF in both senses
        Conj *c1p = toCNF(b->p1, false /*negate*/);
        Conj *c1n = toCNF(b->p1, true /*negate*/);
        Conj *c2p = toCNF(b->p2, false /*negate*/);
        Conj *c2n = toCNF(b->p2, true /*negate*/);
        
        // combine
        if (!negate) {
          //   A <=> B          -->      (B || !A) && (A || !B)
          c2p->disjoinAndDelete(c1n);        // B || !A
          c1p->disjoinAndDelete(c2n);        // A || !B
          c1p->conjoinAndDelete(c2p);
          return c1p;
        }
        else {
          //   !(A <=> B)       -->      (B || A) && (!A || !B)
          c1p->disjoinAndDelete(c2p);        // A || B
          c1n->disjoinAndDelete(c2n);        // !A || !B
          c1p->conjoinAndDelete(c1n);
          return c1p;
        }
      }

      else {
        xfailure("bad BPropCon");
      }
    }

    ASTNEXTC(P_compare, c)
      return literalToConj(encode(c, negate));

    ASTDEFAULTC
      xfailure("bad predicate code");
      
    ASTENDCASEC
  }
}


Pred * owner VML2CNF::materialize(Conj const *conj)
{
  if (!conj->consistent) {
    return new P_lit(SL_UNKNOWN, false);
  }

  if (conj->set.isEmpty()) {
    return new P_lit(SL_UNKNOWN, true);
  }

  Pred *ret = materializeDisj(conj->set.firstC());
  SourceLoc firstLoc = ret->loc;

  // assemble a sequence of AND nodes
  ObjListIter<Disj> iter(conj->set);
  iter.adv();
  for (; !iter.isDone(); iter.adv()) {
    Pred *next = materializeDisj(iter.data());
    ret = new P_binary(firstLoc, ret, BP_AND, next);
  }

  return ret;
}


Pred * owner VML2CNF::materializeDisj(Disj const *disj)
{
  Pred *ret = NULL;
  SourceLoc firstLoc = SL_UNKNOWN;

  // assemble a sequence of OR nodes
  for (Disj::Iter iter(*disj); !iter.isDone(); iter.adv()) {
    LiteralCode code = iter.data();

    Pred *next = materializeLiteral(code);
    if (!ret) {
      ret = next;
      firstLoc = ret->loc;
    }
    else {
      ret = new P_binary(firstLoc, ret, BP_OR, next);
    }
  }

  xassert(ret);
  return ret;
}


Pred * owner VML2CNF::materializeLiteral(LiteralCode code)
{
  // map code to non-negated Pred
  Pred const *positive = literalTable[code / 2];
  
  // clone
  Pred *ret = positive->clone();

  if (code % 2) {
    // negate and return
    return new P_not(ret->loc, ret);
  }
  else {
    // just return
    return ret;
  }
}


// EOF
