// vml_vcgen.h
// verification condition generation for VML

#ifndef VML_VCGEN_H
#define VML_VCGEN_H

#include <fstream.h>        // ofstream

#include "vml.ast.gen.h"    // VML AST
#include "sobjlist.h"       // SObjList

class StringTable;          // strtable.h
class ProverInterface;      // prv_int.h


class VCEnv {
private:     // data
  // interface to prover
  ProverInterface *prover;  // (nullable owner)

  // 'this', as a reference; useful to allow me to easily move
  // code between this class and other classes
  VCEnv &env;

public:      // data
  // for making new StringRefs
  StringTable &str;

  // the entire VML file
  VMLFile *file;

  // all global variables
  SObjList<VD_var> globals;
  
  // all global invariant predicates
  SObjList<Pred> invariants;
                               
  // this is incremented for each S_funCall on the current path,
  // and is used to effectively alpha-convert params/returns
  int callsite;

  // when true, all calls to 'declVar' are treated as if they
  // were calls to 'paramVar'
  bool callSiteParams;

  // when nonzero, calls to 'declVar' retrieve the pre-state
  // versions of the referred variables
  int preState;

  // assumptions made on the current path; this is used to
  // restore the prover's context to a fresh state, after the
  // path analysis is complete
  int assumptions;

  // counter for generating temporary names
  int tempNameCount;

  // what prover, or predicate language, are we using?
  enum ProverType {
    PT_SIMPLIFY,          // Simplify (the default)
    PT_CVC                // CVC, but really Kettle
  } ptype;

  // when non-NULL, we don't actually send the predicates to Simplify;
  // instead, we just emit them to this file
  ofstream *dryRun;       // (nullable owner)

  // when true, we break the global invariant apart when assuming
  // or proving it
  bool breakApartBigFacts;
  
  // when true, we break apart the small facts
  bool breakApartSmallFacts;

  // number of failed proofs
  int errors;

private:     // funcs
  // direct communication with Simplify
  char const *commentStart() const;
  char const *nlCommentStart() const;    // precede with newline
  void feed(rostring text);
  bool prove(rostring text, string *counterexample = NULL);
  void innerAssume(rostring pred);
  string proofObligation(Pred *p);
  bool verboseVerification();

  void printGoal(rostring label, Pred const *p);
  Pred **findConjunction(Pred **p);
  ProverInterface *getProver();

  void foreachGlobalInvariantConjunct(SourceLoc loc, rostring context, 
                                      bool doAssume);
  void foreachConjunct(Pred *pred, rostring context, bool doAssume,
                       int &counter);
  void proveOrAssumeBigFact(Pred *pred, SourceLoc loc, rostring context,
                            bool doAssume);
public:      // funcs
  VCEnv(StringTable &strTable, VMLFile *file);
  ~VCEnv();

  // output VC to a file, but do not try to prove (set 'ptype' first);
  // if 'fname' is NULL then a default filename is chosen for the log
  void doDryRun(char const *fname = NULL);

  // bracket a path analysis
  void beginPath(VProcedure *p, bool funcStart);
  void endPath();

  // yield the global invariant as a single vcgen'd predicate
  Pred* owner vcgenGlobalInv();

  // given a declaration of a variable or constant, yield a VE_var
  // that uses the name that Simplify knows about for that variable
  // at this time; this function is largely responsible for alpha
  // conversion; it also takes into account the modification count
  VE_var* owner declVar(VDeclaration *d);

  // similar to 'declVar', but for when the declared variable is
  // being used as a function parameter (or return value) at a
  // call site; this difference is important for functions that
  // call themselves
  VE_var* owner paramVar(VDeclaration *d);

  // emit a comment; should not contain a newline nor begin with ';'
  void comment(rostring c);
  void commentLoc(SourceLoc loc, rostring c);

  // variation that prints a BLOCK in CVC mode
  void commentLocBlock(SourceLoc, rostring c);
  void endBlock();    // only has effect in CVC mode

  // variation that also prints to cout, as part of the progress narrative
  void pcomment(rostring c);

  // can have newlines in middle, but not end
  void multilineComment(rostring c);

  // comment with AST element
  void commentPred(Pred const *p);
  void commentStmt(VStatement const *stmt);

  // add an assumption to the current proving context
  void assume(Pred* owner p, SourceLoc loc, rostring why);
  void assume(rostring p, SourceLoc loc, rostring why);

  // prove the given predicate; if it fails, it reports an error
  void prove(Pred* owner p, SourceLoc loc, rostring why);

  // retract some assumptions
  void popAssumptions(int howMany, rostring context, rostring kind);

  // optionally breaks assumptions/goals into constituent conjuncts
  void assumeGlobalInvariant(SourceLoc loc, rostring context);
  void proveGlobalInvariant(SourceLoc loc, rostring context);

  // like above but for any (pre-vcgen) predicate
  void proveBigFact(Pred *pred, SourceLoc loc, rostring context);
  void assumeBigFact(Pred *pred, SourceLoc loc, rostring context);
  
  // v->modct++, plus some extra instrumentation
  void scrambleGlobal(VD_var *v, SourceLoc loc);
};

#endif // VML_VCGEN_H
