// vmlmain.cc            see license.txt for copyright and terms of use
// program that parses VML and emits verification conditions

#include "vml_lexer.h"    // VMLLexer
#include "vml.gr.gen.h"   // VMLParse
#include "vml.ast.gen.h"  // VMLFile, etc.
#include "vml_env.h"      // VMLEnv
#include "vml_vcgen.h"    // VCEnv
#include "vml_tocnf.h"    // VML2CNF
#include "glr.h"          // GLR
#include "trace.h"        // TRACE_ARGS
#include "srcloc.h"       // SourceLocManager
#include "exc.h"          // xbase

#include <iostream.h>     // cout
#include <stdio.h>        // remove


void doit(int argc, char **argv)
{
  char const *progname = argv[0];
  TRACE_ARGS()

  if (argc < 2) {
    cout << "usage: " << progname << " [options] input.vml\n"
         << "  options:\n"
         << "    -tr <flags>     turn on tracing (debugging) flags\n"
         << "    -cvc            emit VCs in CVC instead of Simplify format\n"
         << "    -dry            write VC to file (only); do no proving\n"
         << "    -dryf <file>    like -dry but specify output file\n"
         << "    -breakBF        break big facts apart when emitting\n"
         // -testCNF not documented
         ;
    return;
  }

  #define SHIFT argc--; argv++

  // non -tr arguments
  bool cvc = false;
  bool dry = false;
  char const *dryFname = NULL;
  bool breakBF = false;
  bool testCNF = false;

  while (argv[1][0] == '-') {
    if (0==strcmp(argv[1], "-cvc")) {
      cvc = true;
    }
    else if (0==strcmp(argv[1], "-dry")) {
      dry = true;
    }
    else if (0==strcmp(argv[1], "-dryf")) {
      dry = true;
      dryFname = argv[2];
      if (!dryFname[0]) {
        xbase("-dryf requires an argument");
      }
      SHIFT;
    }
    else if (0==strcmp(argv[1], "-breakBF")) {
      breakBF = true;
    }
    else if (0==strcmp(argv[1], "-testCNF")) {
      testCNF = true;
    }
    else {
      xbase(stringc << "unknown argument: " << argv[1]);
    }

    SHIFT;
  }

  SourceLocManager mgr;
  StringTable strTable;

  // ------------- lex/parse --------------
  VMLFile *file;
  {
    // lexer
    VMLLexer lexer(strTable, argv[1]);
    lexer.tokenFunc(&lexer);

    // parser
    VMLParse parse(strTable);
    GLR glr(&parse, parse.makeTables());

    // parse it
    SemanticValue result;
    if (!glr.glrParse(lexer, result) ||
        lexer.errors) {
      xbase("parse error");
    }

    file = (VMLFile*)result;
  }

  // print abstract syntax tree
  if (tracingSys("printAST")) {
    file->debugPrint(cout, 0);
  }

  // pretty-print
  if (tracingSys("pprint")) {
    cout << file->print() << endl;
  }

  if (tracingSys("stopAfterParse")) {
    // by deleting this, I am also implicitly testing that the
    // tree is well-formed (no sharing), since if it is not, 
    // then I will double-dealloc something (perhaps only visible
    // when my debugging heap is enabled)
    delete file;

    return;
  }

  if (testCNF) {
    // convert every TF_axiom to CNF and then print
    FOREACH_ASTLIST(VTopForm, file->forms, iter) {
      if (!iter.data()->isVTF_axiom()) { continue; }
      VTF_axiom const *a = iter.data()->asVTF_axiomC();
      
      cout << "---------- " 
           << (a->label? a->label : "axiom")
           << " at " << toString(a->loc) 
           << " ----------\n"
           ;
      Pred *cnf = toCNF(a->p);
      cout << cnf->print();
      cout << endl;
      delete cnf;
    }
    return;
  }

  // -------------- tcheck ---------------
  {
    VMLEnv env(strTable);
    file->tcheck(env);

    if (env.numErrors()) {
      xbase("tcheck errors");
    }
  }

  if (tracingSys("printTypedAST")) {
    file->debugPrint(cout, 0);
  }

  if (tracingSys("stopAfterTcheck")) {
    return;
  }

  // -------------- vcgen ----------------
  {
    // let "fflowd" imply "fflow"
    if (tracingSys("fflowd")) {
      traceAddSys("fflow");
    }

    char const *fflowOutput = getenv("FFLOW_OUTPUT");
    if (fflowOutput) {
      // remove the file so the new output does not get
      // appended to the old
      remove(fflowOutput);
    }

    VCEnv vcenv(strTable, file);
    if (cvc) {
      vcenv.ptype = VCEnv::PT_CVC;
    }
    if (dry) {
      vcenv.doDryRun(dryFname);
    }
    if (breakBF) {
      vcenv.breakApartBigFacts = true;
      
      // break them all...
      vcenv.breakApartSmallFacts = true;
    }
    file->vcgen(vcenv);

    if (vcenv.errors) {
      xbase("vcenv errors");
    }
  }
}


int main(int argc, char **argv)
{                                 
  // don't double-print error messages
  xBase::logExceptions = false;

  try {
    doit(argc, argv);
    return 0;
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
}
